{-----------------------------------------------------------------------------
 Unit Name: Antialiasing
 Project Name : Superficies
 Author:    Angel Montesinos-Amilibia
 January 2003--October 2008
 Purpose: Antialiasing the contour of a surface
-----------------------------------------------------------------------------}
{

Superficies, workshop for the geometry of surfaces
Copyright (C) 2003-2008, Angel Montesinos-Amilibia

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation,Inc.,
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

You can email me to:
     montesin@uv.es

Or write me to:
     Angel Montesinos
     Departamento de Geometria y Topologia
     Vicente Andrés Estellés, 1
     46100 Burjasot (Valencia)
     Spain
}


unit Antialiasing;

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

interface
uses
  Forms, GR32;

const
  kMaxWidth = 800;
  kHalfMaxWidth= kMaxWidth div 2;

type
  ToCorrect = record
    i,j: Integer;
    c: TColor32;
  end;

  ColorMap = packed array[0..kMaxWidth-1, 0..kMaxWidth-1] of Word;
  ColorMapPtr = ^ColorMap;

procedure Antialiase(var bm: TBitmap32; Map: ColorMapPtr; tag, width: Integer);

implementation
uses
  sysutils, FGPicker;



var
  top : Integer;
  baseR, baseG, baseB: Byte;
  corrList: array of ToCorrect;
  lastTc: Integer;

procedure AddToList(i,j: Integer; c: TColor32);
begin
  inc(lastTc);
  corrList[lastTc].i:= i;
  corrList[lastTc].j:= j;
  corrList[lastTc].c:= c;
end;
  

procedure ScanRow(var bm: TBitmap32; Map: Colormapptr; n: Integer);
{Scan the 'n'th row of 'Map' to find pixels that must be corrected}
var
  h, hIni,  i, jump: Integer;
  half, a, b, c, cc, cb: Double;
  clr: TColor32;
  rr,gg,bb: Byte;
  line: pColor32Array;
begin
{'h' is the order number of the row that is being currently scanned. Motives
 for correcting a pixel [h,n] are that it belongs to the background and there
 is an adjacent pixel that belong to the surface. In the following we will say
 that a pixel is 'b' (in background) or 's' (in surface)}

 {We do not scan directly the first row nor the last}
  h := 1;
  while h < top do begin
    {If (map^[h, n] = 0, the pixel [h,n] is 'b'}
    if (map^[h, n] = 0) then begin
      {register in 'jump' whether the discontinuity is up or down}
      if (map^[h, n - 1] <> 0) then
        jump := -1
      else if (map^[h, n + 1] <> 0) then
        jump := 1
      else begin {no more action is needed for the moment}
        h := h + 1;
        continue;
      end;
      {if (map^[h - 1, n + jump] = 0), there is a corner in the surface at the
       pixel [h,n+jump]}
      if (map^[h - 1, n + jump] = 0) then begin
        hIni := h;
        repeat
          h := h + 1
       {We find the forward extent of the pattern of having a 'b' point in
        row n and an 's' in the row 'n+jump'}
        until (map^[h, n] <> 0) or (map^[h, n + jump] = 0) or (h = top);
        if (map^[h, n + jump] = 0) then begin
        {in this case the surface is convex and has two corners, one at [hIni,n+jump]
         and the other at [h-1,n+jump]. We  shall correct the 'b' pixels [x,n] from
         [hIni,n] to [h-1,n] by blending the 'b' color 'B' with the color 'C'
         in pixel [x,n+jump] by the formula  f[x]=cc[x]*C + (1-cc[x])*B, where cc[x]=
         a + x*(b+ c*x). The coefficients 'a,b,c' are set as to have cc[hIni-1/2]=
         f[h-1/2]= 0, and cc[(hIni+ h-1)/2] = 0.9}
          half := 3.6 / sqr(h - hIni);
          a := (0.5 - hIni) * (h - 0.5) * half;
          b := (h + hIni - 1) * half;
          c := -half;
          for i := hIni to h - 1 do begin
            line:= bm.scanline[n + jump];
            clr := line^[i];
            rr:= clr and $000000FF;
            gg:= (clr and $0000FF00) shr 8;
            bb:= (clr and $00FF0000) shr 16;
            cc := a + i * (b + c * i);
            cb := 1 - cc;
            rr:= round(rr*cc+ baseR*cb);
            gg:= round(gg*cc+ baseG*cb);
            bb:= round(bb*cc+ baseB*cb);
            {The new color is added to the list of corrections}
            AddToList(i,n, $FF000000 + rr + (gg shl 8) + (bb shl 16));
          end;
        end
        else begin
          {In this case the line of 'b' pixels [x,n] terminates at [h,n] with
          an 's' point. Thus, we correct the points in linear manner}
          if (map^[h, n] = 0) then
            h := h + 1;
          if (h - hIni) > 1 then begin
            half := h - hIni;
            a := (h - 0.5) / half;
            b := -1 / half;
            for i := hIni to (h - 1) do begin
              line:= bm.scanline[n + jump];
              clr := line^[i];
              rr:= clr and $000000FF;
              gg:= (clr and $0000FF00) shr 8;
              bb:= (clr and $00FF0000) shr 16;
              cb := a + i * b ;
              cc := 1 - cb;
              rr:= round(rr*cc+ baseR*cb);
              gg:= round(gg*cc+ baseG*cb);
              bb:= round(bb*cc+ baseB*cb);
              AddToList(i,n, $FF000000 + rr + (gg shl 8) + (bb shl 16));
            end;
          end;
        end;
      end
      else begin
      {Now map^[h - 1, n + jump]<> 0. If map^[h - 1, n]=0 then the pixel map^[h, n]
       would have been processed in an instance of the previous case. Therefore, we
       assume that map^[h - 1, n] is an 's' point}
        hIni := h;
        repeat
          h := h + 1
        until (map^[h, n] <> 0) or (map^[h, n + jump] = 0) or (h =
          top);
        if (map^[h, n + jump] = 0) or (h = top) then begin
          {We use a linear correction}
          if (h = top) then
            h := h + 1;
          if (h - hIni) > 1 then
          begin {6}
            half := h - hIni;
            a := (h - 0.5) / half;
            b := -1 / half;
            for i := hIni to (h - 1) do
            begin {7}
              line:= bm.scanline[n + jump];
              clr := line^[i];
              rr:= clr and $000000FF;
              gg:= (clr and $0000FF00) shr 8;
              bb:= (clr and $00FF0000) shr 16;
              cc := a + i * b;
              cb := 1 - cc;
              rr:= round(rr*cc+ baseR*cb);
              gg:= round(gg*cc+ baseG*cb);
              bb:= round(bb*cc+ baseB*cb);
              AddToList(i, n, $FF000000 + rr + (gg shl 8) + (bb shl 16));
            end;
          end;
        end
        else if (map^[h, n + jump] <> 0) then begin
          {We use a quadratic correction as before but the roles of
           cc and cb are interchanged}
          half := 3.6 / sqr(h - hIni);
          a := (0.5 - hIni) * (h - 0.5) * half;
          b := (h + hIni - 1) * half;
          c := -half;
          for i := hIni to h - 1 do
          begin
            line:= bm.scanline[n + jump];
            clr := line^[i];
            rr:= clr and $000000FF;
            gg:= (clr and $0000FF00) shr 8;
            bb:= (clr and $00FF0000) shr 16;
            cb := a + i * (b + c * i);
            cc := 1 - cb;
            rr:= round(rr*cc+ baseR*cb);
            gg:= round(gg*cc+ baseG*cb);
            bb:= round(bb*cc+ baseB*cb);
            AddToList(i, n, $FF000000 + rr + (gg shl 8) + (bb shl 16));
          end;
        end
        else
      end
    end
    else
      h := h + 1;
  end;
end;

procedure ScanColumn(var bm: TBitmap32; Map: Colormapptr; n: Integer);
{Scan the 'n'th column of 'Map' to find pixels that must be corrected. It
 functions as in the ScanRow procedure.}

var
  v, vIni, i, jump: Integer;
  half, a, b, c, cc, cb: Double;
  clr: TColor32;
  rr,gg,bb: Byte;
  line: pColor32Array;
begin {1}
  v := 1;
  while v < top do
  begin {2}
    if (map^[n, v] = 0) then
    begin {3}
      if (map^[n - 1, v] <> 0) then
        jump := -1
      else if (map^[n + 1, v] <> 0) then
        jump := 1
      else
      begin {4}
        inc(v);
        continue;
      end; {4}
      if (map^[n + jump, v - 1] = 0) then
      begin {4}
        vIni := v;
        repeat
          inc(v)
        until (map^[n, v] <> 0) or (map^[n + jump, v] = 0) or (v =
          top);
        if (map^[n + jump, v] = 0) then
        begin {5}
          half := 3.6 / sqr(v - vIni);
          a := (0.5 - vIni) * (v - 0.5) * half;
          b := (v + vIni - 1) * half;
          c := -half;
          for i := vIni to v - 1 do
          begin {6}
            line:= bm.scanline[i];
            clr := line^[n+jump];
            rr:= clr and $000000FF;
            gg:= (clr and $0000FF00) shr 8;
            bb:= (clr and $00FF0000) shr 16;
            cc := a + i*(b + c * i);
            cb := 1 - cc;
            rr:= round(rr*cc+ baseR*cb);
            gg:= round(gg*cc+ baseG*cb);
            bb:= round(bb*cc+ baseB*cb);
            AddToList(n, i, $FF000000 + rr + (gg shl 8) + (bb shl 16));
          end; {6}
        end {5}
        else
        begin {5}
          if (map^[n, v] = 0) then
            inc(v);
          if (v - vIni) > 0 then
          begin {6}
            half := v - vIni;
            a := (v - 0.5) / half;
            b := -1 / half;
            for i := vIni to (v - 1) do
            begin {7}
              line:= bm.scanline[i];
              clr := line^[n+jump];
              rr:= clr and $000000FF;
              gg:= (clr and $0000FF00) shr 8;
              bb:= (clr and $00FF0000) shr 16;
              cb := a + b * i;
              cc := 1 - cb;
              rr:= round(rr*cc+ baseR*cb);
              gg:= round(gg*cc+ baseG*cb);
              bb:= round(bb*cc+ baseB*cb);
              AddToList(n, i, $FF000000 + rr + (gg shl 8) + (bb shl 16));
            end; {7}
          end; {6}
        end; {5}
      end {4}
      else
      begin {4}
        vIni := v;
        repeat
          inc(v)
        until (map^[n, v] <> 0) or (map^[n + jump, v] = 0) or (v =
          top);
        if (map^[n + jump, v] = 0) or (v = top) then
        begin {5}
          if (v = top) then
            inc(v);
          if (v - vIni) > 0 then
          begin {6}
            half := v - vIni;
            a := (v - 0.5) / half;
            b := -1 / half;
            for i := vIni to (v - 1) do
            begin {7}
              line:= bm.scanline[i];
              clr := line^[n+jump];
              rr:= clr and $000000FF;
              gg:= (clr and $0000FF00) shr 8;
              bb:= (clr and $00FF0000) shr 16;
              cc := a + b * i;
              cb := 1 - cc;
              rr:= round(rr*cc+ baseR*cb);
              gg:= round(gg*cc+ baseG*cb);
              bb:= round(bb*cc+ baseB*cb);
              AddToList(n, i, $FF000000 + rr + (gg shl 8) + (bb shl 16));
            end; {7}
          end; {6}
        end {5}
        else if (map^[n + jump, v] <> 0) then
        begin {5}
          half := 3.6 / sqr(v - vIni);
          a := (0.5 - vIni) * (v - 0.5) * half;
          b := (v + vIni - 1) * half;
          c := -half;
          for i := vIni to v - 1 do
          begin {6}
            line:= bm.scanline[i];
            clr := line^[n+jump];
            rr:= clr and $000000FF;
            gg:= (clr and $0000FF00) shr 8;
            bb:= (clr and $00FF0000) shr 16;
            cb := a +i * (b + c * i);
            cc := 1 - cb;
            rr:= round(rr*cc+ baseR*cb);
            gg:= round(gg*cc+ baseG*cb);
            bb:= round(bb*cc+ baseB*cb);
            AddToList(n, i, $FF000000 + rr + (gg shl 8) + (bb shl 16));
          end; {6}
        end {5}
        else
      end {4}
    end {3}
    else
      inc(v);
  end; {2}
end; {1}

procedure Antialiase(var bm: TBitmap32; map: ColorMapPtr; tag, width: Integer);
{A very rude method of antialiasing the frontiers of the colors of a surface, that
is the "lines" along which the color on the drawing changes abruptly}
var
  n: Integer;
  clr: TColor32;
  line: pColor32Array;
begin
  bm.BeginUpdate;
{corrList is the list of pixels that must be modified}
  SetLength(corrList, 4000);
  lastTc:= -1;
{'baseR', etc. are the components of the background color. They do not depend
 on the background point.}
  clr:= gammas32[tag,0];
  baseR:= (clr and $000000FF);
  baseG:= (clr and $0000FF00) shr 8;
  baseB:= (clr and $00FF0000) shr 16;

{It is assumed that 'map' is a square array, that is its height
equals its 'width'}
  top := width-1;
  for n := 1 to top - 1 do begin
  {Reserve room enough  for the corrections that might appear in the next row}
    if high(corrList)-lastTc < width then
      SetLength(corrList, High(corrList)+ width);
  {Scan the 'n'-th row of 'map' for needed corrections}
    ScanRow(bm, map, n);
  end;

{The same as for the rows but now we scan each column of 'map'}
  for n := 1 to top - 1 do begin
    if high(corrList)-lastTc < width then
      SetLength(corrList, High(corrList)+ width);
    ScanColumn(bm, map, n);
  end;

{If there were corrections, make them to the bitmap 'bm'. We leave 'map' untouched}
  if lastTc >= 0 then
    for n:= 0 to lastTc do begin
      line:=bm.ScanLine[corrList[n].j];
      line^[corrList[n].i]:= corrList[n].c;
    end;
  corrList:= nil;
  bm.EndUpdate;
end;

end.
