(*

MIT License

Copyright (c) [2005 - 2018] [Angel Montesinos-Amilibia]

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*)


unit FGPicker;

//{$mode objfpc}{$H+}
//{$mode delphi}

interface

uses
  SysUtils, Variants, Classes, Graphics, Controls, Forms, Dialogs, ExtCtrls,
  LCLType, StdCtrls, GR32, GR32_Image, BGRABitmapTypes,
  BGRADefaultBitmap, BGRABitmap;
const
  lstCtrl= 127; {last abscissa in the twelwe controls}
  lstHS = 175;
  lstIndex= 255;
  addLight= Single(20.0);
type
  Vector3Single= array [1..3] of Single;
  Vector2Int= array[1..2] of Integer;
  ControlPoint = Record
    pb: TPaintBox32;
    clr: TColor32;
    d: Byte; {in pixels, the x coordinate of the control point}
  end;

  HSLColor= Record
    H, S, L: Single;
  end;

  gamut = array[byte, byte] of TColor32;
  BGRAGamut =  array[byte, byte] of TBGRAPixel;

type

  { TGPicker }
  TGPicker = class(TForm)
    AmbLabel: TLabel;
    ALRed: TPaintBox32;
    ALGreen: TPaintBox32;
    ALBlue: TPaintBox32;

    PLLabel: TLabel;
    PLRed: TPaintBox32;
    PLGreen: TPaintBox32;
    PLBlue: TPaintBox32;

    OReflRed: TPaintBox32;
    OReflGreen: TPaintBox32;
    OReflBlue: TPaintBox32;
    ObverseReflLabel: TLabel;

    RReflRed: TPaintBox32;
    RReflGreen: TPaintBox32;
    RReflBlue: TPaintBox32;
    ReverseReflLabel: TLabel;

    gamutObverse: TPaintBox32;
    ObverseColorsLabel: TLabel;

    GamutReverse: TPaintBox32;
    ReverseColorsLabel: TLabel;


    HSBox: TPaintBox32;
    HuesLabel: TLabel;

    ccBox: TPaintBox32;
    ccLabel: TLabel;

    bcBox: TPaintBox32;
    bcLabel: TLabel;

    procedure ComputeGamuts;
    procedure ComputeBGRAGAmuts;
    procedure CreateAll(Sender: TObject);
    procedure DispatchBackgroundChanged;
    procedure DispatchCurvesColorChanged;
    procedure DispatchGamutsChanged;
    procedure DrawControl(i: Integer);
    procedure DrawGamuts;
    procedure DrawHSandMeters;
    procedure FillHSbm;
    procedure FillMeter(bf: TBitmap32; cl: TColor32);
    procedure MouseDownInControl(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure MouseDownInHSBox(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure MouseDownInMBox(sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure PaintAll(Sender: TObject);
  private
    HSbm: TBitmap32;
    AO, AR, PO, PR: Vector3Single;
    OGamutCorner, RGamutCorner: TPoint;
    cColorSeed, bColorSeed: TColor32;
    colorTrio: array [0..2] of TColor32;
    lambda: array[0..lstIndex] of Byte;
  public
    cHSPoint, bHSPoint: TPoint;
    ccPos, bcPos: Integer;
    PickCtrls: array [0..11] of ControlPoint;
    procedure Initiate;
    destructor Destroy; override;
  end;

  {For the use of B-splines in interpolation, consult the memory
  "AproxFuncBSplinesEnglish.pdf" included in this folder}


  var
    gPicker: TGPicker;
    Gamuts: array[1..2] of Gamut;
    BGRAGamuts: array[1..2] of BGRAGamut;
    cColor, bColor: TColor32;
    cBGRAColor, bBGRAColor: TBGRAPixel;

implementation
{$R *.lfm}
  uses
    math,
    {FMainForm,} FSurface, FOde, ResStrings;

(*procedure Initialize;
begin
  gPicker:= TGPicker.Create(nil);
end;

procedure Finalize;
begin
  gPicker.Free;
end;*)

procedure LittleRhomb(bm: TBitmap32; p: TPoint; clBoundary, clInner: TColor32);
var
  i, j,a: Integer;
begin
  for i:= p.x-3 to p.x+3 do
    for j:= p.y-3 to p.y+3 do
      if (i>=0) and (i < bm.width) and (j >=0) and (j< bm.Height) then begin
        a:= abs(i - p.x) + abs(j-p.y);
        if a = 3 then
          bm.Pixel[i,j]:= clBoundary
        else if a <3 then
          bm.Pixel[i,j]:= clInner;
      end;
end;

procedure LittleX(bm: TBitmap32; p: TPoint);
begin
  bm.lineS(p.x-3, p.y-3, p.x+3, p.y+3, clBlack32, True);
  bm.lineS(p.x-3, p.y+3, p.x+3, p.y-3, clBlack32, True);
  bm.SetPixelT(p.x, p.y, clWhite32);
end;

function RGBtoBGRAColor(cl32: TColor32): TBGRAPixel;
begin
  result:= BGRA((cl32 and $00FF0000) shr 16, (cl32 and $0000FF00) shr 8, cl32 and $000000FF,
                 (cl32  and $FF000000) shr 24);
end;



procedure TGPicker.ComputeBGRAGAmuts;
var
  i, j: Integer;
begin
  for i:= 0 to lstIndex do
    for j:= 0 to lstIndex do begin
      BGRAGamuts[1,i,j]:= RGBtoBGRAColor(Gamuts[1,i,j]);
      BGRAGamuts[2,i,j]:= RGBtoBGRAColor(Gamuts[2,i,j]);
    end;
end;

procedure TGPicker.ComputeGamuts;
var
  i, j, r, g, b: Integer;
  step, alpha, beta, s: Single;
begin
  {the following constant, c = 0.6415, depends on the angle of the
   vectors from the object center to the viewPoint and the
   light source used by unit 'FSurface'. It also depends
   on the formula used in  'TSurface.ReflectedLightNoShadows'.
   See the Mathematica notebook 'CalculosColorPicker'. The
   idea is that for any position of the normal to the
   surface at a fixed point, the resulting indexes given
   by 'ReflectedLightNoShadows' be such that
   'c*(index[i,1] + index[i,2]) <= 1'.}


  s:= 0.6415*lstIndex/sqr(AlRed.width);
  for i:= 0 to 2 do begin
    AO[i+1]:= s*(PickCtrls[i].d * PickCtrls[i+6].d);
    AR[i+1]:= s*(PickCtrls[i].d * PickCtrls[i+9].d);
    PO[i+1]:= s*(PickCtrls[i+3].d * PickCtrls[i+6].d);
    PR[i+1]:= s*(PickCtrls[i+3].d * PickCtrls[i+9].d);
  end;
  step:= 1/lstIndex;
  for j:= 0 to lstIndex do begin
    beta:= j*step;
    for i:= 0 to lstIndex do begin
      alpha:= i*step;
      r:= lambda[min(255, round(AO[1]*alpha + PO[1]*beta))];
      g:= lambda[min(255, round(AO[2]*alpha + PO[2]*beta))];
      b:= lambda[min(255, round(AO[3]*alpha + PO[3]*beta))];
      Gamuts[1,i,j]:= r shl 16+  g shl 8 +  b + $FF000000;
      r:= lambda[min(255, round(AR[1]*alpha + PR[1]*beta))];
      g:= lambda[min(255, round(AR[2]*alpha + PR[2]*beta))];
      b:= lambda[min(255, round(AR[3]*alpha + PR[3]*beta))];
      Gamuts[2,i,j]:= r shl 16+  g shl 8 +  b + $FF000000;
    end;
  end;
end;


procedure TGPicker.CreateAll(Sender: TObject);
  var
    i, k: Integer;
    iL, c, d, u: Single;
    be: TCustomBackend;
begin
  iL:= 1/lstIndex;
  c:= lstIndex- 3*addLight;
  d:= 2*addLight- lstIndex;
  for i:= 0 to  lstIndex do begin
    u:= i*iL;
    k:= round(addLight + i + sqr(u)*(c + d*u));
    if (k <= lstIndex) then
      lambda[i]:= k
    else
      lambda[i]:= lstIndex;
  end;
  caption:= eShowColorControls;
  colorTrio[0]:= clRed32;
  colorTrio[1]:= $FF00AA00;
  colorTrio[2]:= clBlue32;

  PickCtrls[0].pb:= ALRed;
  PickCtrls[1].pb:= ALGreen;
  PickCtrls[2].pb:= ALBlue;
  PickCtrls[3].pb:= PLRed;
  PickCtrls[4].pb:= PLGreen;
  PickCtrls[5].pb:= PLBlue;
  PickCtrls[6].pb:= OReflRed;
  PickCtrls[7].pb:= OReflGreen;
  PickCtrls[8].pb:= OReflBlue;
  PickCtrls[9].pb:= RReflRed;
  PickCtrls[10].pb:= RReflGreen;
  PickCtrls[11].pb:= RReflBlue;

  for i:= 0 to 11 do begin
    PickCtrls[i].clr:= colorTrio[i mod 3];
    PickCtrls[i].d:= 90 + random(30);
    be:= PickCtrls[i].pb.buffer.Backend;
    PickCtrls[i].pb.buffer.SetSize(128,7);
    DrawControl(i);
  end;

  OGamutCorner:= Point(8,76);
  be:= GamutObverse.buffer.backend;
  GamutObverse.buffer.SetSize(128,128);

  RGamutCorner:= Point(168,76);
  be:= GamutReverse.buffer.backend;
  GamutReverse.buffer.SetSize(128,128);
  ComputeGamuts;
  DrawGamuts;

  be:= ccBox.buffer.Backend;
  ccBox.buffer.SetSize(176, 12);
  be:= bcBox.buffer.Backend;
  bcBox.buffer.SetSize(176, 12);

  HSbm:= TBitmap32.Create;
  be:= HSbm.Backend;
  HSbm.SetSize(176, 176);
  FillHSbm;
  ccPos:= 141;
  bcPos:= 98;
  cHSPoint:= Point(30,45);
  bHSPoint:= Point(112, 89);
  DrawHSandMeters;

  AmbLabel.caption:= eAmb;
  PLLabel.caption:= ePL;
  ObverseColorsLabel.Caption:= eObverseColors;
  ReverseColorsLabel.caption:= eReverseColors;
  ObverseReflLabel.caption:= eObverseRefl;
  ReverseReflLabel.caption:= eReverseRefl;
  HuesLabel.caption:= eHues;
  ccLabel.caption:=  eCurvesColor;
  bcLabel.caption:= eBackColor;


end;

destructor TGPicker.Destroy;
begin
  HSbm.Free;
  inherited destroy;
end;


procedure TGPicker.DispatchBackgroundChanged;
{Inform the owner that user has changed the background color of tab 'tg'}
begin
    Surface.BackgroundChanged;
end;

procedure TGPicker.DispatchCurvesColorChanged;
begin
  {};
end;

procedure TGPicker.DispatchGamutsChanged;
{Inform the owner that user has changed the gamma of tab 'tg'}
begin
    Surface.GammaDataChanged;
end;

procedure TGPicker.DrawControl(i: Integer);
var
  j: Integer;
  bf: TBitmap32;
begin
  bf:= PickCtrls[i].pb.buffer;
  for j:= 0 to 6 do begin
    bf.LineS(0, j, lstCtrl, j, clLtGray32, True);
    bf.LineS(0, j, PickCtrls[i].d, j, PickCtrls[i].clr, True);
  end;
  PickCtrls[i].pb.Repaint;
end;

procedure TGPicker.DrawGamuts;
var
  step: Single;
  i, j, a, b: Integer;
  OLine, RLine: PColor32Array;
begin
  step:= lstIndex/lstCtrl;
  gamutObverse.buffer.BeginUpdate;
  gamutReverse.buffer.BeginUpdate;
  for j:= 0 to lstCtrl do begin
    OLine:= gamutObverse.buffer.scanline[j];
    RLine:= gamutReverse.buffer.scanline[j];
    b:= round(lstIndex-j*step);
    for i:= 0 to lstCtrl do begin
      a:= round(i*step);
      Oline^[i]:= Gamuts[1,a,b];
      Rline^[i]:= Gamuts[2,a,b];
    end;
  end;
  gamutObverse.buffer.EndUpdate;
  gamutReverse.buffer.EndUpdate;
  gamutObverse.Repaint;
  gamutReverse.Repaint;
end;

procedure TGPicker.DrawHSandMeters;
begin
  HSbm.DrawTo(HSBox.buffer, 0,0);
  cColorSeed:= HSbm.Pixels[cHSPoint.x, cHSPoint.y];
  bColorSeed:= HSbm.Pixels[bHSPoint.x, bHSPoint.y];
  LittleRhomb(HSBox.buffer, cHSPoint, clBlack32, clYellow32);
  LittleX(HSBox.buffer, bHSPoint);
  HSBox.Repaint;
  FillMeter(ccBox.buffer, cColorSeed);
  cColor:= ccBox.buffer.Pixels[ccPos,6];
  cBGRAColor:= RGBtoBGRAColor(cColor);
  LittleRhomb(ccBox.buffer, Point(ccPos, 6), clBlack32, clYellow32);
  ccBox.Repaint;
  FillMeter(bcBox.buffer, bColorSeed);
  bColor:= bcBox.buffer.Pixels[bcPos,6];
  bBGRAColor:= RGBtoBGRAColor(bColor);
  LittleX(bcBox.buffer, Point(bcPos, 6));
  bcBox.Repaint;
end;

procedure TGPicker.FillHSbm;
var
  line: pColor32Array;
  i, j: Integer;
  invw, invh, saturation: Single;
begin
  InvW := 1 / HSbm.Width;
  InvH := 1 / HSbm.Height;
  for j := 0 to HSbm.Height - 1 do begin
    Line := HSbm.ScanLine[j];
    saturation := 1 - j * InvH;
    for i := 0 to HSbm.Width - 1 do
      Line^[i] := HSLtoRGB(i * InvW, Saturation, 0.5);
  end;
end;

procedure TGPicker.FillMeter(bf: TBitmap32; cl: TColor32);
var
  i, j: Integer;
  hsl: HSLColor;
  lines: array [0..11] of PColor32;
begin
  bf.beginUpdate;
  RGBToHSL(cl, hsl.H, hsl.S, hsl.l);
  for j:=0 to 11 do
    lines[j]:= @bf.Scanline[j]^[0];
  for i:= 0 to lstHS do begin
    hsl.L:= i/lstHS;
    cl:= HSLtoRGB(hsl.H, hsl.S, hsl.L);
    for j:= 0 to 11 do begin
      inc(lines[j]);
      lines[j]^:= cl;
    end;
  end;
  bf.endUpdate;
end;

procedure TGPicker.Initiate;  {To be used only after opening a file}
var
  i: Integer;
begin
  for i:= 0 to 11 do
    DrawControl(i);
  ComputeGamuts;
  DrawGamuts;
  DispatchGamutsChanged;
end;

procedure TGPicker.MouseDownInControl(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  i, d: Integer;
begin
  i:= (Sender As TPaintBox32).tag;
  if x < 0 then
    d:= 0
  else if x > lstCtrl then
    d:= lstCtrl
  else
    d:= x;
  PickCtrls[i].d := d;
  DrawControl(i);
  ComputeGamuts;
  DrawGamuts;
  DispatchGamutsChanged;
end;

procedure TGPicker.MouseDownInHSBox(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if (Button = mbLeft) then
    cHSPoint:= Point(x,y)
  else if (Button = mbRight) then
    bHSPoint:=  Point(x,y);
  DrawHSandMeters;
end;


procedure TGPicker.MouseDownInMBox(sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if sender = ccBox then begin
    ccPos:= x;
    FillMeter(ccBox.buffer, cColorSeed);
    cColor:= ccBox.buffer.Pixels[ccPos,6];
    cBGRAColor:= RGBtoBGRAColor(cColor);
    littleRhomb(ccBox.buffer, Point(ccPos, 6), clBlack32, clYellow32);
    DispatchCurvesColorChanged;
  end
  else if sender = bcBox then begin
    bcPos:= x;
    FillMeter(bcBox.buffer, bColorSeed);
    bColor:= bcBox.buffer.Pixels[bcPos,6];
    bBGRAColor:= RGBtoBGRAColor(bColor);
    LittleX(bcBox.buffer, Point(bcPos, 6));
    DispatchBackgroundChanged;
  end
  else
    Exit;
  (Sender as TPaintBox32).Repaint;
end;

procedure TGPicker.PaintAll(Sender: TObject);
begin
 {}
end;


(*initialization
Initialize;

finalization
Finalize;*)

end.
