(*

MIT License

Copyright (c) [2005 - 2018] [Angel Montesinos-Amilibia]

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*)

unit FImagesToSave;

{.$MODE Delphi}

interface

uses
  LCLIntf, LCLType,SysUtils, Classes, Graphics, Controls, Types,
  Forms, Dialogs, StdCtrls, ActnList, ExtCtrls, CheckLst,
  BGRABitmap, BGRABitmapTypes;

type
  { TImagesToSave }

  TImagesToSave = class(TForm)
    CancelButton : TButton;
    OKButton : TButton;
    CheckTexts: TCheckListBox;
    CheckImages: TRadioGroup;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    Label5: TLabel;
    TitleEdit: TEdit;
    SubTextEdit: TEdit;
    theBm: TBGRABitmap;
    procedure Cancelled(Sender: TObject);
    procedure ClickOK(Sender: TObject);
    procedure Close(Sender: TObject; var CloseAction: TCloseAction);
    procedure ComputeLongFormula(var strs: TStringList; bm: TBGRABitmap; theText: AnsiString;
                      maxwidth, blanksFst: Integer);
    procedure ImagesClick(Sender: TObject);
    procedure OnActivate(Sender : TObject);
    procedure PrepareTextsBitmapAndRects;
    procedure TextsItemClick(Sender: TObject; Index: integer);
  private
    rct1, rct2: TRect;
  public
    path: TFilename;
    extension: AnsiString;
    IncludeImages: Array[0..2] of boolean; {'.[0].True': curves only; '.[1].True': curves and surface}
    IncludeTextsUnder: array[0..1] of Boolean; {'.[0].True': Include Input, '.[1].True: Include 'underText'}
    SaveImages: Boolean; {True, after OKing 'ImagesToSave' dialog.}
    procedure SaveImagesToFile;
    destructor Destroy; override;
  end;

var
  ImagesToSave : TImagesToSave;

implementation
{$R *.lfm}
uses
  FMainForm, FSurface, Vectors, FOde, FGPicker, ResStrings;

procedure TImagesToSave.ComputeLongFormula(var strs: TStringList; bm: TBGRABitmap;
                  theText: AnsiString; maxwidth, blanksFst: Integer);
{Adds to 'strs' the lines on which 'theText' is wrapped. 'indentFirst' is the
number of characters in which the first line is indented. 'indentOthers', the
number of chars for the indentation of the remaining lines. 'maxWidth' is the
maximum width that a line can have.}

const
  kReturn = char(13);
  kLineFeed = char(10);
  kEscape = char(27);
  kTab = char($09);
  kDot = '.';
  parenth = [')',']','}'];
  kLeftArrow = 37;
  kUpArrow = 38;
  kDownArrow = 40;
  digits =['0','1','2','3','4','5','6','7','8','9'];
  specialchars = ['+','-','*','/','=','.',',',':',';', ' '];
var
  theLine, theBareText, indTxt, aux: AnsiString;
  theChar, auxChar: AnsiChar;
  i, index, textLength, b, previousLineEnd,
     lineWidth: Integer;
begin
  indTxt:= '';
  if blanksFst > 0 then for i:= 1 to blanksFst do
    indTxt:= indTxt + ' ';
  textLength := Length(theText);
  theBareText:= '';
  if textLength > 0 then begin
    for index := 1 to textLength do begin
      theChar := theText[index];

      {Change returns, line feeds and tabs to blanks}
      if not (theChar in [kReturn, kTab, kLineFeed, kEscape]) then
        theBareText := theBareText + theChar
      else if length(theBareText) > 0 then begin
        auxChar:=theBareText[length(theBareText)];
        if (auxchar<>' ') then
          theBareText := theBareText + ' ';
      end;
    end;
    textLength := Length(theBareText);
  end;
  if textLength = 0 then
    Exit;
  lineWidth:= bm.TextSize(theBareText).cx;
  if (lineWidth <= maxWidth) then begin {text has only one line}
    strs.Add(theBareText);
    Exit;
  end;

  b := 0;
  previousLineEnd:= 0;
  theLine:= '';
  lineWidth:= 0;
  aux:= '';
  index:= 0;
  while index < textLength do begin
  {read a character, measure the length of the line if that character were added.
   If the length is greater than 'maxWidth', don't add it, store the line, record
   the position in 'theBareText' of the last added character, initialize
   the new line with the indentation, and repeat until all of the text has been
   stored. There are exceptions to this, as we shall see}
    inc(index);
    theChar:= theBareText[index];
    aux:= aux + theChar;
    lineWidth := bm.TextSize(aux).cx;
    {We encounter now a character that makes possible to cut the text at it.
     We record its position for further use if necessary. Exception: a dot
     that is part of a decimal number.}
    if (lineWidth <= maxWidth) and (theChar in specialchars) and
           not((theChar= '.') and (index > 1) and (theBareText[index-1] in digits) and
               (index < textLength) and (theBareText[index+1] in digits)
               ) then
      b:= index;

    if index = textLength then begin
      {We have reached the end of the text. We don't check if 'lineWidth' is greater than
      'maxWidth'. Simply add the line}
      strs.Add(aux);
    end
    else  if (lineWidth > maxWidth) then begin  {text must be cut now}
      if b <= previousLineEnd then begin
        {there is no good point for breaking the line. We break it at the previous character}
        theLine:= theLine + Copy(theBareText, previousLineEnd + 1, index -1 - PreviousLineEnd);
        strs.Add(theLine);
        theLine:= indTxt;
        aux:= theLine;;
        previousLineEnd:= index - 1;
      end
      else begin {There is a good point for breaking the line at positionn 'b'}
        theChar:= theBareText[b];
        if theChar in ['+', '-', '*', '/', '=', ' '] then begin
          {don't include the special character in the current line}
          theLine:= theLine + Copy(theBareText, previousLineEnd + 1, b -1 - PreviousLineEnd);
          strs.Add(theLine);
          theLine:= indTxt;
          if theChar <> ' ' then  {include the special character in the new line}
            theLine:= theLine + theChar;
          previousLineEnd:= b;
          aux:= theLine;
        end
        else begin {add the current special char and break the line}
          theLine:= theLine + Copy(theBareText, previousLineEnd + 1, b - PreviousLineEnd);
          strs.Add(theLine);
          theLine:= indTxt;
          aux:= indTxt;
          previousLineEnd:= b;
        end;
      end;
      b:= 0; {we must start again looking for good breaking points}
      lineWidth:= bm.TextSize(theLine).cx;
      index:= previousLineEnd;
    end
  end;
end;

procedure TImagesToSave.OnActivate(Sender : TObject);
var
  i: Integer;
begin
  caption:= eSaveImages;
  label2.caption:= eSelectThings;
  label5.caption:= eTit;
  label3.caption:= eMarkImages;
  checkImages.Items[0]:= eCurves;
  checkImages.Items[1]:= eSurface;
  checkImages.Items[2]:= eBoth;
  label4.caption:= eEpilog;
  titleEdit.Text:= '';
  checkTexts.Items[0]:= eInput;
  checkTexts.Items[1]:= eThisText;
  okButton.caption:= eValidar;
  cancelButton.Caption:= eCancel;
  for i:= 0 to 2 do
    IncludeImages[i]:= (CheckImages.ItemIndex = i);
  extension:= '.png';
  for i:= 0 to 1 do
    IncludeTextsUnder[i]:= CheckTexts.checked[i];
  update;
end;



procedure TImagesToSave.PrepareTextsBitmapAndRects;
const
  w1= 1000; {width of square for only one figure}
  w2=  600; {width of square for each of the two figures}
var
  strL: TStringList;
  tx, ty: AnsiString;
  fh, x, y: Single;
  tshtitle, tshUnder, i, maxTextWidth, countTitle: Integer;
begin
  strL:=TStringList.Create;
  strL.Clear;
  theBm:= TBGRABitmap.Create;
  if IncludeTextsUnder[0] or IncludeTextsUnder[1] then
    theBm.SetSize(w1+12, w1+12+100)
  else
    theBm.SetSize(2*w2+24, w2+12+100);
  if IncludeImages[2] then
    fh:= w2/40
  else
    fh:= w1/40;
  if IncludeImages[2] then
    maxTextWidth:= 1000
  else
    maxTextWidth:= 1000;
  theBm.FontName:= 'DejaVu Serif';
  theBm.FontHeight:= round(fh);
  theBm.FontStyle:= [fsBold];
  theBm.FontAntialias:= False;

  tx:=  titleEdit.Text;
  try
    StrL.BeginUpdate;
    if tx <> '' then begin
      ComputeLongFormula(StrL,theBm, tx, maxTextWidth, 0);
      if StrL.count > 0 then
        tshtitle:= round((0.5 + StrL.count)*fh);
    end
    else
      tshtitle:= 0;
    countTitle:= StrL.count;

    theBm.FontStyle:= [];
    tx:= '';
    ty:=  mainForm.f1Memo.Text;
    if IncludeTextsUnder[0] then with MainForm do begin
      case md of
        ExplicitFO: tx:= 'u'' = '+ ty;
        ExplicitSO: tx:= 'u" = '+ ty;
        ImplicitFO, ImplicitSO, PArcLenFO, PArcLenSO: tx:= '0 = ' + ty;
        Hamilton: tx := eHamiltonian+ ty;
        Lagrange: tx:= eLagrangian + ty;
        else
          ;
      end;
      ComputeLongFormula(strL, theBm, tx, maxTextWidth, 5);
      tx:= '';
      ty:= f2Memo.Text;
      case md of
        ExplicitFO: tx:= 'v'' = '+ ty;
        ExplicitSO: tx:= 'v" = '+ ty;
        ImplicitFO, ImplicitSO: tx:= '0 = ' + ty;
        else
          ;
      end;
      ComputeLongFormula(strL, theBm, tx, maxTextWidth, 5);

      tx:= eDominion + uMinEdit.text+ ' < u < ' +  uMaxEdit.text + ' ,  ' +  vMinEdit.text + ' < v <  ' + vMaxEdit.text;
      case MainForm.md of
        ImplicitFO: tx:= tx + ',  max |vel| = ' + wMaxEdit.Text;
        ImplicitSO: tx:= tx + ',  max |accel| = ' + wMaxEdit.Text;
        PArcLenFO:  tx:= tx + ',  max |w| = ' + wMaxEdit.Text;
        PArcLenSO:  tx:= tx + ',  max |x| = ' + wMaxEdit.Text;
        else
          ;
      end;
      ComputeLongFormula(strL, theBm, tx, maxTextWidth, 5);

      tx:= eParameters;
      if pEdit.visible or qEdit.visible then begin
        if pEdit.visible then begin
           tx:= tx + 'p = ' + pEdit.text;
           if qEdit.visible then
             tx:= tx + ',  ';
        end;
        if qEdit.visible then
          tx:= tx + 'q = '+ qEdit.text;
        strL.Add(tx);
      end;
    end;

    if IncludeTextsUnder[1] and (SubTextEdit.text <> '') then
      ComputeLongFormula(strL, theBm, SubTextEdit.text, maxTextWidth, 4);
    if strL.count > countTitle then
      tshUnder:= round((0.5 + strL.count-countTitle)*fh)
    else
      tshUnder:= 0;
  finally
      StrL.EndUpdate;
  end;
  if IncludeImages[0] or IncludeImages[1]  then begin
    theBm.SetSize(w1+12, w1+12 + tshTitle + tshUnder);
    rct1:= Rect(6, tshTitle + 6, w1+6, tshTitle + w1+6);
  end
  else if IncludeImages[2] then begin
    theBm.SetSize(2*w2 + 24, w2 + 12 + tshTitle + tshUnder);
    rct1:= Rect(6, tshTitle + 6, w2+6, tshTitle + w2+6);
    rct2:= Rect(w2+14, tshTitle + 6, 2*w2 + 14, tshTitle + w2+6);
  end;

  theBm.Rectangle(rct1.left-2, rct1.top-2, rct1.right+2, rct1.bottom+2, BGRA(100,100,100) , dmSet);
  if IncludeImages[2] then
    theBm.Rectangle(rct2.left-2, rct2.top-2, rct2.right+2, rct2.bottom+2, BGRA(100,100,100) , dmSet);

  theBm.FontStyle:= [fsBold];
  y:=-fh/2;
  if countTitle > 0 then for i:= 0 to countTitle-1 do begin
     y:= y + fh;
     x:= (theBm.width - theBm.TextSize(StrL.strings[i]).cx)/2;
     theBm.TextOut(x, y, StrL.strings[i], BGRABlack);
  end;
  theBm.FontStyle:= [];
  y:= rct1.bottom-fh/2;
  if (strL.count > countTitle) then for i:= countTitle to strL.count-1 do begin
     y:= y + fh;
     x:= 0 {(theBm.width - theBm.TextSize(strL.strings[i]).cx)/ 2};
     theBm.TextOut(x, y, strL.strings[i], BGRABlack, taLeftJustify);
  end;

  strL.Free;
end;

procedure TImagesToSave.TextsItemClick(Sender: TObject; Index: integer);
begin
  IncludeTextsUnder[index]:= CheckTexts.checked[index];
end;


procedure TImagesToSave.ImagesClick(Sender: TObject);
var
  i: Integer;
begin
  for i:= 0 to 2 do
    IncludeImages[i]:= (CheckImages.ItemIndex = i);
end;




procedure TImagesToSave.Cancelled(Sender: TObject);
begin
  mainForm.Inform(eCancelledByUser, True);
  Hide;
end;

procedure TImagesToSave.ClickOK(Sender: TObject);
begin
  if (IncludeImages[1] or IncludeImages[2]) and (mainForm.md <> pArcLenFO) then
     ShowMessage(eOnlyForALFO)
  else
     SaveImagesToFile;
end;

procedure TImagesToSave.Close(Sender: TObject; var CloseAction: TCloseAction);
begin
  mainForm.Inform(eCancelledByUser, True);
  Hide;
end;

destructor TImagesToSave.Destroy;
begin
  if Assigned(theBm) then
    theBm.Free;
  inherited;
end;

procedure TImagesToSave.SaveImagesToFile;
var
  dlg: TSaveDialog;
begin
  if (IncludeImages[1] or IncludeImages[2]) and (mainForm.md <> pArcLenFO) then
    mainform.Inform(eOnlyForALFO, True)
  else begin
    mainForm.savedImages:= '';
    dlg:= TSaveDialog.Create(nil);
    dlg.DefaultExt := '.png';
    if dlg.Execute then begin
      path:= dlg.Filename;
      if path[Length(path)-3] <> '.' then
        path:= path + '.png';
      PrepareTextsBitmapAndRects;
      theBm.ClipRect:= rct1;
      if IncludeImages[0] or IncludeImages[2] then
        Ode.DrawCurves2DOnBGRAbm(theBm, rct1);
      if IncludeImages[1] or IncludeImages[2] then
        gPicker.ComputeBGRAGAmuts;
      if IncludeImages[1] then
        surface.DrawSurfaceToTBGRABitmap(theBm, rct1)
      else if IncludeImages[2] then begin
        theBm.ClipRect:= rct2;
        surface.DrawSurfaceToTBGRABitmap(theBm, rct2);
      end;
      theBm.SaveToFile(path);
    end
    else
      mainForm.savedImages:= eCancelledByUser;
    dlg.free;
  end;
  theBm.noClip;
  hide;
end;




end.

