(*

MIT License

Copyright (c) [2005 - 2018] [Angel Montesinos-Amilibia]

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*)

unit FInOut;

{$MODE Delphi}

interface

 uses
  SysUtils,
  Vectors;

 const

   {For File writing and reading}
  kFilter = '2ed';
  kCreator = 'AMonAmi';
  kCreatorLength = Length(kCreator);
  kVersion00 = Integer(0);   //first file version; corresponds to ODEinR2 version < 1.5
  kVersion10 = Integer(1);   //second file version: corresponds to ODEinR2 version  >= 1.5
  kVersion20 = Integer(2);   //third file version, march 2018.
  kVersion30 = Integer(3);
{----------------------}

function ReadFromDisk(pathAndName: TFilename; var mode: Modes): Boolean;
procedure WriteToDisk(pathAndName: TFilename);


implementation

uses
  Classes, Dialogs, Graphics,
  FMainForm, ResStrings, FOde, FGPicker, FSurface;

function ReadStreamInt(Stream : TStream) : integer;
{returns an integer from the stream}
begin
  result:= 1;
  Stream.Read(Result, SizeOf(Integer));
end;

procedure WriteStreamInt(Stream : TStream; Num : integer);
{writes an integer to the stream}
begin
  Stream.Write(Num, SizeOf(Integer));
end;

function ReadStreamStr(Stream : TStream) : string;
{returns a string from the stream}
var
  StrLen : integer;
begin
  {get length of string}
  StrLen := ReadStreamInt(Stream);
  {get memory for result}
  SetLength(Result, StrLen);
  {read characters}
  if StrLen > 0 then
    Stream.Read(Result[1], StrLen);
end;


procedure WriteStreamStr(Stream : TStream; Str : string);
{writes a string to the stream}
var
  StrLen : integer;
begin
  {get length of string}
  StrLen := Length(Str);
  {write length of string}
  WriteStreamInt(Stream, StrLen);
  {write characters}
  if StrLen > 0 then
    Stream.Write(Str[1], StrLen);
end;

procedure ReportIOError(er: Integer);
var
  //e: Integer;
  m: AnsiString;
begin
  case er of
    -1: m := eUnknownSurfaceType;
    -2: m := eBadFileType;
    -3: m := eFileNotAssigned;
    -4: m := eErrorOnWritingFile;
    -5: m := eErrorOnReadingFile;
    -6: m := eVersionTooNew;
    else begin
      {e := GetError;
      m := SysErrorMessage(e); }
    end;
  end;
  ShowMessage(m);
end;


procedure WriteToDisk(pathAndName: TFilename);
  var
    mst: TMemoryStream;
    i,j,len: Integer;
    str: AnsiString;
begin
  mst:= TMemoryStream.Create;
  {$O-}  {Compiler: cease optimizing code.
          Don't suppress this directive. Otherwise the optimizer might change
          the order in which the things are written to disk and create havoc}
  try
    if pathAndName='' then begin
      ReportIOError(-3);
      MainForm.saved:= False;
    end
    else try
      mst.Seek(0,soFromBeginning);
      len:= Length(kCreator);
      str:= kCreator;
      mst.Write(str[1],len);
      WriteStreamInt(mst, kVersion30);

      MainForm.StrToSave[0]:= IntToStr(ord(MainForm.md));
      for i := 0 to 5 do
        WriteStreamStr(mst,MainForm.StrToSave[i]);
      if  MainForm.wMaxEdit.Enabled then
        WriteStreamStr(mst,MainForm.StrToSave[6])
      else
        WriteStreamStr(mst, '5');
      for i := 7 to kLastStr-1 do
        WriteStreamStr(mst,MainForm.StrToSave[i]);
      if MainForm.md in [ExplicitFO, ExplicitSO, ImplicitFO, ImplicitSO] then
        WriteStreamStr(mst,MainForm.StrToSave[kLastStr])
      else
        WriteStreamStr(mst, 'not used for this ODE type');
      len:= -1;
      with Ode do begin
        if ( High(cv2.ac) >=0 ) and (cv2.last >= 0) then begin
          mst.Write(cv2.last, SizeOf(Integer));
          for i:= 0 to cv2.last do
            if ( High(cv2.ac[i].c ) > 0 ) and (cv2.ac[i].last > 0) then begin
              mst.Write(cv2.ac[i].last, SizeOf(Integer));
              mst.Write(cv2.ac[i].color, SizeOf(TColor));
              for j:= 0 to cv2.ac[i].last do
                mst.Write(cv2.ac[i].c[j], SizeOf(Vector2));
            end
            else
              mst.Write(len,SizeOf(Integer))
        end
        else
          mst.Write(len, SizeOf(Integer));
      end;

      if (MainForm.md = PArcLenFO) then with (Ode as TOdeNf) do begin
        if ( High(cv3.ac) >=0 ) and (cv3.last >= 0) then begin
          mst.Write(cv3.last, SizeOf(Integer));
          for i:= 0 to cv3.last do
            if ( High(cv3.ac[i].c ) > 0 ) and (cv3.ac[i].last > 0) then begin
              mst.Write(cv3.ac[i].last, SizeOf(Integer));
              mst.Write(cv3.ac[i].color, SizeOf(TColor));
              for j:= 0 to cv3.ac[i].last do
                mst.Write(cv3.ac[i].c[j], SizeOf(Vector3));
            end
            else
              mst.Write(len,SizeOf(Integer))
        end
        else
          mst.Write(len, SizeOf(Integer));

        if ( High(cvg.ac) >=0 ) and (cvg.last >= 0) then begin
          mst.Write(cvg.last, SizeOf(Integer));
          for i:= 0 to cvg.last do
            if ( High(cvg.ac[i].c ) > 0 ) and (cvg.ac[i].last > 0) then begin
              mst.Write(cvg.ac[i].last, SizeOf(Integer));
              mst.Write(cvg.ac[i].color, SizeOf(TColor));
              for j:= 0 to cvg.ac[i].last do
                mst.Write(cvg.ac[i].c[j], SizeOf(Vector3));
            end
            else
              mst.Write(len,SizeOf(Integer))
        end
        else
          mst.Write(len, SizeOf(Integer));
        for i:= 0 to 11 do
          mst.Write(gPicker.PickCtrls[i].d, SizeOf(Byte));
        mst.Write(surface.uH, SizeOf(Vector3));
        mst.Write(surface.uV, SizeOf(Vector3));
        mst.Write(surface.uN, SizeOf(Vector3));
        mst.Write(surface.gViewPoint, SizeOf(Vector3));
        mst.Write(surface.gFocus, SizeOf(Vector3));
        mst.Write(surface.uc0, SizeOf(Double));
        mst.Write(gPicker.cHSPoint, SizeOf(TPoint));
        mst.Write(gPicker.bHSPoint, SizeOf(TPoint));
        mst.Write(gPicker.ccPos, SizeOf(Integer));
        mst.Write(gPicker.bcPos, SizeOf(Integer));
      end
      else if (MainForm.md = ImplicitFO) then with (Ode as TOdeIf) do  begin
        if ( High(cvg.ac) >=0 ) and (cvg.last >= 0) then begin
          mst.Write(cvg.last, SizeOf(Integer));
          for i:= 0 to cvg.last do
            if ( High(cvg.ac[i].c ) > 0 ) and (cvg.ac[i].last > 0) then begin
              mst.Write(cvg.ac[i].last, SizeOf(Integer));
              mst.Write(cvg.ac[i].color, SizeOf(TColor));
              for j:= 0 to cvg.ac[i].last do
                mst.Write(cvg.ac[i].c[j], SizeOf(Vector2));
            end
            else
              mst.Write(len,SizeOf(Integer))
        end
        else
          mst.Write(len,SizeOf(Integer));
      end;
      len:= -1;
      with Ode do begin
        if ( High(cvAux.ac) >=0 ) and (cvAux.last >= 0) then begin
          mst.Write(cvAux.last, SizeOf(Integer));
          for i:= 0 to cvAux.last do
            if ( High(cvAux.ac[i].c ) > 0 ) and (cvAux.ac[i].last > 0) then begin
              mst.Write(cvAux.ac[i].last, SizeOf(Integer));
              mst.Write(cvAux.ac[i].color, SizeOf(TColor));
              for j:= 0 to cvAux.ac[i].last do
                mst.Write(cvAux.ac[i].c[j], SizeOf(Vector2));
            end
            else
              mst.Write(len,SizeOf(Integer))
        end
        else
          mst.Write(len, SizeOf(Integer));
      end;

      mst.SaveToFile(pathAndName);
      MainForm.Saved:= True;
    except
      MainForm.Saved:= False;
      ReportIOError(-4);
    end;
  finally
    mst.free;
    {$O+} {Compìler: optimize code}
  end;
end;


function ReadFromDisk(pathAndName: TFilename; var mode: Modes): Boolean;
  var
    mst: TMemoryStream;
    i, j, er, len, version: Integer;
    str: AnsiString;
begin
  er:=0;
  mst:= TMemoryStream.Create;
  {$O-} { Compiler: cease optimizing code.
          Don't suppress this directive. Otherwise the optimizer might change
          the order in which the things are written to disk and create havoc}
  try
    if pathAndName='' then begin
      ReportIOError(-3);
      MainForm.saved:= False;
    end
    else begin
      mst.LoadFromFile(pathAndName);
      mst.Seek(0,soFromBeginning);
      len:= Length(kCreator);
      SetLength(str,len);
      mst.Read(str[1], len);
      if (str <> kCreator) then
        er := -2
      else begin
        mst.Read(version, SizeOf(Integer));
        if version < kVersion20 then begin
          er:= -6;
          exit;
        end;
      MainForm.StrToSave[0]:= ReadStreamStr(mst);
      mode:= modes(StrToInt(MainForm.StrToSave[0]));
      for i := 1 to kLastStr do
        MainForm.StrToSave[i]:= ReadStreamStr(mst);
      with MainForm do begin
        f1Memo.text:= StrToSave[1];
        uMinEdit.text:= StrToSave[2];
        uMaxEdit.text:= StrToSave[3];
        vMinEdit.text:= StrToSave[4];
        vMaxEdit.text:= StrToSave[5];
        if mode in [ImplicitFO, ImplicitSO, PArcLenFO, pArcLenSO] then begin
          wMaxEdit.Enabled:= True;
          wMaxEdit.text:= StrToSave[6];
        end;
        pEdit.Text:= StrToSave[7];
        qEdit.Text:= StrToSave[8];
        if mode in [ExplicitFO, ExplicitSO, ImplicitFO, ImplicitSO] then begin
          f2Label.Visible:= True;
          f2Memo.visible:= True;
          f2Memo.Text:= StrToSave[kLastStr];
        end;
        case mode of
          ExplicitFO: Ode:= TOdeEf.Create;
          ExplicitSO: Ode:= TOdeEs.Create;
          ImplicitFO: Ode:= TOdeIf.Create;
          ImplicitSO: Ode:= TOdeIs.Create;
          Hamilton:   Ode:= TOdeH.Create;
          Lagrange:   Ode:= TOdeL.Create;
          PArcLenFO:  Ode:= TOdeNf.Create;
          PArcLenSO:  Ode:= TOdeNs.Create;
          else
          ;
        end;
      end;
      with Ode do begin
        mst.Read(cv2.last,SizeOf(Integer));
        if cv2.last >= 0 then begin
          if High(cv2.ac) < cv2.last then
            SetLength(cv2.ac, cv2.last+1);
          for i:= 0 to cv2.last do begin
            mst.Read(cv2.ac[i].last,SizeOf(Integer));
            if cv2.ac[i].last >= 0 then begin
              mst.Read(cv2.ac[i].color,SizeOf(TColor));
              if High(cv2.ac[i].c) < cv2.ac[i].last then
                SetLength(cv2.ac[i].c, cv2.ac[i].last+1);
              for j:= 0 to cv2.ac[i].last do
                mst.Read(cv2.ac[i].c[j],SizeOf(Vector2))
            end;
          end;
        end;
        if (mode = PArcLenFO) then with (Ode as TOdeNf) do begin
          mst.Read(cv3.last,SizeOf(Integer));
          if cv3.last >= 0 then begin
            if High(cv3.ac) < cv3.last then
              SetLength(cv3.ac, cv3.last+1);
            for i:= 0 to cv3.last do begin
              mst.Read(cv3.ac[i].last,SizeOf(Integer));
              if cv3.ac[i].last >= 0 then begin
                mst.Read(cv3.ac[i].color,SizeOf(TColor));
                if High(cv3.ac[i].c) < cv3.ac[i].last then
                  SetLength(cv3.ac[i].c, cv3.ac[i].last+1);
                for j:= 0 to cv3.ac[i].last do
                  mst.Read(cv3.ac[i].c[j],SizeOf(Vector3))
              end;
            end;
          end;
          mst.Read(cvg.last,SizeOf(Integer));
          if cvg.last >= 0 then begin
            if High(cvg.ac) < cvg.last then
              SetLength(cvg.ac, cvg.last+1);
            for i:= 0 to cvg.last do begin
              mst.Read(cvg.ac[i].last,SizeOf(Integer));
              if cvg.ac[i].last >= 0 then begin
                mst.Read(cvg.ac[i].color,SizeOf(TColor));
                if High(cvg.ac[i].c) < cvg.ac[i].last then
                  SetLength(cvg.ac[i].c, cvg.ac[i].last+1);
                for j:= 0 to cvg.ac[i].last do
                  mst.Read(cvg.ac[i].c[j],SizeOf(Vector3))
              end;
            end;
          end;
          if version = kVersion30 then begin
             for i:= 0 to 11 do
               mst.Read(gPicker.PickCtrls[i].d, SizeOf(Byte));
             mst.Read(surface.uH, SizeOf(Vector3));
             mst.Read(surface.uV, SizeOf(Vector3));
             mst.Read(surface.uN, SizeOf(Vector3));
             mst.Read(surface.gViewPoint, SizeOf(Vector3));
             mst.Read(surface.gFocus, SizeOf(Vector3));
             mst.Read(surface.uc0, SizeOf(Double));
             mst.Read(gPicker.cHSPoint, SizeOf(TPoint));
             mst.Read(gPicker.bHSPoint, SizeOf(TPoint));
             mst.Read(gPicker.ccPos, SizeOf(Integer));
             mst.Read(gPicker.bcPos, SizeOf(Integer));
          end;
        end;
        if (mode = ImplicitFO) then with (Ode as TOdeIf) do begin
          mst.Read(cvg.last,SizeOf(Integer));
          if cvg.last >= 0 then begin
            if High(cvg.ac) < cvg.last then
              SetLength(cvg.ac, cvg.last+1);
            for i:= 0 to cvg.last do begin
              mst.Read(cvg.ac[i].last,SizeOf(Integer));
              if cvg.ac[i].last >= 0 then begin
                mst.Read(cvg.ac[i].color,SizeOf(TColor));
                if High(cvg.ac[i].c) < cvg.ac[i].last then
                  SetLength(cvg.ac[i].c, cvg.ac[i].last+1);
                for j:= 0 to cvg.ac[i].last do
                  mst.Read(cvg.ac[i].c[j],SizeOf(Vector2))
              end;
            end;
          end;
        end;
        if mst.Position < mst.Size -1 then begin
          mst.Read(cvAux.last,SizeOf(Integer));
            if cvAux.last >= 0 then begin
              if High(cvAux.ac) < cvAux.last then
                SetLength(cvAux.ac, cvAux.last+1);
              for i:= 0 to cvAux.last do begin
                mst.Read(cvAux.ac[i].last,SizeOf(Integer));
                if cvAux.ac[i].last >= 0 then begin
                  mst.Read(cvAux.ac[i].color,SizeOf(TColor));
                  if High(cvAux.ac[i].c) < cvAux.ac[i].last then
                    SetLength(cvAux.ac[i].c, cvAux.ac[i].last+1);
                  for j:= 0 to cvAux.ac[i].last do
                    mst.Read(cvAux.ac[i].c[j],SizeOf(Vector2))
                end;
              end;
            end;
          end;
        MainForm.Saved := True;
      end;
    end;
  end;
  finally
    mst.Free;
    str:= '';
    if er <> 0 then
      ReportIOError(er);
    result:= ( er = 0 );
    {$O+} {Compìler: from now on optimize code}
  end;
end;

end.
