(*

MIT License

Copyright (c) [2005 - 2018] [Angel Montesinos-Amilibia]

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*)

unit FMainForm;

{$mode objfpc}{$H+}


interface

uses
  Classes, SysUtils, FileUtil,
  GR32_Image, Forms, LCLType,
  Controls, Graphics, Dialogs, ExtCtrls, ComCtrls, Menus, StdCtrls, ActnList,
  StdActns, Buttons, BGRABitmapTypes, BGRADefaultBitmap,
  BGRABitmap, GR32, Vectors, Types;

const
    kLastStr = 9;
    kcr= #$0D#$0A;
    klastPt = 23;
    cSurfDim= 364;  {width and height of 'Surface'}
    cLastRow= cSurfDim-1;
    cLastCol= cLastRow;

type

  CrossesSet= Record
    lastCross: Integer;
    basePt: TPoint;
    cr: array of TPoint;
  end;

  { TMainForm }

  TMainForm = class(TForm)
    ActionList1: TActionList;
    EnterButton: TButton;
    EditCopy1: TEditCopy;
    EditCut1: TEditCut;
    EditPaste1: TEditPaste;
    f2Memo: TMemo;
    F1Label: TLabel;
    F2Label: TLabel;
    f1Memo: TMemo;
    MenuItem4: TMenuItem;
    AddParamCurve: TMenuItem;
    DeleteDiscriminant1: TMenuItem;
    DeleteSolCurves: TMenuItem;
    DeleteAuxCurves: TMenuItem;
    SaveImages1: TMenuItem;
    qEdit: TEdit;
    pLabel: TLabel;
    qLabel: TLabel;
    Cut1: TMenuItem;
    MenuItem6: TMenuItem;
    Copy1: TMenuItem;
    DirectionsBox: TPaintBox32;
    MenuItem7: TMenuItem;
    DropLastCurve2DItem: TMenuItem;
    DropLastCurve3DItem: TMenuItem;
    OpenDialog1: TOpenDialog;
    SaveDialog1: TSaveDialog;
    pEdit: TEdit;
    Palettes1: TMenuItem;
    View1: TMenuItem;
    Paste1: TMenuItem;
    SurfaceBox: TPaintBox32;
    uMinEdit: TEdit;
    Curves: TPanel;
    uMaxEdit: TEdit;
    vMinEdit: TEdit;
    vMaxEdit: TEdit;
    wMaxEdit: TEdit;
    uMinwMaxLabel: TLabel;
    MainMenu1: TMainMenu;
    File1: TMenuItem;
    Manual1: TMenuItem;
    MenuItem2: TMenuItem;
    EFO: TMenuItem;
    ESO: TMenuItem;
    IFO: TMenuItem;
    ISO: TMenuItem;
    Ham: TMenuItem;
    Lag: TMenuItem;
    ALFO: TMenuItem;
    ALSO: TMenuItem;
    MenuItem3: TMenuItem;
    Edit1: TMenuItem;
    Quit1: TMenuItem;
    SaveAs1: TMenuItem;
    Save1: TMenuItem;
    Open1: TMenuItem;
    InfoBar: TStatusBar;
    //procedure AddImplicitCurve(Sender: TObject);
    procedure AddParametricCurve(Sender: TObject);
    procedure AdjustMenus;
    procedure AdjustInputPanel;
    procedure AdjustInputPanelAfterOpening;
    procedure BeginWatch;
    procedure ChangeFrame(translation: Boolean);
    procedure ChangeODEType(Sender: TObject);
    function CloseCurrentODE(Sender: TObject): Boolean;
    function ComputeSurface: Boolean;
    procedure DrawCrossesFO(var number: Integer; toErase: Boolean; u,v: Integer);
    procedure DropLastCurve2D(Sender: TObject);
    procedure DropLastCurve3D(Sender: TObject);
    procedure EndWatch;
    procedure FillDirectionsBox(u,v,up,vp: Double);
    function FromParamsToDrawing(u,v: Double): TPoint;
    function FromDrawingToParams(x,y: Integer): Vector2;
    procedure Inform(m: AnsiString; ring: Boolean);
    procedure KeyUpInCurves(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure MouseDownInCurves(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure MouseDownSurface(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure MouseLeaveCurves(Sender: TObject);
    procedure MouseMoveInCurves(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure MouseMoveSurface(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure MouseUpInCurves(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure MouseUpSurface(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure OnChangeInput(sender: TObject);
    procedure OnCloseQueryClick(Sender: TObject; var CanClose: boolean);
    procedure OnCreateForm(Sender: TObject);
    procedure OnDeleteAuxCurves(Sender: TObject);
    procedure OnDeleteDiscriminant(Sender: TObject);
    procedure OnDestroyForm(Sender: TObject);
    procedure OnEnterButton(Sender: TObject);
    procedure OnManual(Sender: TObject);
    procedure OnQuit(Sender: TObject);
    procedure OnSaveAsODE(Sender: TObject);
    procedure OnSaveImages(Sender: TObject);
    procedure OnSaveODE(Sender: TObject);
    procedure OnView(Sender: TObject);
    procedure OpenFile(Sender: TObject);
    procedure PaintCurves(Sender: TObject);
    procedure PrepareEnvironment(fromFile: Boolean);
    function ReadForm(fromFile: Boolean): Boolean;
    procedure RefreshCurves(Sender: TObject);
    procedure UpdateParameters(pp,qq: Double);
  private
    pathAndName: TFilename;
    bmDir: TBitmap32;
    rctCurves, rctSurface, rctDirections: TRect;
    inputDataDirty, tracking, trackingSurface, trackingFrame, thereAreCrosses: Boolean;
    crosses: CrossesSet;
    curPos2D: Vector2;
    firstPt, prevPt, curPt, prevPtSurface, curPtSurface: TPoint;
    span3D, dp1, dp2, dp3, dp4, pd1, pd2, pd3, pd4: Double;
    usedParams: set of AnsiChar;
  public
    savedImages: String;
    p, q, span2D: Double; {values of the parameters of functions}
    surfBuffer, bm32: TBitmap32;
    strToSave: array[0..kLastStr] of AnsiString;
    ma, mi, edges, halfEdges, center: Vector3;
    bmBGRA: TBGRABitmap;
    curvesCan: TCanvas;
    alreadyDrawnSurface, alreadyComputedDiscriminant, stop: Boolean;
    saved, busy, alreadyRead: Boolean;
    md: Modes;
  end;

var
  MainForm: TMainForm;

implementation

{$R *.lfm}

uses
  Math,
  ResStrings, FGPicker, Utilities, FunctionParserBytes_64_Linux, FEvaluatorBytes_64_Linux, FOde,
  FGetPointsInR2, FYesNotCancel, FInOut, FSurface, FImagesToSave, FManODEinR2,
  FParametricCurve;

const
{texts of the functions defining the ODE by default}
  f01= 'u^2 + sen(v)';
  f02= 'uv sqrt(1+v^2)';
  f11=  'u''^2 + v^2 sen(u)';
  f12= 'sen(v)/sqrt(1+u''^2+ v''^2)';
  f21= 'u''^2/(1+v^2)+ u'' v'' u- u v''';
  f22= 'u^2- v^2+ 2 u''^3- (u-v) v'' + v''^3';
  f31= 'u" cos(v) + v" sin(u) + u'' sin(2u) - v''cos(2u)';
  f32= 'cos(v") + u'' v'' sin(u)';
  f41= 'p(u^2 + v^2)/2 + q(x^2 + y^2)';
  f51= '(u''^2+v''^2)/2 - p/(u^2+v^2)^(1/3)';
  f61= '(u^2 + 2v) cos(w)^2 + sen(u v) sen( w/2)  +  (v^2 - pu) sin(2w)^2';
  f71= '2(u^2 x^3+v^2)(-1+ p sqrt(u^2+v^2))+ x - v cos(w)+u sen(w)';

procedure TMainForm.AdjustMenus;
begin
  if busy then
    Exit;
  EFO.checked:= (md = ExplicitFO);
  ESO.checked:= (md = ExplicitSO);
  IFO.checked:= (md = ImplicitFO);
  ISO.checked:= (md = ImplicitSO);
  Ham.checked:= (md = Hamilton);
  Lag.checked:= (md = Lagrange);
  ALFO.checked:= (md = PArcLenFO);
  ALSO.checked:= (md = PArcLenSO);
  Save1.Enabled:= alreadyRead;
  SaveAs1.Enabled:= alreadyRead;
  DeleteSolCurves.Enabled:= alreadyRead;
end;

{procedure TMainForm.AddImplicitCurve(Sender: TObject);
begin
  if not Assigned(ImplicitCurve) then
    ImplicitCurve:= TImplicitCurve.Create(self);
  ImplicitCurve.Show;
end;}

procedure TMainForm.AddParametricCurve(Sender: TObject);
begin
  if not Assigned(ParametricCurve) then
    ParametricCurve:= TParametricCurve.Create(self);
  ParametricCurve.Show;
end;

procedure TMainForm.AdjustInputPanel;
const
  std1= '   u Min       u Max         v Min       v Max                ';
begin
  wMaxEdit.text:= ' ';
  wMaxEdit.enabled:= md in[ImplicitFO, ImplicitSO, PArcLenFO, PArcLenSO];
  case md of
    ImplicitFO: uMinwMaxLabel.caption:=   '   u Min       u Max         v Min       v Max        |vel| Max';
    ImplicitSO: uMinwMaxLabel.caption:=   '   u Min       u Max         v Min       v Max       |accel|Max';
    PArcLenFO: uMinwMaxLabel.caption:=    '   u Min       u Max         v Min       v Max         |w| Max';
    PArcLenSO: uMinwMaxLabel.caption:=    '   u Min       u Max         v Min       v Max         |x| Max';
    else
      uMinwMaxLabel.caption:= std1;
  end;

  f2Memo.visible:= md in [ExplicitFO, ExplicitSO, ImplicitFO, ImplicitSO];
  f2label.visible:= f2Memo.visible  or (md= Hamilton);
  DirectionsBox.Visible:= md in [ImplicitFO, ImplicitSO, PArcLenFO, PArcLenSO];
  if DirectionsBox.Visible then begin
    DirectionsBox.buffer.Clear(clWhite32);
    DirectionsBox.Invalidate;
  end;

  case md of
    explicitFO: begin
      F1Label.caption:= '     u''= F(u,v) = ';
      F2Label.caption:= '     v''= G(u,v) = ';
      f1Memo.Text:= f01;
      f2Memo.Text:= f02;
    end;
    explicitSO: begin
      F1Label.caption:= '  u" = F(u,v,u'',v'') = ';
      F2Label.caption:= '  v" = G(u,v,u'',v'') = ';
      f1Memo.Text:= f11;
      f2Memo.Text:= f12;
    end;
    implicitFO: begin
      F1Label.caption:= '0= F(u,v,u'',v'')=';
      F2Label.caption:= '0= G(u,v,u'',v'')=';
      f1Memo.Text:= f21;
      f2Memo.Text:= f22;
      wMaxEdit.text:= '20';
    end;
    implicitSO: begin
      wMaxEdit.text:= '20';
      F1Label.caption:= '0=F(u,v,u'',v'',u",v")=';
      F2Label.caption:= '0=G(u,v,u'',v'',u",v")=';
      f1Memo.Text:= f31;
      f2Memo.Text:= f32;
      wMaxEdit.text:= '20';
    end;
    Hamilton: begin
      F1Label.caption:= 'H(u,v,x,y)=';
      F2Label.caption:= eMomentaNames;
      f1Memo.Text:= f41;
    end;
    Lagrange: begin
      F1Label.caption:= 'L(u,v,u'',v'')=';
      F2Label.caption:= '';
      f1Memo.Text:= f51;
    end;
    PArcLenFO: begin
      wMaxEdit.text:= 'pi';
      F1Label.caption:= '        0  = F(u,v,w) =';
      f1Memo.Text:= f61;
    end;
    PArcLenSO: begin
      wMaxEdit.text:= '20';
      F1Label.caption:= '       0 = F(u,v,w,x) =';
      f1Memo.Text:= f71;
    end;
  end;
  update;
end;

procedure TMainForm.AdjustInputPanelAfterOpening;
const
  std1= '   u Min       u Max         v Min       v Max                ';
begin
  wMaxEdit.enabled:= md in[ImplicitFO, ImplicitSO, PArcLenFO, PArcLenSO];
  case md of
    ImplicitFO: uMinwMaxLabel.caption:=   '   u Min       u Max         v Min       v Max        |vel| Max';
    ImplicitSO: uMinwMaxLabel.caption:=   '   u Min       u Max         v Min       v Max       |accel|Max';
    PArcLenFO: uMinwMaxLabel.caption:=    '   u Min       u Max         v Min       v Max         |w| Max';
    PArcLenSO: uMinwMaxLabel.caption:=    '   u Min       u Max         v Min       v Max         |x| Max';
    else
      uMinwMaxLabel.caption:= std1;
  end;

  f2Memo.visible:= md in [ExplicitFO, ExplicitSO, ImplicitFO, ImplicitSO];
  f2Label.visible:= f2Memo.visible or (md = Hamilton);
  DirectionsBox.Visible:= md in [ImplicitFO, ImplicitSO, PArcLenFO, PArcLenSO];
  if DirectionsBox.Visible then begin
    DirectionsBox.buffer.Clear(clWhite32);
    DirectionsBox.Invalidate;
  end;


  case md of
    explicitFO: begin
      F1Label.caption:= '     u''= F(u,v) = ';
      F2Label.caption:= '     v''= G(u,v) = ';
    end;
    explicitSO: begin
      F1Label.caption:= '  u" = F(u,v,u'',v'') = ';
      F2Label.caption:= '  v" = G(u,v,u'',v'') = ';
    end;
    implicitFO: begin
      F1Label.caption:=  '    0= F(u,v,u'',v'')=';
      F2Label.caption:=  '    0= G(u,v,u'',v'')=';
    end;
    implicitSO: begin
      F1Label.caption:= '0=F(u,v,u'',v'',u",v")=';
      F2Label.caption:= '0=G(u,v,u'',v'',u",v")= ';
    end;
    Hamilton: begin
      F1Label.caption:= eHamiltonian + kcr +
                           '           H(u,v,x,y)=';
      F2Label.caption:= eMomentaNames;
    end;
    Lagrange: begin
      F1Label.caption:= eLagrangian + kcr +
                           '           L(u,v,u'',v'')=';
      F2Label.caption:= '';
    end;
    PArcLenFO:  F1Label.caption:= ePArcLenFO + kcr +
                          '        0  = F(u,v,w) =';
    PArcLenSO:
      F1Label.caption:= ePArcLenSO + kcr +
                        '       0 = F(u,v,w,x) =';
  end;
  update;
end;

procedure TMainForm.ChangeFrame(translation: Boolean);
{scale the domain or translate it, and draw everything accordingly. The idea
 is to give the feeling that the curves and discriminant are translated
 without change when 'translation'= true. This means that the domain is to
 be shifted in the opposite sense. When 'translation'= False, the user changes
 the horizontal scale by moving the mouse from right to left or conversely, and
 the vertical by moving the mouse up and down.}
const
  cMul = 3;
var
  shX, shY, i: Integer;
  scX, scY: Double;
  rslt: Boolean;
begin
  if busy or (not alreadyRead) then
    Exit;
  if (curPt.X <> prevPt.X) or (curPt.Y <> prevPt.Y) then begin
    saved:= False; {because there are new data to save}
    {compute shift of mouse}
    shX:= curPt.X - prevPt.X;
    shY:= curPt.Y - prevPt.Y ;
    if translation then begin  {translate the domain}
      mi[1]:= mi[1] - shX * dp1;
      mi[2]:= mi[2] - shY * dp3;
      ma[1]:= ma[1] - shX * dp1;
      ma[2]:= ma[2] - shY * dp3;
    end
    else begin
        {compute new scale and change the domain accordingly}
      scX:= (rctCurves.right-cMul*shx)/rctCurves.right;
      scY:= (rctCurves.bottom+cMul*shy)/rctCurves.bottom;
      ma[1]:= scX*halfEdges[1] + center[1];
      mi[1]:= center[1] - scX*halfEdges[1];
      ma[2]:= scy*halfEdges[2] + center[2];
      mi[2]:= center[2] - scy*halfEdges[2];
    end;
    for i:=1 to 2 do begin
      edges[i]:= ma[i]-mi[i];
      halfEdges[i]:= edges[i]/2;
      center[i]:= (ma[i]+mi[i])/2;
    end;
    rslt:= (ma[1] > mi[1]) and (ma[2] > mi[2]);
    if not rslt then
       raise Exception.Create(eDomOrBm);

      {update edit fields}
    uMinEdit.text:= FormatFloatString(mi[1], 4 + ord(mi[1] < 0));
    vMinEdit.text:= FormatFloatString(mi[2],4 + ord(mi[2] < 0));
    uMaxEdit.text:= FormatFloatString(ma[1], 4 + ord(ma[1] < 0));
    vMaxEdit.text:= FormatFloatString(ma[2], 4 + ord(ma[2] < 0));
    update;
    try
      {update coordinate change between window and domain}
      dp1:= edges[1]/curves.width;
      dp2:= mi[1];
      dp3:= -edges[2]/curves.height;
      dp4:= ma[2];

      pd1:= curves.width/edges[1];
      pd2:= -pd1*mi[1];
      pd3:= -curves.height/edges[2];
      pd4:= -pd3*ma[2];
      ode.Initiate(mi, ma);
      PaintCurves(nil);
      curves.invalidate;
      if md = PArcLenFO then
        alreadyDrawnSurface:= False;
    except on E: Exception do begin
        Inform(E.message, True);
        Exit;
      end;
    end;
  end;
end;


procedure TMainForm.ChangeODEType(Sender: TObject);
var
  newMode: Modes;
begin
  newMode:= Modes((Sender as TMenuItem).tag);
  if (newMode = md) or busy then
    Exit;
  if not CloseCurrentODE(Sender) then {give user the chance of saving data}
    Exit;
  if md = PArcLenFO then begin
    SurfaceBox.buffer.Clear(clWhite32);
    surfaceBox.Repaint;
  end;
  md:= newMode;
  AdjustMenus;
  AdjustInputPanel;
  alreadyDrawnSurface:= False;
  stop:= False;
  alreadyComputedDiscriminant:= False;
  alreadyRead:= False;
  inputDataDirty:= True;
  saved:= True;
  tracking:= False;
  busy:= False;
  thereAreCrosses:= False;
  SurfaceBox.visible:= (md = PArcLenFO);
  Ode.Free;
  case md of
    explicitFO: Ode:= TOdeEf.Create;
    explicitSO: Ode:= TOdeEs.Create;
    ImplicitFO: Ode:= TOdeIf.Create;
    ImplicitSO: Ode:= TOdeIs.Create;
    Hamilton:   Ode:= TOdeH.Create;
    Lagrange:   Ode:= TOdeL.Create;
    PArcLenFO:  Ode:= TOdeNf.Create;
    PArcLenSO:  Ode:= TOdeNs.Create;
    else
       ;
  end;
  if Assigned(gPicker) then
    gPicker.Initiate;
  curves.Invalidate;
end;


function TMainForm.CloseCurrentODE(Sender: TObject): Boolean;
{When changing ODE type or opening a file, this routine
 asks the user whether the ODE data must be saved}
var
  res: Word;
begin
  result := False;
  if busy then
    Exit;
  if alreadyRead and (not saved) then begin
    with TYesNotCancel.Create(nil) do {Ask the user "Save It?"}
      try
        res:= ShowModal;
      finally
        release;
      end;
    if res = mrYes then begin
      OnSaveODE(sender);
      if not saved then
        Exit;
    end
    else if res = mrCancel then
      Exit;
  end;
  result := True;
end;

function TMainForm.ComputeSurface: Boolean;
begin
  result:= True;
  try
    BeginWatch;
    try
      Surface.ComputeImplicitSurface(trackingSurface);
      SurfBuffer.Draw(0,0,bm32);
      (Ode as TOdeNf).DrawCurves3D;
      alreadyDrawnSurface:= True;
      SurfaceBox.Repaint;
    except on Exception do begin
        result:= False;
        Inform(eSurfaceFailed,True);
      end;
    end;
  finally
    EndWatch;
  end;
end;

procedure TMainForm.DrawCrossesFO(var number: Integer; toErase: Boolean; u,v: Integer);
{Draw the DirectionsBox that can be initial conditions for the ODE at (u,v)}
  var
    i: Integer;
    x, dir: Vector2;
    a, b, c: Double;
    co, si: Double;
    dirs: Directions;
begin
  if (not alreadyRead) or inputDataDirty then
    Exit;
  number:= 0;
  curvesCan.Pen.Mode:= pmXoR; {thus we shall be able to erase it afterwards}
  curvesCan.Pen.Color:= clRed;
  if thereAreCrosses and (Crosses.lastCross > -1) then begin {erase previous vectors}
    for i:= 0 to Crosses.lastCross do begin
      curvesCan.moveto(crosses.basePt.x, crosses.basePt.y);
      curvesCan.lineto(crosses.cr[i].x, crosses.cr[i].y);
    end;
    curvesCan.Ellipse(crosses.basePt.x-3,crosses.basePt.y-3,
        crosses.basePt.x+3, crosses.basePt.y+3);
  end;
  thereAreCrosses:= False;
  crosses.lastCross:= -1;
  if not toErase and PtInRect(rctCurves,Point(u,v)) then begin {draw new vectors}
    x:= FromDrawingToParams(u,v);
    SetLength(dirs.FOSO, 5);
    dirs.last:= -1;
    ode.GetInitDirections(dirs, dir, x);  {find DirectionsBox at 'x'}
    thereAreCrosses:= ( dirs.last > -1);
    crosses.lastCross:= dirs.last;
    if High(crosses.cr) < crosses.lastCross then
      SetLength(crosses.cr, crosses.lastCross +3);
    if thereAreCrosses then begin
      {set proportion parameters}
      a:= rctCurves.right/12;
      b:= a*edges[1]/edges[2];

      crosses.basePt.x:= u;
      crosses.basePt.y:= v;
      if not (md = PArcLenFO) then
        for i:= 0 to crosses.lastCross do begin
          si:= dirs.FOSO[i,2];
          co:= dirs.FOSO[i,1];

          {limit the size of the vector}
          c:= hypot(si,co);
          if c > span2D/2 then begin
            c:= span2D/(2*c);
            co:= co*c;
            si:= si*c;
          end;

          {compute the tip of the vector}
          crosses.cr[i].x:= round(u + a*co);
          crosses.cr[i].y:= round(v - b*si);

          {draw the vector}
          curvesCan.moveto(u, v);
          curvesCan.lineto(crosses.cr[i].x, crosses.cr[i].y);
        end
      else if md = PArcLenFO then begin
        for i:= 0 to crosses.lastCross do begin
          {all crosses are of same length}
          sincos(dirs.FOSO[i,1],si,co);
          crosses.cr[i].x:= round(u + a*co);
          crosses.cr[i].y:= round(v - b*si);
          curvesCan.moveto(u, v);
          curvesCan.lineto(crosses.cr[i].x, crosses.cr[i].y);
        end
      end;
      curvesCan.Ellipse(u-3,v-3,u+3, v+3);
    end;
  end;
  number:= dirs.last+1;
  curvesCan.Pen.Mode:= pmCopy;
  dirs.FOSO:= nil;
end;

procedure TMainForm.DropLastCurve2D(Sender: TObject);
begin
  if not busy then begin
    Ode.DropLastCurve;
    Invalidate;
  end;
end;

procedure TMainForm.DropLastCurve3D(Sender: TObject);
begin
  if (not busy) and (md = PArcLenFO) then begin
    surfBuffer.Draw(0,0,bm32);
    (Ode as TOdeNf).DropLastCurve3D;
    (Ode as TOdeNf).DrawCurves3D;
    SurfaceBox.invalidate;
    Invalidate;
  end;
end;

procedure TMainForm.EndWatch;
{End the possibility of stopping a calculation and declare the
 SurfaceBox renderer as idle, so that interaction with the program
 is permitted from now on}
begin
  EnterButton.Font.Color:= clNavy;
  EnterButton.caption := '';
  EnterButton.Repaint;
  busy:= False;
  stop := False;
end;


procedure TMainForm.FillDirectionsBox(u,v,up,vp: Double);
const
  cl1= Cardinal($00FF0000);
  cl2= Cardinal($000000FF);
var
  i, j, w, h: Integer;
  stepx, stepy, x, y: Double;
  line: PColor32Array;
  lines: array[1..10] of PColor32Array;
  bm: TBitmap32;
  cl: TColor32;
begin
  bm:= DirectionsBox.Buffer;
  w:= DirectionsBox.width;
  h:= DirectionsBox.height;
  stepx:= 2*ma[3]/w;
  stepy:= 2*ma[3]/h;
  bm.BeginUpdate;
  if md = ImplicitFO then begin
    y:= ma[3]+stepy;
    for j:= 0 to h-1 do begin
      y-= stepy;
      line:= bm.Scanline[j];
      x:= -ma[3]-stepx;
      for i:= 0 to w-1 do begin
        x+= stepx;
        cl:= Cardinal($FF00FF00 + ord( theF1.ff([u,v,x,y]) >= 0 )*cl1 + ord(theF2.ff([u,v,x,y]) >= 0)*cl2);
        line^[i]:=cl;
      end;
    end;
    bm.PenColor:= $FFBBBBBB;
    bm.DrawMode:= dmOpaque;
    bm.moveto(0, h div 2);
    bm.linetoS(w-1, h div 2);
    bm.moveto(w div 2, 0);
    bm.linetoS(w div 2, h-1);
  end
  else if md = ImplicitSO then begin
    y:= ma[3]+stepy;
    for j:= 0 to h-1 do begin
      y-= stepy;
      line:= bm.Scanline[j];
      x:= -ma[3]-stepx;
      for i:= 0 to w-1 do begin
        x+= stepx;
        cl:= Cardinal($FF002200 + ord(theF1.ff([u,v,up,vp,x,y]) >= 0)*cl1 +
             ord(theF2.ff([u,v,up,vp,x,y]) >= 0)*cl2);
        line^[i]:= cl;
      end;
    end;
    bm.PenColor:= $FFBBBBBB;
    bm.DrawMode:= dmOpaque;
    bm.moveto(0, h div 2);
    bm.linetoS(w-1, h div 2);
    bm.moveto(w div 2, 0);
    bm.linetoS(w div 2, h-1);
  end
  else if md = PArcLenFO then begin
    bm.Clear(clWhite);
    x:= -ma[3]-stepx;
    for j:= 1 to 10 do
      lines[j]:= bm.Scanline[h div 2 - 5 + j];
    for i:= 0 to w-1 do begin
      x+= stepx;
      cl:= Cardinal($FF002200  + cl2 +ord(theF1.ff([u,v,x]) >= 0)*cl1);
      for j:= 1 to 10 do
        lines[j]^[i]:= cl;
     end;
  end
  else if md = PArcLenSO then begin
    bm.Clear(clWhite);
    x:= -ma[3]-stepx;
    for j:= 1 to 10 do
      lines[j]:= bm.Scanline[h div 2 - 5 + j];
    for i:= 0 to w-1 do begin
      x+= stepx;
      cl:= Cardinal($FF002200 + cl2 +ord(theF1.ff([u,v,up,x]) >= 0)*cl1);
      for j:= 1 to 10 do
        lines[j]^[i]:= cl;
    end;
  end;
  bm.EndUpdate;
  DirectionsBox.Invalidate;
end;


function TMainForm.FromParamsToDrawing(u,v: Double): TPoint;
begin
  result.x:= round(pd1*u + pd2);
  result.y:= round(pd3*v + pd4);
end;

function TMainForm.FromDrawingToParams(x,y: Integer): Vector2;
begin
  result[1]:= dp1*x + dp2;
  result[2]:= dp3*y + dp4;
end;


procedure TMainForm.Inform(m: AnsiString; ring: Boolean);
begin
  if (m <> '') and ring then
    Beep;
  InfoBar.Panels[0].text := m;
end;

procedure TMainForm.KeyUpInCurves(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if (Key in [vk_Shift, vk_Control]) and trackingFrame then begin
    trackingFrame:= False;
    if md in [ImplicitFO, PArcLenFO] then
      OnEnterButton(EnterButton);
  end
end;



procedure TMainForm.BeginWatch;
{Begin the possibility of stopping a calculation and declare the
 SurfaceBox renderer as busy, so that interaction with the program
 is not permitted from now on}
begin
  stop := False;
  busy:= True;
  EnterButton.Font.Color := clRed;
  EnterButton.caption := eStop;
  EnterButton.Invalidate;
end;


procedure TMainForm.MouseDownInCurves(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  dummy: Vector2;
  num: Integer;
begin
  if busy then begin
    stop:= True;  {thus, if along a calculation, the program calls
      'Application.ProcessMessages', the variable 'InputForm.stop' will have
      the value 'True' and then it must finish the calculation}
    Exit;
  end;
  if not AlreadyRead then
    Exit;
  tracking:= (mbLeft = Button) and not (md in [ExplicitFO, ImplicitFO, PArcLenFO]);
  curPt:= Point(x,y);
  prevPt:= curPt;
  firstPt:= curPt;
  if tracking then begin
    {prepare pen for drawing a sticking line}
    curvesCan.Brush.Color:= clWhite;
    curvesCan.Pen.Color := clPurple;
    curvesCan.Pen.Mode := pmXor;
    curvesCan.moveto(x,y);
    curvesCan.lineto(x,y);
    curvesCan.Pen.Mode := pmCopy;
  end
  else begin
    DrawCrossesFO(num, true, prevPt.x, prevPt.y);
    curPos2D:= FromDrawingToParams(x,y);
    Ode.ComputeCurve(curPos2D, dummy);
    Curves.Invalidate;
    saved:= False; {because there are new curves to save}
  end;
end;

procedure TMainForm.MouseDownSurface(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  pos: Vector3;
begin
  if Shift= [ssLeft] then begin
    if Surface.PickedPt(x,y,pos) then begin
      (ODe as TOdeNf).ComputeCurveByPoint3D(pos);
      (ODe as TOdeNf).DrawCurves3D;
      Curves.Invalidate;
      SurfaceBox.invalidate;
    end;
    trackingSurface:= False;
  end
  else if ssRight in Shift then begin
     trackingSurface:= True;
     curPtSurface:= Point(x,y);
  end;
end;

procedure TMainForm.MouseLeaveCurves(Sender: TObject);
begin
  if trackingFrame then begin
    trackingFrame:= False;
    if md in [ImplicitFO, PArcLenFO] then begin
      if md = ImplicitFO then
        (Ode as TOdeIf).ComputeVGCurve
      else begin
        (Ode as TOdeNf).ComputeVGCurve;
        surface.SetUp(bm32, mi, ma, 1);
        ComputeSurface;
      end;
      Ode.DrawCurves2D(1);
    end;
    Update;
  end;
end;


procedure TMainForm.MouseMoveInCurves(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
var
   firstParam, curParam, dir, v2, a2: Vector2;
   strCoord: AnsiString;
   drs: Directions;
   ders: array of Double;
   num, i: Integer;
   goodPt: Boolean;
begin
  if busy  or (not alreadyRead) then
    Exit;
  prevPt:= curPt;
  curPt:= Point(x, y);
  if (Shift = []) and trackingFrame then begin
    trackingFrame:= False;
    if md in [ImplicitFO, PArcLenFO] then
      OnEnterButton(EnterButton);
  end
  else if (Shift = [ssCtrl]) or  (Shift = [ssShift]) then begin
    {translate domain intervals where the ODE is defined}
    if not trackingFrame then begin
       trackingFrame:= True;
       Exit;
    end
    else
      ChangeFrame(Shift = [ssCtrl]);
  end
  else
    if not (md in [ExplicitFO, ImplicitFO, PArcLenFO]) then begin
      if tracking and (Shift = [ssLeft]) then begin
        {prepare pen for drawing sticking line}
        curvesCan.Pen.Color := clPurple;
        curvesCan.Pen.Mode := pmXor;

        {erase previous sticking line}
        curvesCan.moveto(firstPt.X, firstPt.Y);
        curvesCan.lineto(prevPt.X, prevPt.Y);

        {draw the new sticking line}
        curvesCan.moveto(firstPt.X, firstPt.Y);
        curvesCan.lineto(curPt.X, curPt.Y);
        curvesCan.Pen.Mode := pmCopy;

        {compute the vector represented by the sticking line}
        firstParam:= FromDrawingToParams(firstPt.x, firstPt.y);
        curParam:= FromDrawingToParams(curPt.x, curPt.y);
        for i:= 1 to 2 do
          dir[i]:= curParam[i]- firstParam[i];


        SetLength(drs.FOSO,2);
        drs.last:= -1;

        {find appropriate accelerations, i.e. that may serve as initial conditions of a
         second order ODE}
        if (md in [Lagrange, Hamilton]) then begin
          SetLength(ders, 4);
          Ode.Derivative(ders, [firstParam[1], firstParam[2], dir[1], dir[2]]);
        end
        else
          Ode.GetInitDirections(drs, dir, firstParam);
        if md in [ImplicitSO, PArcLenSO] then
          FillDirectionsBox(firstParam[1], firstParam[2], dir[1], dir[2]);

        {write to the status bar info about the point and the vector represented
         by the sticking line}
        strCoord := 'u= '+FormatFloatString(firstParam[1], 7) + ' ,  ';
        strCoord := strCoord + 'v= ' + FormatFloatString(firstParam[2],7)+' ; ';
        if md = PArcLenSO then
          strCoord := strCoord + 'w= '+ FormatFloatString(dir[1],7)+' ; '
        else  begin
          if md = Hamilton then begin
            strCoord := strCoord + 'x= '+ FormatFloatString(dir[1],7)+' , ';
            strCoord := strCoord + 'y= '+ FormatFloatString(dir[2],7)+' ; ';
            strCoord := strCoord + 'u''= '+ FormatFloatString(ders[0],7)+' , ';
            strCoord := strCoord + 'v''= '+ FormatFloatString(ders[1],7)+' ; ';
            strCoord := strCoord + 'u"= '+ FormatFloatString(ders[2],7)+' , ';
            strCoord := strCoord + 'v"= '+ FormatFloatString(ders[3],7);
          end
          else if md = Lagrange then begin
            strCoord := strCoord + 'u''= '+ FormatFloatString(ders[0],7)+' , ';
            strCoord := strCoord + 'v''= '+ FormatFloatString(ders[1],7)+' ; ';
            strCoord := strCoord + 'u"= '+ FormatFloatString(ders[2],7)+' , ';
            strCoord := strCoord + 'v"= '+ FormatFloatString(ders[3],7);
          end
          else begin
            strCoord := strCoord + 'u''= '+ FormatFloatString(dir[1],7)+' , ';
            strCoord := strCoord + 'v''= '+ FormatFloatString(dir[2],7)+' ; ';
          end;
          if md = explicitSO then
            try
              a2[1]:= theF1.ff([curParam[1], curParam[2], dir[1],dir[2]]);
              a2[2]:= theF2.ff([curParam[1], curParam[2], dir[1],dir[2]]);
              strCoord := strCoord + 'u"= '+ FormatFloatString(a2[1],7)+' , ';
              strCoord := strCoord + 'v"= '+ FormatFloatString(a2[2],7);
            except on exception do
              Inform(eBadPointEncountered, false);
            end
        end;

        {write to the status bar the number of appropriate accelerations}
        if md in [ImplicitSO, PArcLenSO] then
          strCoord:= strCoord + IntToStr(drs.last+1) + eAccelerations;
        InfoBar.Panels[0].text:= strCoord;
        drs.FOSO:= nil;
      end
      else if PtInRect(rctCurves, Point(x,y)) then begin
        v2:= FromDrawingToParams(x,y);
        strCoord := 'u= '+FormatFloatString(v2[1], 7) + ' ,  ';
        strCoord := strCoord + 'v= ' + FormatFloatString(v2[2],7);
        Inform(strCoord, false);
      end;
    end
    else if PtInRect(rctCurves, Point(x,y)) then begin  {mode is of first order}
      DrawCrossesFO(num, false, curPt.X, curPt.y);

      {give info about the point where is the mouse}
      v2:= FromDrawingToParams(x,y);
      strCoord := 'u= '+FormatFloatString(v2[1], 7) + ' ,  ';
      strCoord := strCoord + 'v= ' + FormatFloatString(v2[2],7);

      if md = ImplicitFO then begin
        {add info about the number of appropriate velocities that may serve as
        initial conditions}
        strCoord := strCoord + '  ; '+ IntToStr(num) + eVelocities;
        FillDirectionsBox(v2[1], v2[2], dir[1], dir[2]); {'dir[1], dir[2]' are not used}
      end
      else if md = PArcLenFO then begin
        strCoord := strCoord + '  ; '+ IntToStr(num) + eVelocities;
        FillDirectionsBox(v2[1], v2[2], dir[1], dir[2]);
      end;
      try
        if md = explicitFO then begin
          dir[1]:= theF1.ff([v2[1], v2[2]]);
          dir[2]:= theF2.ff([v2[1], v2[2]]);
          strCoord := strCoord + ' ; u''= '+ FormatFloatString(dir[1],7)+' , ';
          strCoord := strCoord + 'v''= '+ FormatFloatString(dir[2],7);
        end;
        InfoBar.Panels[0].text:= strCoord;
        except on exception do
          Inform(eBadPointEncountered, False);
      end;
  end;
  drs.FOSO:= nil;
  ders:= nil;
end;

procedure TMainForm.MouseMoveSurface(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
  if trackingSurface and (ssRight in Shift) then begin
    prevPtSurface:= curPtSurface;
    curPtSurface:= Point(x,y);
    surface.MoveSurface(prevPtSurface, curPtSurface, Shift);
    alreadyDrawnSurface:= True;
    surfaceBox.Buffer.BeginUpdate;
    surfaceBox.Buffer.Draw(0,0,bm32);
    surfaceBox.Buffer.EndUpdate;
    //(Ode as TOdeNf).DrawCurves3D;
    surfaceBox.Repaint;
  end;
end;

procedure TMainForm.MouseUpInCurves(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
{user has left pressing the mouse button}
var
  param, dir:  Vector2;
begin
  if busy or (not alreadyRead)  then
    Exit;
  if (not (md in [ExplicitFO, ImplicitFO, PArcLenFO]))  and tracking  then begin
    {erase sticking line}
    curvesCan.Pen.Color := clPurple;
    curvesCan.Pen.Mode := pmXor;
    curvesCan.moveto(firstPt.x, firstPt.y);
    curvesCan.lineto(curPt.X, curPt.Y);
    curvesCan.Pen.Mode := pmCopy;
    curPt.x:= X;
    curPt.y:= Y;
    {compute the vector represented by the last sticking line}
    param:= FromDrawingToParams(firstPt.x, firstPt.y);
    dir:= FromDrawingToParams(curPt.x, curPt.y);
    dir[1]:= dir[1]- param[1]; dir[2]:= dir[2]- param[2];
    ode.ComputeCurve(param, dir);
    saved:= False; {because there are new curves}
    curves.Repaint;
  end;
  tracking:= False;
  trackingFrame:= False;
end;

procedure TMainForm.MouseUpSurface(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  trackingSurface:= False;
  surface.uGrain:= 1;
  ComputeSurface;
  (Ode as TOdeNf).DrawCurves3D;
end;


procedure TMainForm.OnChangeInput(Sender: TObject);
begin
  if busy or (not alreadyRead) then
    Exit;
  if Sender is TCustomEdit then begin
      enterButton.Color:= clInfoBk;
      enterButton.caption:= eEnter;
      EnterButton.Repaint;
  end;
end;

procedure TMainForm.OnCloseQueryClick(Sender: TObject; var CanClose: boolean);
  {User wants to close the application}
var
  res: Word;
begin
  if busy then
    Exit;
  if alreadyRead and not saved then begin
    with TYesNotCancel.Create(nil) do {Ask the user "Save It?"}
      try
        res:= ShowModal;
      finally
        Release;
      end;
    if res = mrNo then
      CanClose := True
    else if res = mrYes then begin
      OnSaveODE(sender);
      canClose := saved;
    end
    else
      canClose := False;
  end
  else
    CanClose := True;
  end;



procedure TMainForm.OnCreateForm(Sender: TObject);
begin
  InfoBar.Panels[0].text:= '';
  rctSurface:= Rect(0, 0, SurfaceBox.width, SurfaceBox.Height);
  rctDirections:= Rect(0, 0, DirectionsBox.width, DirectionsBox.Height);
  rctCurves:= Rect(0, 0, curves.width, curves.Height);
  bmBGRA:= TBGRABitmap.Create(curves.width, curves.Height, BGRAWhite);
  curvesCan:= Curves.Canvas;
  surfBuffer:=  SurfaceBox.buffer;
  bmDir:= DirectionsBox.buffer;
  md:= none;
  bm32:= TBitmap32.Create;
  bm32.Width:= SurfaceBox.width;
  bm32.Height:= SurfaceBox.Height;
  gPicker:= TGPicker.Create(Application);
  ChangeODEType(EFO);
  SetLength(crosses.cr,10);
  crosses.lastCross:= -1;
  pathAndName:= '';
  caption:= 'ODEinR2';

  File1.caption:= eFile;
  EFO.caption:= eExplicitFO;
  ESO.caption:= eExplicitSO;
  IFO.caption:= eImplicitFO;
  ISO.caption:= eImplicitSO;
  Ham.caption:= eHamilton;
  Lag.caption:= eLagrange;
  ALFO.caption:= ePArcLenFO;
  ALSO.caption:= ePArcLenSO;
  Open1.caption:= eOpen;
  Save1.caption:= eSave;
  SaveAs1.Caption:= eSaveAs;
  SaveImages1.caption:= eSaveImages;
  Quit1.caption:= eExit;
  Edit1.caption:= eEdit;
  Cut1.caption:= eCut;
  Copy1.caption:= eCopy;
  Paste1.caption:= ePaste;
  DropLastCurve2DItem.caption:= eDropLastCurve2D;
  DropLastCurve3DItem.caption:= eDropLastCurve3D;
  DeleteSolCurves.caption:= eDropSolutionCurves;
  DeleteDiscriminant1.caption:= eDeleteDiscriminant;
  DeleteAuxCurves.caption:= eDeleteAuxiliaryCurves;
  View1.caption:= eView;
  Palettes1.caption:= eShowColorControls;
  AddParamCurve.caption:= eParametricCurve;
  EnterButton.caption:= eEnter;
end;

procedure TMainForm.OnDeleteAuxCurves(Sender: TObject);
begin
  ode.InitCurves(False, False, True);
  curves.Repaint;
end;

procedure TMainForm.OnDeleteDiscriminant(Sender: TObject);
begin
  ode.InitCurves(False, True, False);
  curves.Repaint;
end;

procedure TMainForm.OnDestroyForm(Sender: TObject);
begin
  bmBGRA.Free;
  Ode.Free;
  //gPicker.Free; {do not free gPicker because its owner is 'Application'}
  bm32.Free;
  crosses.cr:= nil;
end;

procedure TMainForm.OnEnterButton(Sender: TObject);
begin
  if (EnterButton.caption = eStop) then begin
    stop:= True;
    Exit;
  end;
  DirectionsBox.buffer.Clear(clWhite);
  DirectionsBox.invalidate;

    {read edit fields and parse functions }
  alreadyRead:= ReadForm(sender <> EnterButton);
  ode.InitCurves(True, True, True);
  inputDataDirty := not alreadyRead;  {something failed}
  if alreadyRead then begin
    curvesCan.FillRect(RctCurves);
    if md in [ImplicitFO, PArcLenFO] then begin
      if (md = PArcLenFO) then begin
        (Ode as TOdeNf).ComputeVGCurve;
        alreadyDrawnSurface:= False;
        surface.SetUp(bm32, mi, ma, 1);
        alreadyDrawnSurface:= ComputeSurface;
      end
      else if (Sender = EnterButton) then
        (Ode as TOdeIf).ComputeVGCurve;
    end;
    SurfaceBox.Visible:= md = PArcLenFO;
    MainForm.saved := False;
    AdjustMenus;
    EnterButton.Color:= clDefault;
    EnterButton.caption:= '';
  end;
  Invalidate;
end;

procedure TMainForm.OnManual(Sender: TObject);
begin
  Manual.Show;
end;

procedure TMainForm.OpenFile(Sender: TObject);
var
  str: AnsiString;
begin
  if busy then
    Exit;
  if not CloseCurrentODE(Sender) then
    Exit;
  alreadyRead:= False;
  try
    OpenDialog1.Filter:= eODEinR2Filter;
    OpenDialog1.Title:= eOpenFile;
    OpenDialog1.Options := [ofFileMustExist, ofNoChangeDir];
    OpenDialog1.Filename := '';
    if OpenDialog1.Execute then begin
      pathAndName := OpenDialog1.FileName;
      ode.Free;
      alreadyRead:= ReadFromDisk(pathAndName, md);
    end;
    if alreadyRead then begin
      AdjustMenus;
      AdjustInputPanelAfterOpening;
      alreadyDrawnSurface:= False;
      stop:= False;
      alreadyComputedDiscriminant:= (md in [ImplicitFO, PArcLenFO]);
      alreadyRead:= False;
      inputDataDirty:= True;
      tracking:= False;
      busy:= False;
      thereAreCrosses:= False;
      SurfaceBox.visible:= (md = PArcLenFO);
      alreadyRead:= ReadForm(True);
      if  not alreadyRead then
         raise Exception.Create(eErrorOnReadingFile);
      Ode.Initiate(mi, ma);
      stop:= False;
      alreadyComputedDiscriminant:= True;
      if Assigned(gPicker) then begin
        gPicker.show;
        gPIcker.Initiate;
        gPicker.DrawHSandMeters;
      end;
      If md = PArcLenFO then begin
        BeginWatch;
        surface.SetUpAfterOpeningFile(bm32, mi, ma);
        alreadyDrawnSurface:= ComputeSurface;
        EndWatch;
        (Ode as TOdeNf).DrawCurves3D;
        SurfaceBox.Repaint;
      end;
      Curves.RePaint;
      str:= ExtractFileName(pathAndName);
      caption:= ExtracTitle(str, kFilter);
      saved:= True;
    end;
    except on E: Exception do begin
        Inform(E.Message, true);
        alreadyRead:= False;
      end;
  end;
end;



procedure TMainForm.OnQuit(Sender: TObject);
{User has clicked the menu item 'Quit'}
var
  cc: Boolean;
begin
  if busy then
    Exit;
  cc := False;
  OnCloseQuery(sender, cc); {Give user chance for saving data}
  {cc is the answer to the OnCloseQuery}
  if cc then
    Application.Terminate;
end;

procedure TMainForm.OnSaveODE(Sender: TObject);
begin
  if Busy or (not alreadyRead) or saved then
    Exit;
  if (pathAndName= '') then
    OnSaveAsODE(sender)
  else begin
    WriteToDisk(pathAndName);
    saved:= True;
    Save1.Enabled:= False;
  end;
end;


procedure TMainForm.OnSaveAsODE(Sender: TObject);
var
  theName:  string[255];
begin
  if busy or (not AlreadyRead) then
    Exit;
  SaveDialog1.FileName := caption;
  SaveDialog1.DefaultExt := '.2ed';
  SaveDialog1.Filter := eODEinR2Filter;
  if pathAndName <> '' then
    SaveDialog1.InitialDir := ExtractFileDir(pathAndName);
  if SaveDialog1.Execute then begin
    theName := SaveDialog1.FileName;
    pathAndName:= theName;
    WriteToDisk(pathAndName);
    caption:= ExtractFilename(pathAndName);
    saved:= True;
  end;
end;

procedure TMainForm.OnSaveImages(Sender: TObject);
begin
  beginWatch;
  ImagesToSave.Show;
  endWatch;
  if savedImages <> '' then
    Inform(savedImages, True);
end;

procedure TMainForm.OnView(Sender: TObject);
begin
  if (Sender = Palettes1) then begin
    gPicker.show;
    gPIcker.Initiate;
  end
  else if (md = PArcLenFO) and not alreadyDrawnSurface then
    alreadyDrawnSurface:= ComputeSurface;
end;

procedure TMainForm.PaintCurves(Sender: TObject);
begin
  if alreadyRead then begin
    bmBGRA.Fill(BGRAWhite);
    ode.DrawCurves2D(1);
    curvesCan.pen.mode:= pmCopy;
    bmBGRA.Draw(curvesCan, rctCurves, True);
  end;
end;

procedure TMainForm.PrepareEnvironment(fromFile: Boolean);
var
  i: Integer;
begin
      {Set basic data  }
  for i:=1 to 3 do begin
    edges[i]:= ma[i]-mi[i];
    center[i]:= (ma[i]+mi[i])/2;
    halfEdges[i]:= edges[i]/2;
  end;
  span3D:= EuclNorm(edges);
  span2D:= hypot(ma[1]-mi[1],ma[2]-mi[2]);
  dp1:= edges[1]/rctCurves.width;
  dp2:= mi[1];
  dp3:= -edges[2]/rctCurves.height;
  dp4:= ma[2] ;

  pd1:= 1/dp1;
  pd2:= -dp2/dp1;
  pd3:= 1/dp3;
  pd4:= -dp4/dp3;
  inputDataDirty:= False;
  tracking:= False;
  thereAreCrosses:= False;
  try
     {prepare the ODE solver for receiving computing orders}
    ode.Initiate(mi, ma);
  except on E: Exception do
    begin
      Inform(E.message, True);
      Exit;
    end;
  end;
end;

function TMainForm.ReadForm(fromFile: Boolean): Boolean;
{Read edit fields, parse functions, set basic data and prepare ODE solver for
 acting on this equation}
var
  params, usedvars: AnsiString;
  valid: Boolean;
  vset: set of AnsiChar;
begin
  result:= True;
  InfoBar.Panels[0].text:= '';
  mi[3]:= 0; ma[3]:= 0;
  try
    StrToSave[1] := Trim(f1Memo.Text);

    StrToSave[2] := Trim(uMinEdit.Text);
    ReadConstantField(StrToSave[2], mi[1]);

    StrToSave[3] := Trim(uMaxEdit.Text);
    ReadConstantField(StrToSave[3], ma[1]);

    if mi[1] >= ma[1] then
      raise Exception.Create(eUMinLessThanUMax);

    StrToSave[4] := Trim(vMinEdit.Text);
    ReadConstantField(StrToSave[4], mi[2]);

    StrToSave[5] := Trim(vMaxEdit.Text);
    ReadConstantField(StrToSave[5], ma[2]);

    if mi[2] >= ma[2] then
      raise Exception.Create(eVMinLessThanVMax);

    if md in [ImplicitFO, ImplicitSO, PArcLenFO, PArcLenSO] then begin
      StrToSave[6]:= Trim(wMaxEdit.Text);
      ReadConstantField(StrToSave[6], ma[3]);
      mi[3]:= -ma[3];
    end
    else
      StrToSave[6]:= '5';

    StrToSave[7] := Trim(pEdit.Text);
    ReadConstantField(StrToSave[7], p);

    StrToSave[8] := Trim(qEdit.Text);
    ReadConstantField(StrToSave[8], q);

    if not (md in [Hamilton, Lagrange, PArcLenFO, PArcLenSO]) then
      StrToSave[9] := Trim(f2Memo.Text);

    params:= 'PQ';

    with TheParser do
    case md of
      ExplicitFO: begin
          ParseNewText(StrToSave[1],'UV', params, md, false);
          valid:= varsSet <= ['U','V'];
        end;
      ExplicitSO: begin
          ParseNewText(StrToSave[1], 'UVWX', params, md, false);
          valid:=  varsSet <= ['U','V','W','X'];
        end;
      ImplicitFO: begin
          ParseNewText(StrToSave[1], 'UVWX', params, md, false);
          valid:= (varsSet <= ['U','V','W','X']) and (('W' in varsSet) or
             ('X' in varsSet));
        end;
      ImplicitSO: begin
          ParseNewText(StrToSave[1], 'UVWXYZ', params, md, false);
          valid:= (varsSet <= ['U','V','W','X','Y','Z']) and (('Y' in varsSet) or ('Z' in varsSet));
        end;
      Hamilton: begin
          ParseNewText(StrToSave[1],'UVXY', params, md, false);
          valid:= (varsSet <= ['U','V','X','Y']) and (('U' in varsSet) or ('V' in varsset)) and
          (('X' in varsSet) or ('Y' in varsset));
        end;
      Lagrange: begin
          ParseNewText(StrToSave[1], 'UVWX', params, md, false);
          valid:= (('U' in varsSet) or ('V' in varsset)) and  (('W' in varsSet) or ('X' in varsset));
        end;
      PArcLenFO: begin
          ParseNewText(StrToSave[1], 'UVW', params, md, false); {Addition of 'r' to
             params, because we shall use the value 'r*w' instead of 'w'. See 'ParseNewText'}
          valid:= ('W' in varsSet);
        end;
      PArcLenSO: begin
          ParseNewText(StrToSave[1], 'UVWX', params, md, false);
          valid:= ('X' in varsSet);
        end;
      else  ;
    end; {case}
    if not valid then
      raise Exception.Create(eBadEquations);

    vset:= TheParser.varsSet;
    usedVars:= theParser.UsedVarsAndParams;

    theF1.Build(theParser);

   { if md = PArcLenFO then begin
      r := (4*ma[3])/(ma[1] - mi[1]+ ma[2]- mi[2]);
      ma[3]:= ma[3]/r;
      mi[3]:= -ma[3];
      theF1.UpdateParameters([p, q, r]);
    end
    else}
      theF1.UpdateParameters([p,q]);


    with TheParser do
    case md of
      ExplicitFO: begin
          ParseNewText(StrToSave[9],'UV', params, md, false);
          valid:=  varsSet <= ['U','V'];
        end;
      ExplicitSO: begin
          ParseNewText(StrToSave[9], 'UVWX', params, md, false);
          valid:=  varsSet <= ['U','V','W','X'];
        end;
      ImplicitFO: begin
          ParseNewText(StrToSave[9], 'UVWX', params, md, false);
          valid:= (varsSet <= ['U','V','W','X']) and (('W' in varsSet) or
             ('X' in varsSet));
        end;
      ImplicitSO: begin
          ParseNewText(StrToSave[9], 'UVWXYZ', params, md, false);
          valid:= (varsSet <= ['U','V','W','X','Y','Z']) and (('Y' in varsSet) or ('Z' in varsSet));
        end;
      else  ;
    end; {case}
    if not valid then
      raise Exception.Create(eBadEquations);
    if md = ImplicitSO then begin
       vset:= vset + TheParser.varsSet;
       valid:= ['Y','Z'] <= vset;
       if not valid then
          raise Exception.Create(eBadEquations);
    end;

    if md in [explicitFO, ExplicitSO, ImplicitFO, ImplicitSO] then begin
      theF2.Build(theParser);
      theF2.UpdateParameters([p,q]);
      usedvars:= usedvars + theParser.UsedVarsAndParams;
    end;

    usedParams:= [];

    {pLabel.visible:=  pos('P',usedvars)> 0;
    pEdit.visible:= pLabel.visible;

    qLabel.visible:=  pos('Q',usedvars)> 0;
    qEdit.visible:= qLabel.visible;}

    PrepareEnvironment(fromFile);
  except
    on E: Exception do
      begin
        result := False;
        alreadyRead:= False;
        Inform(E.message,True);
      end;
  end;
end;

procedure TMainForm.UpdateParameters(pp,qq: Double);
{called when user has moved some of the ticks in the main window
 border that denote the parameters}
begin
  {update saved parameter values}
  StrToSave[7]:= pEdit.Text;
  StrToSave[8]:= qEdit.Text;

  {update parameters in functions}
  if md = PArcLenFO then
    TheF1.UpdateParameters([p,q])
  else
    TheF1.UpdateParameters([p,q]);
  if md in [ExplicitFO, ExplicitSO, ImplicitFO, ImplicitSO] then
    theF2.UpdateParameters([p,q]);
end;



procedure TMainForm.RefreshCurves(Sender: TObject);
begin
  ode.InitCurves(True, False, False);
  curves.Repaint;
end;





end.

