(*

MIT License

Copyright (c) [2005 - 2018] [Angel Montesinos-Amilibia]

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

*)


unit FOde;

{.$MODE Delphi}
{$mode objfpc}{$H+}

interface
uses
  Graphics, classes, SysUtils, GR32, BGRABitmap, BGRABitmapTypes,
  FRunKut, Vectors, FGetPointsInR2, FImplicitCurveDescent, FGPicker;
const
  {cGVColor= $00BBBBBB;} {gray color for the discriminant}
  cGVColor32= $FFAAAAAA;

  {masks for drawing curves in case TODEIf}
  kcv3DR2= byte(1);
  kcv3DS= byte(2);
  kcvgR2= byte(4);
  kcvgS= byte(8);

type
  cv2D= record
    color: TBGRAPixel;
    last: Integer;
    c: array of Vector2;
  end;

  cv2DArray= record
    last: Integer;
    ac: array of cv2D;
  end;

  cv3D= record
    color: TColor32;
    last: Integer;
    c: array of Vector3;
  end;

  cv3DArray= record
    last: Integer;
    ac: array of cv3D;
  end;


  TOde = class
    cv2, cvAux: cv2DArray;
    BGRAbm: TBGRABitmap;
    can: TCanvas;
    mi, ma, edges: Vector3;
    sense: sgn;
    procedure ComputeCurve(par, dir: Vector2); virtual; abstract;
    constructor Create; virtual;
    function Derivative(var deriv: array of double; here:
           array of double): Boolean; virtual; abstract;
    destructor Destroy; override;
    procedure DrawCurves2D(w: Integer); virtual;
    procedure DrawOneCurve2D(crv: cv2D; w: Integer);
    procedure DrawCurves2DOnBGRAbm(bm: TBGRABitmap; r: TRect); virtual; overload;
    function DropLastCurve: Boolean;
    procedure FromParamsToDraw(x: array of Double; var p: TPointF); overload;
    procedure FromParamsToDraw(x: array of Double; var p: TPoint); overload;
    procedure GetInitDirections(var dirs: directions; var dir: Vector2; par: Vector2); virtual; abstract;
    procedure InitCurves(solutions, vgCurve, auxCurves: Boolean); virtual;
    procedure Initiate(mi0, ma0: Vector3); virtual;
    function NumberOfPieces( const lastX, nextX: array of double): Integer;
    procedure RecordNode(x: array of double; key: Byte);
    procedure RecordNodeAux(x: array of double; key: Byte);
  private
    pd1, pd2, pd3, pd4: Double;
    span, rad, radTotal, aux: Double;
    auxInt: Integer;
    xx, vv: Vector2;
    function ComputeMatrix: Boolean;
  protected
    procedure AdjustToSurf(var x: array of double); virtual; abstract;
    function CheckCameBack(firstX, prevX, curX: array of double): Boolean; virtual; abstract;
    function CheckRange(nextX: array of double): Boolean; virtual; abstract;
    function Shift(x1,x2: array of double): Double; virtual; abstract;

  end;


  TOdeEf = class(TOde)  {for Explicit first order ODE in R²}
    const
      klastPt= 23;
    var
      arrowScale: Double;
      tangs: array[0..klastPt, 0..klastPt] of Vector2;
    procedure ComputeCurve(par, dir: Vector2); override;
    constructor Create; override;
    function Derivative(var deriv: array of double; here:
           array of double): Boolean; override;
    procedure GetInitDirections(var dirs: directions; var dir: Vector2; par: Vector2); override;
  protected
    procedure AdjustToSurf(var x: array of double); override;
    function CheckCameBack(firstX, prevX, curX: array of double): Boolean; override;
    function CheckRange(nextX: array of double): Boolean; override;
    function Shift(x1,x2: array of double): Double; override;
  end;

  TOdeEs = class(TOde)   {for Explicit second order ODE in R²}
    procedure ComputeCurve(par, dir: Vector2); override;
    function Derivative(var deriv: array of double; here:
           array of double): Boolean; override;
    procedure GetInitDirections(var dirs: directions; var dir: Vector2; par: Vector2);override;
  protected
    procedure AdjustToSurf(var x: array of double); override;
    function CheckCameBack(firstX, prevX, curX: array of double): Boolean; override;
    function CheckRange(nextX: array of double): Boolean; override;
    function Shift(x1,x2: array of double): Double; override;
  end;

  TOdeIf = class(TOde)  {For Implicit ODE of first order in R²}
    cvg: cv2DArray;
    ecVG: RealProcList;
    ecVGgrad: RealProcGradList;
    constructor Create; override;
    procedure ComputeCurve(par, dir: Vector2); override;
    procedure ComputeVGCurve;
    function dd1(p, v: Vector2): Double;
    function dd2(p, v: Vector2): Double;
    function Derivative(var deriv: array of double; here:
           array of double): Boolean; override;
    destructor Destroy; override;
    procedure Det(var f: Double; x: array of double);
    procedure DetGrad(var gr: array of double; x: array of double);
    procedure DrawCurves2D(w: Integer); override;
    procedure DrawCurves2DOnBGRAbm(bm: TBGRABitmap; r: TRect); override;
    procedure Eq1(var f: Double; x: Vector2);
    procedure Eq2(var f: Double; x:  Vector2);
    procedure GetInitDirections(var dirs: directions; var dir: Vector2; par: Vector2); override;
    procedure Grad1(var f: array of Double; x: array of double);
    procedure Grad2(var f: array of Double; x: array of double);
    procedure InitCurves(solutions, vgCurve, auxCurves: Boolean); override;
    procedure Initiate(mi0, ma0: Vector3);  override;
    procedure RecordVGNode(x: array of double; key: Byte);
    procedure vg1(var f: Double; x: array of double);
    procedure vg2(var f: Double; x: array of double);
  protected
    procedure AdjustToSurf(var x: array of double); override;
    function CheckCameBack(firstX, prevX, curX: array of double): Boolean; override;
    function CheckRange(nextX: array of double): Boolean; override;
    function CheckRange4(nextX: array of double): Boolean;
    function Shift(x1,x2: array of double): Double; override;
  end;

  TOdeIs = class(TOde) {For Implicit ODE of second order in R²}
    procedure ComputeCurve(par, dir: Vector2); override;
    function dd1(p, v: Vector2): Double;
    function dd2(p, v: Vector2): Double;
    procedure Eq1(var f: Double; x:  Vector2);
    procedure Eq2(var f: Double; x:  Vector2);
    procedure GetInitDirections(var dirs: directions; var dir: Vector2; par: Vector2); override;
    procedure Gradients(var gr1, gr2: array of double;
       const x: array of double);
  protected
    procedure AdjustToSurf(var x: array of double); override;
    function CheckCameBack(firstX, prevX, curX: array of double): Boolean; override;
    function CheckRange(nextX: array of double): Boolean; override;
    function Derivative(var deriv: array of double; here:
           array of double): Boolean; override;
    function Shift(x1,x2: array of double): Double; override;
  end;

  TOdeL = class(TOde)   {For Lagrange equations  in R²}
    procedure ComputeCurve(par, dir: Vector2); override;
    function Derivative(var deriv: array of double; here:
           array of double): Boolean; override;
    procedure DerivativesLagrangian(var fu, fv, fux, fuy, fvx, fvy, fxx, fxy, fyy:
           Double; const x: array of double);
    procedure GetInitDirections(var dirs: directions; var dir: Vector2; par: Vector2);override;
  protected
    procedure AdjustToSurf(var x: array of double); override;
    function CheckCameBack(firstX, prevX, curX: array of double): Boolean; override;
    function CheckRange(nextX: array of double): Boolean; override;
    function Shift(x1,x2: array of double): Double; override;
  end;

  TOdeH = class(TOde)   {For Hamilton equations  in R²}
    procedure ComputeCurve(par, dir: Vector2); override;
    function Derivative(var deriv: array of double; here:
           array of double): Boolean; override;
    procedure GetInitDirections(var dirs: directions; var dir: Vector2; par: Vector2);override;
  protected
    procedure AdjustToSurf(var x: array of double); override;
    function CheckCameBack(firstX, prevX, curX: array of double): Boolean; override;
    function CheckRange(nextX: array of double): Boolean; override;
    function Shift(x1,x2: array of double): Double; override;
  end;

  TOdeNf = class(TOde) {For first order ODE on normalized curves in R²}
    cv3, cvg: cv3DArray;
    surfBuffer: TBitmap32;
    ecVG: RealProcList;
    ecVGgrad: RealProcGradList;
    curvesKey: Byte; {kcv3DR2= 1;  kcv3DS= 2; kcvgR2= 4 ; kcvgS= 8}
    function SF(const x: Vector3): Double;
    function SFgrad(const x: vector3): Vector3;
    procedure ComputeCurve(par, dir: Vector2); override;
    procedure ComputeCurveByPoint3D(x: Vector3);
    procedure ComputeVGCurve;
    constructor Create; override;
    function Derivative(var deriv: array of double; here:
           array of double): Boolean; override;
    function DerivativeFrom3D(var deriv: array of double; here:
                     array of double): Boolean;
    destructor Destroy; override;
    function Direction(var vel: array of double;
           x: array of double): Boolean;
    function DirectionFrom3D(var vel: array of double;
       x: array of double): Boolean;
    procedure DrawCurves2D(w: Integer); override;
    procedure DrawCurves2DOnBGRAbm(bm: TBGRABitmap; r: TRect); override;
    procedure DrawCurves3D;
    function DropLastCurve3D: Boolean;
    procedure GetInitDirections(var dirs: directions; var dir: Vector2; par: Vector2); override;
    procedure Gradient(var f: array of Double; x: array of double);
    procedure InitCurves(solutions, vgCurve, auxCurves: Boolean); override;
    procedure Initiate(mi0, ma0: Vector3); override;
    procedure RecordCV3Node(x: array of double; key: Byte);
    procedure RecordVGNode(x: array of double; key: Byte);
    procedure VertGrad(var f: Double; x: array of double);
    procedure vg1(var f: Double; x: array of double);
  protected
    procedure AdjustToSurf(var x: array of double); override;
    function CheckCameBack(firstX, prevX, curX: array of double): Boolean; override;
    function CheckRange(nextX: array of double): Boolean; override;
    function CheckRange3D(nextX: array of double): Boolean;
    function Shift(x1,x2: array of double): Double; override;
  end;

  TOdeNs = class(TOde)  {For second order ODE on normalized curves in R²}
    procedure ComputeCurve(par, dir: Vector2); override;
    function Derivative(var deriv: array of double; here:
           array of double): Boolean; override;
    function Direction(var vel: array of double;
       x: array of double): Boolean;
    procedure GetInitDirections(var dirs: directions; var dir: Vector2; par: Vector2); override;
    procedure Gradient(var f: array of Double; x: array of double);
  protected
    procedure AdjustToSurf(var x: array of double); override;
    function CheckCameBack(firstX, prevX, curX: array of double): Boolean; override;
    function CheckRange(nextX: array of double): Boolean; override;
    function Shift(x1,x2: array of double): Double; override;
  end;

var
  Ode: TOde;

implementation
  uses
    LCLIntf, LCLType, Math, Types, BGRACanvas, BGRAPen,
    FEvaluatorBytes_64_Linux, FSurface,
    FMainForm, ResStrings;
  const
    cTolFO= 1.0e-9;
    cExTolFO = 1.0e-10;
    cExTolSO = 1.0e-10;
    cTwoPi = 2*pi;
    cHalfPi= pi/2;
    kTolFactor = 0.999;

function fmod(x, y : Double) : Double;
    {returns x mod y}
var
  a : Double;
begin
  if abs(y) < 1.0E-24 then
    Result := y
  else
    begin
      a := round(x / y);
      result := x - a * y;
    end;end;

{------------------------  TOde  ---------------------------}
function TOde.ComputeMatrix: Boolean;
var
  w, h: Integer;
begin
  w:= BGRAbm.width-1;
  h:= BGRAbm.height-1;
  result:= (w > 0) and (h > 0) and (edges[1] > 0) and (edges[2] > 0);
  if result then begin
    pd1:= w/edges[1];
    pd2:= -mi[1]*pd1;
    pd3:= -h/edges[2];
    pd4:= -pd3*ma[2];
  end;
end;

constructor TOde.Create;
var
  i: Integer;
begin
  inherited Create;
  BGRAbm:= MainForm.bmBGRA;
  BGRAbm.Fill(BGRAWhite, 0, BGRAbm.NbPixels);
  can:= MainForm.curvesCan;
  can.Brush.Color:= clWhite;
  can.FillRect(0, 0, BGRAbm.width, BGRAbm.Height);
  SetLength(cv2.ac, 20);
  cv2.last:= -1;
  for i:= 0 to high(cv2.ac) do begin
    SetLength(cv2.ac[i].c, 50);
    cv2.ac[i].last:= -1;
  end;
end;


procedure TOde.DrawCurves2D(w: Integer);

var
  i,j: Integer;
  cc: Array Of TPointF;
begin
  BGRAbm.PenStyle:= psSolid;
  if ( High(cv2.ac) >=0 ) and (cv2.last >=0) then
    for i:= 0 to cv2.last do
      DrawOneCurve2D(cv2.ac[i], w);
  if ( High(cvAux.ac) >=0 ) and (cvAux.last >=0) then begin
    with BGRAbm do begin
      CustomPenStyle:= BGRAPenStyle(4,4);
      for i:= 0 to cvAux.last do
        if cvAux.ac[i].last > 0 then begin
          SetLength(cc, cvAux.ac[i].last+1);
          for j:= 0 to cvAux.ac[i].last do
            FromParamsToDraw(cvAux.ac[i].c[j], cc[j]);
          DrawPolyLineAntialias(cc, BGRABlack, 1);
          cc:= nil;
        end;
        CustomPenStyle:= SolidPenStyle;
      end;
  end;
end;


procedure TOde.DrawCurves2DOnBGRAbm(bm: TBGRABitmap; r: TRect);
var
  i, j, w: Integer;
  pd1p, pd2p, pd3p, pd4p: Double;
  cc: Array Of TPointF;
begin
  pd1p:= pd1;
  pd2p:= pd2;
  pd3p:= pd3;
  pd4p:= pd4;
  pd1:= (r.right-r.left)/edges[1];
  pd2:= r.left - pd1*mi[1];
  pd3:= (r.top - r.bottom)/edges[2];
  pd4:= r.top - pd3*ma[2];
  w:= 1 + (r.bottom - r.top) div 500;
  try
    if cv2.last >= 0 then
      for i:= 0 to cv2.last do
        if cv2.ac[i].last > 0 then begin
          SetLength(cc, cv2.ac[i].last+1);
          for j:= 0 to cv2.ac[i].last do
            FromParamsToDraw(cv2.ac[i].c[j], cc[j]);
          bm.DrawPolyLineAntialias(cc, cv2.ac[i].color, w);
          cc:= nil;
        end;
    if ( High(cvAux.ac) >=0 ) and (cvAux.last >=0) then begin
      with bm do begin
        CustomPenStyle:= BGRAPenStyle(2*w, 2*w);
        for i:= 0 to cvAux.last do
          if cvAux.ac[i].last > 0 then begin
            SetLength(cc, cvAux.ac[i].last+1);
            for j:= 0 to cvAux.ac[i].last do
              FromParamsToDraw(cvAux.ac[i].c[j], cc[j]);
            DrawPolyLineAntialias(cc, BGRABlack, 1);
            cc:= nil;
          end;
          CustomPenStyle:= SolidPenStyle;
        end;
    end;
  finally
    pd1:= pd1p;
    pd2:= pd2p;
    pd3:= pd3p;
    pd4:= pd4p;
  end;
end;

procedure TOde.DrawOneCurve2D(crv: cv2D; w: Integer);
var
  i: Integer;
  cc: ArrayOfTPointF;
begin
  if crv.last > 0 then begin
    SetLength(cc, crv.last+1);
    for i:= 0 to crv.last do
      FromParamsToDraw(crv.c[i], cc[i]);
    BGRAbm.DrawPolyLineAntialias(cc, crv.color, w);
    cc:= nil;
  end;
end;

destructor TOde.Destroy;
var
  i: Integer;
begin
  if High(cv2.ac)>=0 then begin
    for i:= 0 to High(cv2.ac) do
      cv2.ac[i].c:= nil;
    cv2.ac:= nil;
  end;
  inherited Destroy;
end;

function TOde.DropLastCurve: Boolean;
{take last curve part as non existent by decrementing the parts count}
begin
  result:=  (high(cv2.ac) >=0) and  (cv2.last >=0);
  if result then begin
    cv2.ac[cv2.last].last:= -1;
    cv2.ac[cv2.last].c:= nil;
    dec(cv2.last);
  end;
end;

procedure TOde.FromParamsToDraw(x: array of Double; var p: TPointF);
begin
  p.x := pd1 * x[0] + pd2;
  p.y := pd3 * x[1] + pd4;
end;

procedure TOde.FromParamsToDraw(x: array of Double; var p: TPoint);
begin
  p.x := round(pd1 * x[0] + pd2);
  p.y := round(pd3 * x[1] + pd4);
end;

procedure TOde.InitCurves(solutions, vgCurve, auxCurves: Boolean);
var
  i: Integer;
begin
  if solutions then begin
    cv2.last:= -1;
    for i:= 0 to high(cv2.ac) do
      cv2.ac[i].last:= -1;
  end;
  if auxCurves then begin
    cvAux.last:= -1;
    for i:= 0 to high(cvAux.ac) do
      cvAux.ac[i].last:= -1;
  end;
end;

procedure TOde.Initiate(mi0, ma0: Vector3);
var
  i: Integer;
begin
  mi:= mi0;
  ma:= ma0;
  for i:= 1 to 3 do
    edges[i]:= ma[i]-mi[i];
  ComputeMatrix;
  span:= EuclDistance(ma, mi);
  rad:= hypot(ma[1]- mi[1], ma[2]- mi[2])/BGRAbm.width;
  radTotal:= span/BGRAbm.width;
end;

function TOde.NumberOfPieces(const lastX, nextX: array of double): Integer;
{number of pieces to divide the interval for filling the intermediate points
 with extrapolation in the ODE integrator}
var
  a: Double;
  p1, p2: TPointF;
begin
  FromParamsToDraw(lastX, p1);
  FromParamsToDraw(nextX, p2);
  a:= hypot(p1.x-p2.x, p1.y-p2.y);
  if a < 0.02 then
    result:= 0
  else
    result:= 1 + round(a/6);
end;

procedure TOde.RecordNode(x: array of double; key: Byte); 
var
  i,ind: Integer;
begin
  if key = 1 then begin
    inc(cv2.last);
    if high(cv2.ac) < cv2.last then begin
      SetLength(cv2.ac, cv2.last+10);
      for i:= 1 to 9 do
        cv2.ac[cv2.last+i].last:= -1;
    end;
    cv2.ac[cv2.last].last:= -1;
    cv2.ac[cv2.last].color:= cBGRAColor;
  end;
  inc(cv2.ac[cv2.last].last);
  if high(cv2.ac[cv2.last].c) < cv2.ac[cv2.last].last then
    SetLength(cv2.ac[cv2.last].c, cv2.ac[cv2.last].last+100);
  ind:= cv2.ac[cv2.last].last;
  for i:= 1 to 2 do
    cv2.ac[cv2.last].c[ind,i]:= x[i-1];
end;

procedure TOde.RecordNodeAux(x: array of double; key: Byte);
var
  i,ind: Integer;
begin
  if key = 1 then begin
    inc(cvAux.last);
    if high(cvAux.ac) < cvAux.last then begin
      SetLength(cvAux.ac, cvAux.last+10);
      for i:= 1 to 9 do
        cvAux.ac[cvAux.last+i].last:= -1;
    end;
    cvAux.ac[cvAux.last].last:= -1;
    cvAux.ac[cvAux.last].color:= BGRABlack;
  end;
  inc(cvAux.ac[cvAux.last].last);
  if high(cvAux.ac[cvAux.last].c) < cvAux.ac[cvAux.last].last then
    SetLength(cvAux.ac[cvAux.last].c, cvAux.ac[cvAux.last].last+100);
  ind:= cvAux.ac[cvAux.last].last;
  for i:= 1 to 2 do
    cvAux.ac[cvAux.last].c[ind,i]:= x[i-1];
end;


{--------------------- End of TOde -------------------------------------------}

{------------------------ TOdeEf----------------------------------------------}

procedure TOdeEf.AdjustToSurf(var x: array of double);
begin
  {not used}
end;

function TOdeEf.CheckCameBack(firstX, prevX, curX: array of double): Boolean;
var
  d: Double;
begin
  result:= EuclDistanceToSegment(d, 1, firstX, prevX, curX);
  result := result and (d < rad);
end;


function TOdeEf.CheckRange(nextX: array of double): Boolean;
var
    i: Integer;
begin
  result:= True;
  for i:= 0 to 1 do
    result:= result and (mi[i+1] < nextX[i]) and (nextX[i] < ma[i+1]);
end; 
 
procedure TOdeEf.ComputeCurve(par, dir: Vector2);
var
    x: Vector2;
begin
  if CheckRange(par) then begin
    MainForm.BeginWatch;
    try
      RK.SetEnvironment(2, rad, cExTolFO, @Derivative, @CheckRange,
         @RecordNode, nil, @CheckCameBack, nil, nil);
      x:= par;
        sense:= 1;
       {curve forward}
        rk.ComputeIntegralCurve(x, sense);
        if sense <> 0 then begin
          sense:= -1;
          x:= par;
         rk.ComputeIntegralCurve(x, sense);
        end;
    finally
      MainForm.EndWatch;
    end;
  end;
end;


constructor TOdeEf.Create;
begin
  inherited create;
  arrowScale:= 1/2;
end;


function TOdeEf.Derivative(var deriv: array of double; here:
           array of double): Boolean;
begin
  deriv[0]:= theF1.ff([here[0], here[1]]);
  deriv[1]:= theF2.ff([here[0], here[1]]);
  if sense = -1 then begin
    deriv[0]:= -deriv[0];
    deriv[1]:= -deriv[1];
  end;
  result:= True;
end;  


procedure TOdeEf.GetInitDirections(var dirs: directions; var dir: Vector2; par: Vector2);
begin
  {get the initial directions for an explicit first order ODE. They are
   simply the values at 'x' of the functions defining the ODE}
  SetLength(dirs.FOSO,1);
  dirs.last:= 0;
  dirs.FOSO[0,1]:= theF1.ff([par[1],par[2]]);
  dirs.FOSO[0,2]:= theF2.ff([par[1],par[2]]);
end;


function TOdeEf.Shift(x1,x2: array of double): Double;
var
    i: Integer;
begin
  result:= 0;
  for i:= 0 to 1 do
    result:= result + sqr(x2[i]-x1[i]);
  result:= sqrt(result);
end; 

{------------------------------ End of TOdeEf --------------------------------------}


{-------------------------------  TOdeEs -------------------------------------------}

procedure TOdeEs.AdjustToSurf(var x: array of double);
begin
  {not used}
end;

function TOdeEs.CheckCameBack(firstX, prevX, curX: array of double): Boolean;
var
  a, b, d: Double;
  i: Integer;
begin
  result:= EuclDistanceToSegment(d, 1, firstX, prevX, curX);
  if result then begin
    a:= 0;
    b:= 0;
    for i:= 2 to 3 do begin
      a := a + abs(2*firstX[i]- prevX[i]- curX[i]);
      b:= b + abs(curX[i]);
    end;
    result := (d < rad) and (a < 0.1*b)
  end;
end;  


function TOdeEs.CheckRange(nextX: array of double): Boolean;
var
  i: Integer;
begin
  result:= True;
  for i:= 0 to 1 do
    result:= result and (mi[i+1] < nextX[i]) and (nextX[i] < ma[i+1]) and
       (abs(nextX[2+i]) < 1.0e7);
end;
 
procedure TOdeEs.ComputeCurve(par, dir: Vector2);
var
  i: Integer;
  x, xIni: Vector4;
begin
  if (par[1] >= mi[1]) and (par[1] <= ma[1]) and (par[2] >= mi[2]) and (par[2] <= ma[2]) then begin
    RK.SetEnvironment(4, rad, cExTolSO, @Derivative, @CheckRange,
         @RecordNode, nil, @CheckCameBack, nil, @Shift);
    xIni[1]:= par[1];
    xIni[2]:= par[2];
    xIni[3]:= dir[1];
    xIni[4]:= dir[2];
    for i:= 1 to 4 do
      x[i]:= xIni[i];
    try
      MainForm.BeginWatch;
      sense:= 1;
      {compute curve forward. In second order cases we don't compute the curve
         backward: see Manual}
      rk.ComputeIntegralCurve(x, sense);
    finally
      MainForm.EndWatch;
    end;
  end;
end;


function TOdeEs.Derivative(var deriv: array of double; here:
           array of double): Boolean;
var
  i: Integer;
begin
  deriv[0]:= here[2];
  deriv[1]:= here[3];
  deriv[2]:= theF1.ff([here[0], here[1], here[2],here[3]]);
  deriv[3]:= theF2.ff([here[0], here[1], here[2],here[3]]);
  if sense = -1  then for i:= 0 to 3 do
    deriv[i]:= -deriv[i];
  result:= True;
end;

procedure TOdeEs.GetInitDirections(var dirs: directions; var dir: Vector2; par: Vector2);
begin
  SetLength(dirs.FOSO,1);
  dirs.last:= 0;
  dirs.FOSO[0,1]:= theF1.ff([par[1],par[2], dir[1], dir[2]]);
  dirs.FOSO[0,2]:= theF2.ff([par[1],par[2], dir[1], dir[2]]);
end;

  

function TOdeEs.Shift(x1,x2: array of double): Double;
 var
    a: Double;
    i: Integer;
begin
  a:= 0;
  for i:= 2 to 3 do
    a:= a + sqr(x2[i]-x1[i]);
  a:= ln(1+a);
  result:= 0;
  for i:= 0 to 1 do
    result:= result + sqr(x2[i]-x1[i]);
  result:= a + sqrt(result);
end; 

{------------------------------ Ends  TOdeEs --------------------------------------}

{---------------------------------  TODEIf   --------------------------------------}




procedure TOdeIf.AdjustToSurf(var x: array of double);
var
  gr1, gr2: Vector4;
  a, b, f1, f2, g11, g12, g22, determ, cut, h: Double;
  i, count: Integer;
begin
  count:= 0;
  f1:= theF1.ff([x[0], x[1], x[2],x[3],0,0]);
  f2:= theF2.ff([x[0], x[1], x[2],x[3],0,0]);
  h:= hypot(f1,f2);
  grad1(gr1, x);
  grad2(gr2, x);
  g11:= sqr(gr1[1])+sqr(gr1[2])+sqr(gr1[3])+sqr(gr1[4]);
  g22:= sqr(gr2[1])+sqr(gr2[2])+sqr(gr2[3])+sqr(gr2[4]);
  cut:= 1.0e-12 * radTotal * sqrt(g11+g22);
  while (h > cut) and (count < 5) do begin
    inc(count);
    g12:= gr1[1]*gr2[1] + gr1[2]*gr2[2]+ gr1[3]*gr2[3]+gr1[4]*gr2[4] ;
    determ:= g11*g22-sqr(g12);
    if determ > 1.0e-30 then begin
      a:= (f2*g12-f1*g22)/determ;
      b:= (f1*g12-f2*g11)/determ;
      for i:= 1 to 4 do
        x[i-1]:= x[i-1]+a*gr1[i]+ b*gr2[i];
      f1:= theF1.ff([x[0], x[1], x[2], x[3],0,0]);
      f2:= theF2.ff([x[0], x[1], x[2], x[3],0,0]);
      h:= hypot(f1,f2);
      grad1(gr1, x);
      grad2(gr2, x);
      g11:= sqr(gr1[1])+sqr(gr1[2])+sqr(gr1[3])+sqr(gr1[4]);
      g22:= sqr(gr2[1])+sqr(gr2[2])+sqr(gr2[3])+sqr(gr2[4]);
      cut:= 1.0e-12 * radTotal * sqrt(g11+g22);
    end
    else begin
      x[0]:= kElError;
      break;
    end;
  end;
  if h > cut then
    x[0]:= kElError;
end; 

function TOdeIf.CheckCameBack(firstX, prevX, curX: array of double): Boolean;
var
  d: Double;
begin
  result:= EuclDistanceToSegment(d, 1, firstX, prevX, curX);
  result := result and (d < rad);
end;  

function TOdeIf.CheckRange(nextX: array of double): Boolean;
var
  i: Integer;
begin
  result:= True;
  for i:= 0 to 1 do
    result:= result and (mi[i+1] < nextX[i]) and(ma[i+1] > nextX[i]);
end;

function TOdeIf.CheckRange4(nextX: array of double): Boolean;
var
  i: Integer;
begin
  result:= True;
  for i:= 0 to 1 do
    result:= result and (mi[i+1] < nextX[i]) and(ma[i+1] > nextX[i]);
  for i:= 2 to 3 do
    result:= result and (mi[3] < nextX[i]) and (ma[3] > nextX[i]);
end;

procedure TOdeIf.ComputeCurve(par, dir: Vector2);
var
  dirs: directions;
  i: Integer;
  dummy: Vector2;
  x,xIni: Vector4;
begin
  SetLength(dirs.FOSO,6);
  dirs.last:= -1;
  GetInitDirections(dirs,dummy, par); {find and store admissible initial directions}
  if (dirs.last >=0) and CheckRange(par) then begin
    rk.SetEnvironment(4, rad, cExTolFO,
       @Derivative, @CheckRange,
       @RecordNode, nil, @CheckCameBack, @AdjustToSurf, @Shift);
    for i:= 0 to dirs.last do begin
      xIni[1]:= par[1];
      xIni[2]:= par[2];

      {the found initial velocity}
      xIni[3]:= dirs.FOSO[i,1];
      xIni[4]:= dirs.FOSO[i,2];

      x:= xIni;
      try
        MainForm.BeginWatch;
        sense:= 1;
        {compute curve forward}
        rk.ComputeIntegralCurve(x, sense);
        DrawOneCurve2D(cv2.ac[cv2.last], 1);
        x:= xIni;
        if sense <> 0 then begin
          sense:= -1;
          rk.ComputeIntegralCurve(x, sense);
          DrawOneCurve2D(cv2.ac[cv2.last], 1);
        end;
      finally
        MainForm.EndWatch;
      end;
    end;
  end;
  dirs.FOSO:= nil;
end;  

procedure TOdeIf.ComputeVGCurve;
{compute discriminant}
var
  miC, maC: Vector4;
  i: Integer;
begin
  cvg.last:= -1;
  aux:= 0;
  auxInt:= 0;
  for i:= 1 to 2 do begin
    miC[i]:= mi[i];
    maC[i]:= ma[i];
  end;
  for i:= 3 to 4 do begin
    miC[i]:= -ma[3];
    maC[i]:= ma[3];
  end;
  try
    MainForm.BeginWatch;
    {find and store curve}
    ImplicitCurveDescent.ComputeCurve(ecVG, ecVGGrad, 4, @CheckRange,
       @RecordVGNode, cTolFO, miC, maC, 47, rad);
  finally
    MainForm.EndWatch;
  end;
end;

constructor TOdeIf.Create;
var
  i: Integer;
begin
  inherited Create;
  SetLength(ecVG,3);
  SetLength(ecVGGrad,3);
  SetLength(cvg.ac, 20);
  cvg.last:= -1;
  for i:= 0 to high(cvg.ac) do begin
    SetLength(cvg.ac[i].c, 50);
    cvg.ac[i].last:= -1;
  end;
end;

function TOdeIf.dd1(p, v: Vector2): Double;
{Directional @Derivative of 'theF1' at the point (xx,p) with respect to
 the vector 'v'. It is the limit when h->0 of
      (theF1(xx,p+h*v) - theF1(xx,p))/h.            }
var
 fd: FirstDeriv;
begin
  fd:= theF1.FirstD([xx[1],xx[2],p[1],p[2],0,0,v[1],v[2]]);
  result:= fd.dfa0;
end; 

function TOdeIf.dd2(p, v: Vector2): Double;
{Directional @Derivative of 'theF1' at the point (xx,p) with respect to
 the vector 'v'. It is the limit when h->0 of
      (theF1(xx,p+h*v) - theF1(xx,p))/h.            }
var
 fd: FirstDeriv;
begin
  fd:= theF2.FirstD([xx[1],xx[2],p[1],p[2],0,0,v[1],v[2]]);
  result:= fd.dfa0;
end;

procedure TOdeIf.Det(var f: Double; x: array of double);
{the function that defines the discriminant in the ImplicitFO equations}
var
  gr1, gr2: Vector4;
begin
  grad1(gr1, x);
  grad2(gr2, x);
  f:= gr1[3]*gr2[4]-gr1[4]*gr2[3];
end;

procedure TOdeIf.DetGrad(var gr: array of double; x: array of double);
{the gradient of 'ImplicitFODet'}
var
  f1x, f1y, f1uu, f1vv, f1ux, f1vx, f1xx, f1xy, f1uy, f1vy, f1yy,
  f2x, f2y, f2uu, f2vv, f2ux, f2vx, f2xx, f2xy, f2uy, f2vy, f2yy: Double;
  sd: secondDeriv;
begin
  sd:= theF1.SecondD([x[0], x[1], x[2], x[3], 0,0,1,0,0,0,0,0]);
  f1x:= sd.dfa0;
  f1xx:= sd.ddfa0;

  sd:= theF1.SecondD([x[0], x[1], x[2], x[3], 0,0,0,1,0,0,0,0]);
  f1y:= sd.dfa0;
  f1yy:= sd.ddfa0;

  sd:= theF1.SecondD([x[0], x[1], x[2], x[3], 0,0,1,1,0,0,0,0]);
  f1xy:= 0.5*(sd.ddfa0 - f1xx - f1yy);

  sd:= theF1.SecondD([x[0], x[1], x[2], x[3], 1,0,0,0,0,0,0,0]);
  f1uu:= sd.ddfa0;

  sd:= theF1.SecondD([x[0], x[1], x[2], x[3], 0,1,0,0,0,0,0,0]);
  f1vv:= sd.ddfa0;

  sd:= theF1.SecondD([x[0], x[1], x[2], x[3], 1,0,1,0,0,0,0,0]);
  f1ux:= 0.5*(sd.ddfa0 - f1xx - f1uu);

  sd:= theF1.SecondD([x[0], x[1], x[2], x[3], 1,0,0,1,0,0,0,0]);
  f1uy:= 0.5*(sd.ddfa0 - f1yy - f1uu);

  sd:= theF1.SecondD([x[0], x[1], x[2], x[3], 0,1,1,0,0,0,0,0]);
  f1vx:= 0.5*(sd.ddfa0 - f1xx - f1vv);

  sd:= theF1.SecondD([x[0], x[1], x[2], x[3], 0,1,0,1,0,0,0,0]);
  f1vy:= 0.5*(sd.ddfa0 - f1yy - f1vv);

  sd:= theF2.SecondD([x[0], x[1], x[2], x[3], 0,0,1,0,0,0,0,0]);
  f2x:= sd.dfa0;
  f2xx:= sd.ddfa0;

  sd:= theF2.SecondD([x[0], x[1], x[2], x[3], 0,0,0,1,0,0,0,0]);
  f2y:= sd.dfa0;
  f2yy:= sd.ddfa0;

  sd:= theF2.SecondD([x[0], x[1], x[2], x[3], 0,0,1,1,0,0,0,0]);
  f2xy:= 0.5*(sd.ddfa0 - f2xx - f2yy);

  sd:= theF2.SecondD([x[0], x[1], x[2], x[3], 1,0,0,0,0,0,0,0]);
  f2uu:= sd.ddfa0;

  sd:= theF2.SecondD([x[0], x[1], x[2], x[3], 0,1,0,0,0,0,0,0]);
  f2vv:= sd.ddfa0;

  sd:= theF2.SecondD([x[0], x[1], x[2], x[3], 1,0,1,0,0,0,0,0]);
  f2ux:= 0.5*(sd.ddfa0 - f2xx - f2uu);

  sd:= theF2.SecondD([x[0], x[1], x[2], x[3], 1,0,0,1,0,0,0,0]);
  f2uy:= 0.5*(sd.ddfa0 - f2yy - f2uu);

  sd:= theF2.SecondD([x[0], x[1], x[2], x[3], 0,1,1,0,0,0,0,0]);
  f2vx:= 0.5*(sd.ddfa0 - f2xx - f2vv);

  sd:= theF2.SecondD([x[0], x[1], x[2], x[3], 0,1,0,1,0,0,0,0]);
  f2vy:= 0.5*(sd.ddfa0 - f2yy - f2vv);

  {the following formula comes from the fact that 'ImplicitFODet' can be
   written as    f1x*f2y - f1y*f2x  }
  gr[0]:= f1ux*f2y + f1x*f2uy - f1uy*f2x- f1y*f2ux;
  gr[1]:= f1vx*f2y + f1x*f2vy - f1vy*f2x- f1y*f2vx;
  gr[2]:= f1xx*f2y + f1x*f2xy - f1xy*f2x- f1y*f2xx;
  gr[3]:= f1xy*f2y + f1x*f2yy - f1yy*f2x- f1y*f2xy;
end;


function TOdeIf.Derivative(var deriv: array of double; here:
           array of double): Boolean;
var
  gr1, gr2: Vector4;
  determ, ind1, ind2: Double;
  i: Integer;
begin
  grad1(gr1, here);
  grad2(gr2, here);
  determ:= gr1[3]*gr2[4]-gr1[4]*gr2[3];

  deriv[0]:= here[2];  {of course, the @Derivative of u is u', the third coordinate}
  deriv[1]:= here[3];  {of course, the @Derivative of v is v', the fourth coordinate}

  {solve the linear equation}
  result:= abs(determ) > 1.0e-20;
  if result then begin
    ind1:= -here[2]*gr1[1]-here[3]*gr1[2];
    ind2:= -here[2]*gr2[1]-here[3]*gr2[2];
    determ:= 1/determ;
    deriv[2]:= determ*(ind1*gr2[4]- ind2*gr1[4]);
    deriv[3]:= determ*(ind2*gr1[3]- ind1*gr2[3]);
  end;
  if (sense= -1) then for i:= 0 to 3 do
    deriv[i]:= - deriv[i];
end;



destructor TOdeIf.Destroy;
var
  i: Integer;
begin
  if High(cvg.ac)>=0 then begin
    for i:= 0 to High(cvg.ac) do
      cvg.ac[i].c:= nil;
    cvg.ac:= nil;
  end;
  ecVG:= nil;
  ecVGGrad:= nil; 
  inherited Destroy;
end;

procedure TOdeIf.DrawCurves2D(w: Integer);
var
   i: Integer;
begin
  inherited;
  if ( High(cvg.ac) >=0 ) and (cvg.last >=0) then
    for i:= 0 to cvg.last do
      DrawOneCurve2D(cvg.ac[i], 2*w);
end;

procedure TOdeIf.DrawCurves2DOnBGRAbm(bm: TBGRABitmap; r: TRect);
var
  i, j, w: Integer;
  pd1p, pd2p, pd3p, pd4p: Double;
  cc: ArrayOfTPointF;
begin
  pd1p:= pd1;
  pd2p:= pd2;
  pd3p:= pd3;
  pd4p:= pd4;
  pd1:= (r.right-r.left)/edges[1];
  pd2:= r.left - pd1*mi[1];
  pd3:= (r.top - r.bottom)/edges[2];
  pd4:= r.top - pd3*ma[2];
  w:= 1 + (r.bottom - r.top) div 500;
  try
    if ( High(cv2.ac) >=0 ) and (cv2.last >=0) then
      for i:= 0 to cv2.last do
        if cv2.ac[i].last > 0 then begin
          SetLength(cc, cv2.ac[i].last+1);
          for j:= 0 to cv2.ac[i].last do
            FromParamsToDraw(cv2.ac[i].c[j], cc[j]);
          bm.DrawPolyLineAntialias(cc, cv2.ac[i].color, w);
          cc:= nil;
        end;
    w:= 1 + (r.bottom - r.top) div 500;
    if ( High(cvg.ac) >=0 ) and (cvg.last >=0) then
      for i:= 0 to cvg.last do
        if cvg.ac[i].last > 0 then begin
          SetLength(cc, cvg.ac[i].last+1);
          for j:= 0 to cvg.ac[i].last do
            FromParamsToDraw(cvg.ac[i].c[j], cc[j]);
          bm.DrawPolyLineAntialias(cc, cvg.ac[i].color, 1.5*w);
          cc:= nil;
        end;
    if (High(cvAux.ac) >=0 ) and (cvAux.last >=0) then begin
      with bm do begin
        CustomPenStyle:= BGRAPenStyle(2*w, 2*w);
        for i:= 0 to cvAux.last do
          if cvAux.ac[i].last > 0 then begin
            SetLength(cc, cvAux.ac[i].last+1);
            for j:= 0 to cvAux.ac[i].last do
              FromParamsToDraw(cvAux.ac[i].c[j], cc[j]);
            DrawPolyLineAntialias(cc, BGRABlack, 1);
            cc:= nil;
          end;
          CustomPenStyle:= SolidPenStyle;
        end;
    end;
  finally
    pd1:= pd1p;
    pd2:= pd2p;
    pd3:= pd3p;
    pd4:= pd4p;
  end;
end;

procedure TOdeIf.Eq1(var f: Double; x: Vector2);
begin
  f:= theF1.ff([xx[1],xx[2],x[1],x[2]]);
end;

procedure TOdeIf.Eq2(var f: Double; x: Vector2);
begin
  f:= theF2.ff([xx[1],xx[2],x[1],x[2]]);
end;

procedure TOdeIf.GetInitDirections(var dirs: directions; var dir: Vector2; par: Vector2);
  var
    i: Integer;
    minCorner0, maxCorner0: Vector2;
begin
  try
    {get initial directions for an Implicit first order ODE}

    {prepare data for passing them to 'ComputePointsInR2'}
    xx:= par;
    minCorner0[1]:= -ma[3];
    minCorner0[2]:= minCorner0[1];
    maxCorner0[1]:= ma[3];
    maxCorner0[2]:= maxCorner0[1];

    {Get the points in the configuration domain in R^4, where the functions
     'eq1,eq2' both vanish}
    GetPointsInR2.ComputePointsInR2(@eq1, @eq2, @dd1, @dd2,
        minCorner0, maxCorner0, 1.0e-7, ma[3]/100, 100, 600);

     {store results}
    if GetPointsInR2.pointsFound.last >=0 then begin
      SetLength(dirs.FOSO,GetPointsInR2.pointsFound.last+1);
      dirs.last:= GetPointsInR2.pointsFound.last;
      for i:= 0 to GetPointsInR2.pointsFound.last do
        dirs.FOSO[i]:= GetPointsInR2.pointsFound.FOSO[i];
    end;
  except on exception do
    MainForm.Inform(eBadPointEncountered, False);
  end;
end;

procedure TOdeIf.Grad1(var f: array of Double; x: array of double);
  var
    fd: FirstDeriv;
  begin
    fd:= theF1.firstD([x[0],x[1],x[2],x[3],1,0,0,0]);
    f[0]:= fd.dfa0;
    fd:= theF1.firstD([x[0],x[1],x[2],x[3],0,1,0,0]);
    f[1]:= fd.dfa0;
    fd:= theF1.firstD([x[0],x[1],x[2],x[3],0,0,1,0]);
    f[2]:= fd.dfa0;
    fd:= theF1.firstD([x[0],x[1],x[2],x[3],0,0,0,1]);
    f[3]:= fd.dfa0;
  end;

procedure TOdeIf.Grad2(var f: array of Double; x: array of double);
  var
    fd: FirstDeriv;
  begin
    fd:= theF2.firstD([x[0],x[1],x[2],x[3],1,0,0,0]);
    f[0]:= fd.dfa0;
    fd:= theF2.firstD([x[0],x[1],x[2],x[3],0,1,0,0]);
    f[1]:= fd.dfa0;
    fd:= theF2.firstD([x[0],x[1],x[2],x[3],0,0,1,0]);
    f[2]:= fd.dfa0;
    fd:= theF2.firstD([x[0],x[1],x[2],x[3],0,0,0,1]);
    f[3]:= fd.dfa0;
  end;

procedure TOdeIf.InitCurves(solutions, vgCurve, auxCurves: Boolean);
var
  i: Integer;
begin
  inherited;
  if vgCurve then
    for i:= 0 to high(cvg.ac) do begin
      SetLength(cvg.ac[i].c, 50);
      cvg.ac[i].last:= -1;
    end;
end;

procedure TOdeIf.Initiate(mi0, ma0: Vector3);
      {prepare everything for drawing curves or obtaining other data}
begin
  inherited Initiate(mi0, ma0);
  with MainForm do begin
    ecVG[0]:= @vg1;
    ecVGGrad[0]:= @grad1;
    ecVG[1]:= @vg2;
    ecVGGrad[1]:= @grad2;
    ecVG[2]:= @Det;
    ecVGGrad[2]:= @DetGrad;
  end;
end;

procedure TOdeIf.RecordVGNode(x: array of double; key: Byte);
{store the new node and draw the segment defined with the previous one if key=0}
   var
    i,ind: Integer;
begin
  if key = 1 then begin
    inc(cvg.last);
    if high(cvg.ac) < cvg.last then begin
      SetLength(cvg.ac, cvg.last+10);
      for i:= 1 to 9 do
        cvg.ac[cvg.last+i].last:= -1;
    end;
    cvg.ac[cvg.last].last:= -1;
    cvg.ac[cvg.last].color:= BGRA(200, 200, 200, 255);
  end;
  inc(cvg.ac[cvg.last].last);
  if high(cvg.ac[cvg.last].c) < cvg.ac[cvg.last].last then
    SetLength(cvg.ac[cvg.last].c, cvg.ac[cvg.last].last+100);
  ind:= cvg.ac[cvg.last].last;
  for i:= 1 to 2 do
    cvg.ac[cvg.last].c[ind,i]:= x[i-1];
end;    
  
function TOdeIf.Shift(x1,x2: array of double): Double;
 var
    i: Integer;
begin
  result:= 0;
  for i:= 0 to 1 do
    result:= result + sqr(x2[i]-x1[i]);
  result:= sqrt(result);
end;

procedure TOdeIf.vg1(var f: Double; x: array of double);
begin
  f:= theF1.ff(x);
end;

procedure TOdeIf.vg2(var f: Double; x: array of double);
begin
  f:= theF2.ff(x);
end;


{------------------------------ End of TOdeIf --------------------------------------}

{---------------------------------  TODEIs   --------------------------------------}


procedure TOdeIs.AdjustToSurf(var x: array of double);
{similar to the 'AdjustToSurfImplicitFO'.}
var
  gr1, gr2: Array[0..5] of Double;
  a, b, f1, f2, g11, g12, g22, det, cut, h: Double;
  i, count: Integer;
begin
  count:= 0;
  f1:= theF1.ff([x[0], x[1], x[2], x[3], x[4], x[5]]);
  f2:= theF2.ff([x[0], x[1], x[2], x[3], x[4], x[5]]);
  h:= hypot(f1,f2);
  Gradients(gr1, gr2, x);
  g11:= sqr(gr1[0])+sqr(gr1[1])+sqr(gr1[2])+sqr(gr1[3])+ sqr(gr1[4])+sqr(gr1[5]);
  g22:= sqr(gr2[0])+sqr(gr2[1])+sqr(gr2[2])+sqr(gr2[3])+ sqr(gr2[4])+sqr(gr2[5]);
  cut:= 1.0e-12 * radTotal * sqrt(g11+g22);
  while (h > cut) and (count < 5) do begin
    inc(count);
    g12:= gr1[0]*gr2[0] + gr1[1]*gr2[1]+ gr1[2]*gr2[2]+ gr1[3]*gr2[3]+
          gr1[4]*gr2[4]+ gr1[5]*gr2[5];
    det:= g11*g22-sqr(g12);
    if det > 1.0e-30 then begin
      a:= (f2*g12-f1*g22)/det;
      b:= (f1*g12-f2*g11)/det;
      for i:= 0 to 5 do
        x[i]:= x[i]+a*gr1[i]+ b*gr2[i];
      f1:= theF1.ff([x[0], x[1], x[2], x[3], x[4], x[5]]);
      f2:= theF2.ff([x[0], x[1], x[2], x[3], x[4], x[5]]);
      h:= hypot(f1,f2);
      Gradients(gr1, gr2, x);
      g11:= sqr(gr1[0])+sqr(gr1[1])+sqr(gr1[2])+sqr(gr1[3])+ sqr(gr1[4])+sqr(gr1[5]);
      g22:= sqr(gr2[0])+sqr(gr2[1])+sqr(gr2[2])+sqr(gr2[3])+ sqr(gr2[4])+sqr(gr2[5]);
      cut:= 1.0e-12 * radTotal * sqrt(g11+g22);
    end
    else begin
      x[0]:= kElError;
      break;
    end;
  end;
  if h > cut then
    x[0]:= kElError;
end;

function TOdeIs.CheckCameBack(firstX, prevX, curX: array of double): Boolean;
var
  a, b, d: Double;
  i: Integer;
begin
  result:= EuclDistanceToSegment(d, 1, firstX, prevX, curX);
  if result then begin
    a:= 0;
    b:= 0;
    for i:= 2 to 3 do begin
      a := a + abs(2*firstX[i]- prevX[i]- curX[i]);
      b:= b + abs(curX[i]);
    end;
    result := (d < rad) and (a < 0.1*b)
  end;
end;  

function TOdeIs.CheckRange(nextX: array of double): Boolean;
var
  i: Integer;
begin
  result:= True;
  for i:= 0 to 1 do
    result:= result and (mi[i+1] < nextX[i]) and(ma[i+1] > nextX[i]);
end;

procedure TOdeIs.ComputeCurve(par, dir: Vector2);
{computes and draws a solution to an Implicit Implicit second order ODE}
type
  vector6= array[1..6] of Double;
var
    dirs: directions;
    i: Integer;
    x, xIni: Vector6;
begin
  SetLength(dirs.FOSO,6);
  dirs.last:= -1;
  GetInitDirections(dirs, dir, par); {find and store admissible initial accelerations}
  if (dirs.last >=0) and CheckRange(par) then begin
    RK.SetEnvironment(6, rad, cExTolSO, @Derivative, @CheckRange,
         @RecordNode, nil, @CheckCameBack, @AdjustToSurf, @Shift);
    for i:= 0 to dirs.last do begin
      xIni[1]:= par[1];
      xIni[2]:= par[2];
      xIni[3]:= dir[1];
      xIni[4]:= dir[2];

      {the found initial acceleration}
      xIni[5]:= dirs.FOSO[i,1];
      xIni[6]:= dirs.FOSO[i,2];

      x:= xIni;
      try
        MainForm.BeginWatch;
        sense:= 1;
        {compute curve forward. In second order cases we don't compute the curve
         backward: see Manual}
        RK.ComputeIntegralCurve(x, sense);
        MainForm.EndWatch;
      finally
        MainForm.EndWatch;
      end;
    end;
  end;
  dirs.FOSO:= nil;
end; 


function TOdeIs.dd1(p, v: Vector2): Double;
{Directional @Derivative of 'theF1' at the point (xx,vv,p) with respect to
 the vector 'v'. It is the limit when h->0 of
    (theF1(xx,vv,p+h*v)- theF1(xx,vv,p))/h.}
var
 fd: FirstDeriv;
begin
  fd:= theF1.FirstD([xx[1],xx[2],vv[1],vv[2],p[1],p[2],0,0,0,0,v[1],v[2]]);
  result:= fd.dfa0;
end; 

function TOdeIs.dd2(p, v: Vector2): Double;
{Directional @Derivative of 'theF1' at the point (xx,vv,p) with respect to
 the vector 'v'. It is the limit when h->0 of
    (theF1(xx,vv,p+h*v)- theF1(xx,vv,p))/h.}
var
 fd: FirstDeriv;
begin
  fd:= theF2.FirstD([xx[1],xx[2],vv[1],vv[2],p[1],p[2],0,0,0,0,v[1],v[2]]);
  result:= fd.dfa0;
end;

function TOdeIs.Derivative(var deriv: array of double; here:
      array of double): Boolean;
{Gives the derivatives of the variables u,v,u',v',u'',v'' along a curve
 solution of a Implicit implicit second order ODE. See comments on
 'DerivativeFOImplicit'.}
var
  gr1, gr2: Array[0..5] of Double;
  det, ind1, ind2: Double;
  i: Integer;
begin
  deriv[0]:= here[2];
  deriv[1]:= here[3];
  deriv[2]:= here[4];
  deriv[3]:= here[5];
  Gradients(gr1, gr2, here);
  det:= gr1[4]*gr2[5]- gr1[5]*gr2[4];
  result:= abs(det) > 1.0e-20;
  if result then begin
    ind1:= -(here[2]*gr1[0]+here[3]*gr1[1]+here[4]*gr1[2]+here[5]*gr1[3]);
    ind2:= -(here[2]*gr2[0]+here[3]*gr2[1]+here[4]*gr2[2]+here[5]*gr2[3]);
    det:= 1/det;
    deriv[4]:= det*(ind1*gr2[5]- ind2*gr1[5]);
    deriv[5]:= det*(ind2*gr1[4]- ind1*gr2[4]);
    if (sense = -1) then for i:= 0 to 5 do
      deriv[i]:= - deriv[i];
  end
end;    

procedure TOdeIs.Eq1(var f: Double; x: Vector2);
{wrap for the first function defining a Implicit implicit ODE of second order}
begin
  f:= theF1.ff([xx[1],xx[2],vv[1], vv[2], x[1],x[2]]);
end;

procedure TOdeIs.Eq2(var f: Double; x: Vector2);
{wrap for the first function defining a Implicit implicit ODE of second order}
begin
  f:= theF2.ff([xx[1],xx[2],vv[1], vv[2], x[1],x[2]]);
end;

 

procedure TOdeIs.GetInitDirections(var dirs: directions; var dir: Vector2; par: Vector2);
{get admissible initial accelerations for an ODE of second order}
  var
    i: Integer;
    minCorner0, maxCorner0: Vector2;
begin
  try
    xx:= par;
    vv:= dir;
    minCorner0[1]:= -ma[3];
    minCorner0[2]:= minCorner0[1];
    maxCorner0[1]:= ma[3];
    maxCorner0[2]:= maxCorner0[1];
    GetPointsInR2.ComputePointsInR2(@eq1,@eq2,
     @dd1, @dd2, 
      minCorner0, maxCorner0, 1.0e-7, ma[3]/100, 50, 800);
    if GetPointsInR2.pointsFound.last >=0 then begin
      SetLength(dirs.FOSO,GetPointsInR2.pointsFound.last+1);
      dirs.last:= GetPointsInR2.pointsFound.last;
      for i:= 0 to GetPointsInR2.pointsFound.last do
        dirs.FOSO[i]:= GetPointsInR2.pointsFound.FOSO[i];
    end; 
  except on exception do
    MainForm.Inform(eBadPointEncountered, False);
  end;
end;  

procedure TOdeIs.Gradients(var gr1, gr2: array of double;
       const x: array of double);
{gradient of the functions 'theF1' and 'theF2'}
var
  i: Integer;
  xv: array[0..11] of Double;
  fp: FirstDeriv;
begin
  for i:= 0 to 5 do begin
    xv[i]:= x[i];
    xv[6+i]:= 0;
  end;
  for i:= 0 to 5 do begin
    xv[6+i]:= 1.0;
    fp:= theF1.FirstD(xv);
    gr1[i]:= fp.dfa0;
    fp:= theF2.FirstD(xv);
    gr2[i]:= fp.dfa0;
    xv[6+i]:= 0;
  end;
end;


function TOdeIs.Shift(x1,x2: array of double): Double;
  var
    a: Double;
    i: Integer;
begin
  a:= 0;
  for i:= 3 to 5 do
    a:= a + sqr(x2[i]-x1[i]);
  a:= ln(1+a);
  result:= 0;
  for i:= 0 to 2 do
    result:= result + sqr(x2[i]-x1[i]);
  result:= a + sqrt(result);
end;   

{------------------------------  End ofTOdeIs ------------------------------------}

{------------------------------  TOdeL begins ------------------------------------}

procedure TOdeL.AdjustToSurf(var x: array of double);
begin
  {not used}
end;

function TOdeL.CheckCameBack(firstX, prevX, curX: array of double): Boolean;
var
  a, b, d: Double;
  i: Integer;
begin
  result:= EuclDistanceToSegment(d, 1, firstX, prevX, curX); 
  if result then begin
    a:= 0;
    b:= 0;
    for i:= 2 to 3 do begin
      a := a + abs(2*firstX[i]- prevX[i]- curX[i]);
      b:= b + abs(curX[i]);
    end;
    result := (d < rad) and (a < 0.1*b)
  end;
end; 

function TOdeL.CheckRange(nextX: array of double): Boolean;
{Check whether 'nextX' is in the domain and the curvature < 10^7}
var
  i: Integer;
begin
  result:= True;
  for i:= 0 to 1 do
    result:= result and (mi[i+1] < nextX[i]) and (nextX[i] < ma[i+1]) and
       (abs(nextX[2+i]) < 1.0e7);
end;

procedure TOdeL.ComputeCurve(par, dir: Vector2);
{compute a curve solution to Lagrange equations derived from a Lagrangian
 function}
var
  i: Integer;
  x, xIni: Vector4;
begin
  RK.SetEnvironment(4, rad, cExTolFO, @Derivative, @CheckRange,
         @RecordNode, nil, @CheckCameBack, nil, @Shift);

    {initial conditions chosen by user}
  xIni[1]:= par[1];
  xIni[2]:= par[2];
  xIni[3]:= dir[1];
  xIni[4]:= dir[2];

  for i:= 1 to 4 do
    x[i]:= xIni[i];
  try
    MainForm.BeginWatch;
    sense:= 1;
      {compute curve forward. In second order cases we don't compute the curve
       backward: see Manual}
    rk.ComputeIntegralCurve(x, sense);
  finally
    MainForm.EndWatch;
  end;
end;

function TOdeL.Derivative(var deriv: array of double; here:
      array of double): Boolean;
{Gives the derivatives of the variables 'u,v,u',v' along a curve solution
 of the Lagrange equations}
var
  fu, fv, fux, fuy, fvx, fvy, fxx, fxy, fyy, a0, b0, d: Double;
begin
  DerivativesLagrangian(fu, fv, fux, fuy, fvx, fvy, fxx, fxy, fyy, here);
  a0:= fu- fux*here[2] - fvx*here[3];
  b0:= fv- fuy*here[2] - fvy*here[3];
  d:= fxx*fyy - sqr(fxy);
  result:= abs(d) > 1.0e-18;
  if result then begin
    deriv[0]:= here[2];
    deriv[1]:= here[3];
    deriv[2]:= (a0*fyy - b0*fxy)/d;
    deriv[3]:= (b0*fxx - a0*fxy)/d;
  end;
end; 

procedure TOdeL.DerivativesLagrangian(var fu, fv, fux, fuy, fvx, fvy, fxx, fxy, fyy:
       Double; const x: array of double);
{first and second derivatives of the Lagrangian function}
var
  fuu, fvv: Double;
  sd: secondDeriv;
begin
  sd:= theF1.SecondD([x[0], x[1], x[2], x[3], 1,0,0,0,0,0,0,0]);
  fu:= sd.dfa0;
  fuu:= sd.ddfa0;

  sd:= theF1.SecondD([x[0], x[1], x[2], x[3], 0,1,0,0,0,0,0,0]);
  fv:= sd.dfa0;
  fvv:= sd.ddfa0;

  sd:= theF1.SecondD([x[0], x[1], x[2], x[3], 0,0,1,0,0,0,0,0]);
  fxx:= sd.ddfa0;

  sd:= theF1.SecondD([x[0], x[1], x[2], x[3], 0,0,0,1,0,0,0,0]);
  fyy:= sd.ddfa0;

  sd:= theF1.SecondD([x[0], x[1], x[2], x[3], 0,0,1,1,0,0,0,0]);
  fxy:= 0.5*(sd.ddfa0 - fxx - fyy);

  sd:= theF1.SecondD([x[0], x[1], x[2], x[3], 1,0,1,0,0,0,0,0]);
  fux:= 0.5*(sd.ddfa0 - fxx - fuu);

  sd:= theF1.SecondD([x[0], x[1], x[2], x[3], 1,0,0,1,0,0,0,0]);
  fuy:= 0.5*(sd.ddfa0 - fyy - fuu);

  sd:= theF1.SecondD([x[0], x[1], x[2], x[3], 0,1,1,0,0,0,0,0]);
  fvx:= 0.5*(sd.ddfa0 - fxx - fvv);

  sd:= theF1.SecondD([x[0], x[1], x[2], x[3], 0,1,0,1,0,0,0,0]);
  fvy:= 0.5*(sd.ddfa0 - fyy - fvv);
end;

procedure TOdeL.GetInitDirections(var dirs: directions; var dir: Vector2; par: Vector2);
begin
  {not used}
end;


function TOdeL.Shift(x1,x2: array of double): Double;
  var
    a: Double;
    i: Integer;
begin
  a:= 0;
  for i:= 2 to 3 do
    a:= a + sqr(x2[i]-x1[i]);
  a:= ln(1+a);
  result:= 0;
  for i:= 0 to 1 do
    result:= result + sqr(x2[i]-x1[i]);
  result:= a + sqrt(result);
end; 

{------------------------------  End ofTOdeL ------------------------------------}

{------------------------------  TOdeH begins ------------------------------------}

procedure TOdeH.AdjustToSurf(var x: array of double);
begin
  {not used}
end;


function TOdeH.CheckCameBack(firstX, prevX, curX: array of double): Boolean;
var
  a, b, d: Double;
  i: Integer;
begin
  result:= EuclDistanceToSegment(d, 1, firstX, prevX, curX);
  if result then begin
    a:= 0;
    b:= 0;
    for i:= 2 to 3 do begin
      a := a + abs(2*firstX[i]- prevX[i]- curX[i]);
      b:= b + abs(curX[i]);
    end;
    result := (d < rad) and (a < 0.1*b)
  end;
end; 

function TOdeH.CheckRange(nextX: array of double): Boolean;
{Check whether 'nextX' is in the domain and the curvature < 10^7}
var
  i: Integer;
begin
  result:= True;
  for i:= 0 to 1 do
    result:= result and (mi[i+1] < nextX[i]) and (nextX[i] < ma[i+1]) and
       (abs(nextX[2+i]) < 1.0e7);
end;

procedure TOdeH.ComputeCurve(par, dir: Vector2);
{compute a curve solution to Hamilton equations derived from a Hamiltonian
 function}
var
  i: Integer;
  x, xIni: Vector4;
begin
  RK.SetEnvironment(4, rad, cExTolFO, @Derivative, @CheckRange,
         @RecordNode, nil, @CheckCameBack, nil, @Shift);
    {initial conditions chosen by user}
  xIni[1]:= par[1];
  xIni[2]:= par[2];
  xIni[3]:= dir[1];
  xIni[4]:= dir[2];

  for i:= 1 to 4 do
    x[i]:= xIni[i];
  try
    MainForm.BeginWatch;
    try
      sense:= 1;
      {compute curve forward. In second order cases we don't compute the curve
       backward: see Manual}
      rk.ComputeIntegralCurve(x, sense);
    except on E: Exception do
      MainForm.Inform(E.Message, True);
    end;
  finally
    MainForm.EndWatch;
  end;
end;


function TOdeH.Derivative(var deriv: array of double; here:
      array of double): Boolean;
{Gives the derivatives of the variables 'u,v,x,y' along a curve solution
 of the Hamilton equations}
var
  i: Integer;
  fd: FirstDeriv;
begin
  fd:= theF1.FirstD([here[0], here[1], here[2], here[3], 1,0,0,0]);
  deriv[2]:= -fd.dfa0;
  fd:= theF1.FirstD([here[0], here[1], here[2], here[3], 0,1,0,0]);
  deriv[3]:= -fd.dfa0;
  fd:= theF1.FirstD([here[0], here[1], here[2], here[3], 0,0,1,0]);
  deriv[0]:= fd.dfa0;
  fd:= theF1.FirstD([here[0], here[1], here[2], here[3], 0,0,0,1]);
  deriv[1]:= fd.dfa0;

  if (sense = -1) then for i:= 0 to 3 do
    deriv[i]:= -deriv[i];
  result:= True;
end;


procedure TOdeH.GetInitDirections(var dirs: directions; var dir: Vector2; par: Vector2);
begin
  {not used}
end;

function TOdeH.Shift(x1,x2: array of double): Double;
var
  i: Integer;
begin
  result:= 0;
  for i:= 0 to 3 do
    result:= result + sqr(x2[i]-x1[i]);
  result:= sqrt(result);
end; 

{------------------------------  End ofTOdeL ------------------------------------}

{------------------------------  TOdeNf begins ------------------------------------}

procedure TOdeNf.AdjustToSurf(var x: array of double);
{the solution curves are curves on a surface in R^3. Little
 local errors make the curves drift off the surface, and
 this routine projects orthogonally 'x' on the surface for
 correcting that deviation}
  var
    gr: Vector3;
    a,b, cut: Double;
    i, count: Integer;
    rslt: Boolean;
begin
  count:= 0;
  gr:= SfGrad(x); {it is orthogonal to the surface}

  {radTotal is more or less the length in R^3 that projects to
   a length of a pixel. Thus, EuclNorm(gr)*radTotal is the change of
   the function along a pixel. Imposing this 'cut' is to say that
   the error must be less than 1.0e-12 times what the function
   changes along a pixel }
  cut:= 1.0e-12*EuclNorm(gr)*radTotal;

  a:= Sf(x);
  rslt:= ( abs(a) < cut );
  if not rslt then begin
    while (not rslt) and (count < 20) do begin
      inc(count);
      gr:= SfGrad(x);
      b:= EuclEnergy(gr);
      if b > 1.0e-24 then begin
        b:= -a/b;
        for i:= 0 to 2 do
          x[i]:= x[i]+b*gr[i+1]; {this is the first order approximation
             to the point where 'x' projects to the surface}
      end;
      a:= Sf(x);
      rslt:= ( abs(a) < cut );
    end;
    if not rslt then
      x[0]:= kElError;
  end;
end; 

function TOdeNf.CheckCameBack(firstX, prevX, curX: array of double): Boolean;
var
  d: Double;
begin
  result:= EuclDistanceToSegment(d, 1, firstX, prevX, curX);
  result := result and (d < rad);
  d:= abs(prevX[2]+ curX[2]-2*firstX[2])/2;
  result:= result and (abs(d) < 0.1);
end;  

function TOdeNf.CheckRange(nextX: array of double): Boolean;
{Check whether 'nextX' is in the domain. Here the angle 'w' is limited by 12*pi}
  var
    i: Integer;
begin
  result:= True;
  for i:= 0 to 1 do
    result:= result and (mi[i+1] < nextX[i]) and (nextX[i] < ma[i+1]);
end;

function TOdeNf.CheckRange3D(nextX: array of double): Boolean;
{Check whether 'nextX' is in the domain. Here the angle 'w' is limited also}
  var
    i: Integer;
begin
  result:= True;
  for i:= 0 to 2 do
    result:= result and (mi[i+1] < nextX[i]) and (nextX[i] < ma[i+1]);
end;


procedure TOdeNf.ComputeCurve(par, dir: Vector2);
  var
    dirs: directions;
    i: Integer;
    x: Vector3;
    dummy: Vector2;
begin
  SetLength(dirs.FOSO,6);
  dirs.last:= -1;
  GetInitDirections(dirs,dummy,par); {find and store in 'dirs' the admissible initial
                               directions}
  if dirs.last >=0 then begin
    rk.SetEnvironment(3, rad, cExTolFO, @Derivative, @CheckRange,
        @RecordNode, nil, @CheckCameBack,  @AdjustToSurf,
        nil);
    try
      MainForm.BeginWatch;
      for i:= 0 to dirs.last do begin
        x[1]:= par[1]; x[2]:= par[2]; x[3]:= dirs.FOSO[i,1];
        if CheckRange(x) then begin
          sense:= 1;
          {compute curve forward}
          rk.ComputeIntegralCurve(x, sense);
          if (sense <> 0) then begin
            x[1]:= par[1]; x[2]:= par[2]; x[3]:= dirs.FOSO[i,1];
            sense:= -1;
            {compute curva backward with same initial condition}
            rk.ComputeIntegralCurve(x, sense);
          end;
        end;
      end;
    finally
      MainForm.EndWatch;
    end;
  end;
  dirs.FOSO:= nil;
end;

procedure TOdeNf.ComputeCurveByPoint3D(x: Vector3);
{Compute the solution to a Normalized implicit first order ODE starting
 from a point 'x' in the surface S on which F(u,v,w)=0. See Manual}
  var
   y: Vector3;
begin
  BGRAbm.PenStyle:= psSolid;
  RK.SetEnvironment(3, rad, cExTolFO, @DerivativeFrom3D, @CheckRange3D,
         @RecordCV3Node, nil, nil, @AdjustToSurf, @Shift);
  y:= x;
  try
    MainForm.BeginWatch;
    sense:= 1;

    {compute curve forward}
    RK.ComputeIntegralCurve(y, sense);
    if sense <> 0 then begin
      sense:= -1;
      y:= x;
      {compute curve backward}
      RK.ComputeIntegralCurve(y, sense);
    end;
  finally
    MainForm.EndWatch;
  end;
end;


procedure TOdeNf.ComputeVGCurve;
{compute criminant and discriminant}
begin
  cvg.last:= -1;
  aux:= 0;
  auxInt:= 0;
  MainForm.BeginWatch;
  try
    {find and store curve}
    ImplicitCurveDescent.ComputeCurve(ecVG, ecVGGrad, 3, @CheckRange3D,
         @RecordVGNode, cTolFO, mi, ma, 43, rad);
    DrawCurves3D;
    MainForm.update;
  finally
    MainForm.EndWatch;
  end;
end;   

constructor TOdeNf.Create;
var
  i: Integer;
begin
  inherited Create;
  SetLength(ecVG,2);
  SetLength(ecVGGrad,2);
  SetLength(cv3.ac, 20);
  cv3.last:= -1;;
  for i:= 0 to high(cv3.ac) do begin
    SetLength(cv3.ac[i].c, 50);
    cv3.ac[i].last:= -1;
  end;
  SetLength(cvg.ac,5);
  cvg.last:= -1;
  for i:= 0 to high(cvg.ac) do begin
    SetLength(cvg.ac[i].c, 20);
    cvg.ac[i].last:= -1;
  end;
end; 

function TOdeNf.Derivative(var deriv: array of double; here:
    array of double): Boolean;
{Gives the @Derivative of (u,v,w) along the solution curve}
  var
    i: Integer;
begin
  result:= Direction(deriv, here);
  if result and (sense= -1) then for i:= 0 to 2 do
    deriv[i]:= - deriv[i];
end;  

function TOdeNf.DerivativeFrom3D(var deriv: array of double; here:
    array of double): Boolean;
{Gives the @Derivative of (u,v,w) along the solution curve, when this curve
 is being drawn primarily as a curve on the surfade F(u,v,w)=0: see Manual}
  var
    i: Integer;
begin
  result := DirectionFrom3D(deriv,here);
  if result and (sense= -1) then for i:= 0 to 2 do
    deriv[i]:= - deriv[i];
end;

destructor TOdeNf.Destroy;
  var
    i: Integer;
begin
  if High(cv3.ac)>=0 then begin
    for i:= 0 to High(cv3.ac) do
      cv3.ac[i].c:= nil;
    cv3.ac:= nil;
  end;
  if High(cvg.ac)>=0 then begin
    for i:= 0 to High(cvg.ac) do
      cvg.ac[i].c:= nil;
    cvg.ac:= nil;
  end;
  ecVG:= nil;
  ecVGGrad:= nil;
  inherited Destroy;
end; 

function TOdeNf.Direction(var vel: array of double ;
       x: array of double): Boolean;
{Gives the @Derivative of (u,v,w) along a solution curve of a PArclenFO ODE as a unit vector.
 Let g=(g1,g2,g3) be the gradient of 'F' at the
 point 'x'. Since we can assume that cos(w)= u' and sin(w)= v', we will have
 from F(u,v,w)=0 that
    <g,(u',v',w')> = g1*cos(w)+g2*sin(w)+ g3*w'=0.
 Hence,
    w'= -(g1*cos(w)+g2*sin(w))/g3,
 which is the formula implemented below.}
  var
    co,si: Double;
    gr: Vector3;
begin
  gr:= SFGrad(x);
  sincos(x[2],si,co);
  vel[2]:= -(gr[1]*co + gr[2]*si);
  result:= abs(vel[2]) < 1.0e12*abs(gr[3]);
  if result then begin
    vel[0]:= co;
    vel[1]:= si;
    vel[2]:= vel[2]/gr[3];
  end;
end;  

function TOdeNf.DirectionFrom3D(var vel: array of double;
       x: array of double): Boolean;
{Almost equal to 'Direction'. There is an important difference
 however. When gr[3]=0, then this function results in True if
    gr[1]*co + gr[2]*si <> 0,
 while 'DirectionFrom3D' results in False when gr[3]=0 and
    gr[1]*co + gr[2]*si <> 0.
 So, this function allows the curve to cross the criminant (that is the
 curve where gr[3]=0), while 'Direction3Normalized' avoids it}
  var
    co,si: Double;
    gr: Vector3;
begin
  gr:= SFGrad(x);
  sincos(x[2],si,co);
  vel[2]:= -(gr[1]*co + gr[2]*si);
  vel[0]:= co*gr[3];
  vel[1]:= si*gr[3];
  result:= EuclNorm(vel) > 1.0e-25;
end;

procedure TOdeNf.DrawCurves2D(w: Integer);
var
  i, j: Integer;
  cc: ArrayOfTPointF;
begin
  inherited;
  try
    MainForm.BeginWatch;
    if (cv3.last >= 0) then
      for i:= 0 to cv3.last do
        if cv3.ac[i].last > 0 then begin
          SetLength(cc, cv3.ac[i].last+1);
          for j:= 0 to cv3.ac[i].last do
            FromParamsToDraw(cv3.ac[i].c[j], cc[j]);
          BGRAbm.DrawPolyLineAntialias(cc, RGBToBGRAColor(cv3.ac[i].color), w);
        end;
    if cvg.last >= 0 then
      for i:= 0 to cvg.last do
        if cvg.ac[i].last > 0 then begin
          SetLength(cc, cvg.ac[i].last+1);
          for j:= 0 to cvg.ac[i].last do
            FromParamsToDraw(cvg.ac[i].c[j], cc[j]);
          BGRAbm.DrawPolyLineAntialias(cc, RGBToBGRAColor(cvg.ac[i].color), 2*w);
        end;
    if ( High(cvAux.ac) >=0 ) and (cvAux.last >=0) then begin
      with BGRAbm do begin
        CustomPenStyle:= BGRAPenStyle(4,4);
        for i:= 0 to cvAux.last do
          if cvAux.ac[i].last > 0 then begin
            SetLength(cc, cvAux.ac[i].last+1);
            for j:= 0 to cvAux.ac[i].last do
              FromParamsToDraw(cvAux.ac[i].c[j], cc[j]);
            DrawPolyLineAntialias(cc, BGRABlack, 1);
            cc:= nil;
          end;
          CustomPenStyle:= SolidPenStyle;
        end;
    end;
  finally
    MainForm.EndWatch;
    cc:= nil;
  end;
end;

procedure TOdeNf.DrawCurves2DOnBGRAbm(bm: TBGRABitmap; r: TRect);
var
  i, j, w: Integer;
  pd1p, pd2p, pd3p, pd4p: Double;
  cc: ArrayOfTPointF;
begin
  pd1p:= pd1;
  pd2p:= pd2;
  pd3p:= pd3;
  pd4p:= pd4;
  pd1:= (r.right-r.left)/edges[1];
  pd2:= r.left - pd1*mi[1];
  pd3:= (r.top - r.bottom)/edges[2];
  pd4:= r.top - pd3*ma[2];
  bm.PenStyle:= psSolid;
  w:= 1 + (r.bottom - r.top) div 500;
  try
    if (High(cv2.ac) >=0 ) and (cv2.last >=0) then
      for i:= 0 to cv2.last do
        if cv2.ac[i].last > 0 then begin
          SetLength(cc, cv2.ac[i].last+1);
          for j:= 0 to cv2.ac[i].last do
            FromParamsToDraw(cv2.ac[i].c[j], cc[j]);
          bm.DrawPolyLineAntialias(cc, cv2.ac[i].color, w);
        end;
    if  (High(cv3.ac) >=0 ) and  (cv3.last >= 0) then
      for i:= 0 to cv3.last do
        if cv3.ac[i].last > 0 then begin
          SetLength(cc, cv3.ac[i].last+1);
          for j:= 0 to cv3.ac[i].last do
            FromParamsToDraw(cv3.ac[i].c[j], cc[j]);
          bm.DrawPolyLineAntialias(cc, RGBToBGRAColor(cv3.ac[i].color), w);
        end;
    w:= 1 + (r.bottom - r.top) div 400;
    if (High(cvg.ac) >=0 ) and (cvg.last >=0) then
      for i:= 0 to cvg.last do
        if cvg.ac[i].last > 0 then begin
          SetLength(cc, cvg.ac[i].last+1);
          for j:= 0 to cvg.ac[i].last do
            FromParamsToDraw(cvg.ac[i].c[j], cc[j]);
          bm.DrawPolyLineAntialias(cc, RGBToBGRAColor(cvg.ac[i].color), 1.5*w);
        end;
    if ( High(cvAux.ac) >=0 ) and (cvAux.last >=0) then begin
      with bm do begin
        CustomPenStyle:= BGRAPenStyle(2*w, 2*w);
        for i:= 0 to cvAux.last do
          if cvAux.ac[i].last > 0 then begin
            SetLength(cc, cvAux.ac[i].last+1);
            for j:= 0 to cvAux.ac[i].last do
              FromParamsToDraw(cvAux.ac[i].c[j], cc[j]);
            DrawPolyLineAntialias(cc, BGRABlack, 1);
            cc:= nil;
          end;
          CustomPenStyle:= SolidPenStyle;
        end;
    end;
  finally
    pd1:= pd1p;
    pd2:= pd2p;
    pd3:= pd3p;
    pd4:= pd4p;
    cc:= nil;
  end;
end;

procedure TOdeNf.DrawCurves3D;
var
  i, j: Integer;
  x, y: Vector3;
begin
  if not MainForm.alreadyDrawnSurface then
    exit;
  try
    mainForm.BeginWatch;
    if (cv3.last >= 0) then
      for i:= 0 to cv3.last do
        if cv3.ac[i].last > 0 then begin
          y:= cv3.ac[i].c[0];
          for j:= 1 to cv3.ac[i].last do begin
            x:= y;
            y:= cv3.ac[i].c[j];
            surface.DrawSegment3D(surfBuffer, kTolFactor, x,y, cv3.ac[i].color);
          end;
        end;
    if cvg.last >= 0 then
      for i:= 0 to cvg.last do
      if cvg.ac[i].last > 0 then begin
        y:= cvg.ac[i].c[0];
        for j:= 1 to cvg.ac[i].last do begin
          x:= y;
          y:= cvg.ac[i].c[j];
          surface.DrawSegment3D(surfBuffer, kTolFactor, x,y, cGVColor32);
        end;
      end;
  finally
    mainForm.EndWatch;
  end;
end;


function TOdeNf.DropLastCurve3D: Boolean;
begin
  result:= (high(cv3.ac) >=0) and  (cv3.last >=0);
  if result then begin
    cv3.ac[cv3.last].last:= -1;
    cv3.ac[cv3.last].c:= nil;
    dec(cv3.last);
  end;
end;



procedure TOdeNf.GetInitDirections(var dirs: directions; var dir: Vector2; par: Vector2);
{get initial directions for a Normalized implicit ODE of first order}
  const
    cGrid= 401;
    cAbsoluteTol= 1.0e-12;
  var
    t, r, preval, curval, minAdvance, step: Double;
    i: Integer;
    s1, s2: TValueSign;

  procedure AddAPoint(s: Double);
  begin
    inc(dirs.last);
    if High(dirs.FOSO) > dirs.last then
      SetLength(dirs.FOSO, dirs.last+10);
    dirs.FOSO[dirs.last,1]:= s;
  end;

  procedure RegulaFalsi(var Result: Double; a, b, fa, fb: Double);
  {It is assumed that fa=theF(foc+a*dir), fb=theFfoc+b*dir), sign(fa)
  <> sign(fb),a < b. Gives a value of  t, a<= t <= b, such that
  |theF(foc+t*v)| < verySmallValOfF.}
  var
    s, ft, a1, b1, fa1, fb1: Double;
  begin
    a1 := a;
    b1 := b;
    fa1 := fa;
    fb1 := fb;
    s := a;
    while (abs(b1 - a1) > minAdvance) do begin
      s := (a1 * fb1 - b1 * fa1) / (fb1 - fa1);
      ft:= theF1.ff([par[1],par[2],s]);
      if abs(ft) < cAbsoluteTol then
        break;
      if ft * fa1 > 0 then begin
        a1 := s;
        fa1 := ft;
      end
      else begin
        b1 := s;
        fb1 := ft;
      end;
    end;
    Result := s;
  end;

begin
  t:= mi[3];
  step:= (ma[3]-mi[3])/cGrid;
  minAdvance:= step/1000;
  curVal:= theF1.ff([par[1],par[2],t]);
  if abs(curVal) < cAbsoluteTol then
    AddAPoint(t);

  for i:= 1 to cGrid do begin
    t+= step;
    preVal:= curVal;
    curVal:= theF1.ff([par[1],par[2],t]);
    if abs(curVal) < cAbsoluteTol then begin
      AddAPoint(t);
      continue;
    end;
    s1:= sign(curval);
    s2:= sign(preval);
    if s1 <> s2 then begin
      RegulaFalsi(r, t-step, t, preVal, curVal);
      AddAPoint(r);
    end;
  end;
end;

procedure TOdeNf.Gradient(var f: array of Double; x: array of double);
begin
  f[0]:= theF1.firstD([x[0], x[1], x[2], 1, 0, 0]).dfa0;
  f[1]:= theF1.firstD([x[0], x[1], x[2], 0, 1, 0]).dfa0;
  f[2]:= theF1.firstD([x[0], x[1], x[2], 0, 0, 1]).dfa0;
end;

procedure TOdeNf.InitCurves(solutions, vgCurve, auxCurves: Boolean);
{erase curves data, or better take them as inexistent by putting
 its last index as -1}
var
  i: Integer;
begin
  inherited;
  if solutions then begin
    cv3.last:= -1;
    if high(cv3.ac)>=0 then
      for i:= 0 to high(cv3.ac) do
        cv3.ac[i].last:= -1;
  end;
  if vgCurve then
    for i:= 0 to high(cvg.ac) do begin
      SetLength(cvg.ac[i].c, 50);
      cvg.ac[i].last:= -1;
  end;
end; 

procedure TOdeNf.Initiate(mi0, ma0: Vector3);
begin
  inherited Initiate(mi0, ma0);
  ecVG[0]:= @vg1;
  ecVGGrad[0]:= @gradient;
  ecVG[1]:= @VertGrad;
  ecVGGrad[1]:= nil;
  surfBuffer:= MainForm.surfBuffer;
end;

procedure TOdeNf.RecordCV3Node(x: array of double; key: Byte);
var
  i,ind: Integer;
begin
  if key = 1 then begin
    inc(cv3.last);
    if high(cv3.ac) < cv3.last then begin
      SetLength(cv3.ac, cv3.last+10);
      for i:= 1 to 9 do
        cv3.ac[cv3.last+i].last:= -1;
    end;
    cv3.ac[cv3.last].last:= -1;
    cv3.ac[cv3.last].color:= cColor;
  end;
  inc(cv3.ac[cv3.last].last);
  if high(cv3.ac[cv3.last].c) < cv3.ac[cv3.last].last then
    SetLength(cv3.ac[cv3.last].c, cv3.ac[cv3.last].last+100);
  ind:= cv3.ac[cv3.last].last;
  for i:= 1 to 3 do
    cv3.ac[cv3.last].c[ind,i]:= x[i-1];
end;

procedure TOdeNf.RecordVGNode(x: array of double; key: Byte);
{store the new node and draw the segment defined with the previous one if key=0}
   var
    i,ind: Integer;
begin
  if key = 1 then begin
    inc(cvg.last);
    if high(cvg.ac) < cvg.last then begin
      SetLength(cvg.ac, cvg.last+10);
      for i:= 1 to 9 do
        cvg.ac[cvg.last+i].last:= -1;
    end;
    cvg.ac[cvg.last].last:= -1;
    cvg.ac[cvg.last].color:= cGVColor32;
  end;
  inc(cvg.ac[cvg.last].last);
  if high(cvg.ac[cvg.last].c) < cvg.ac[cvg.last].last then
    SetLength(cvg.ac[cvg.last].c, cvg.ac[cvg.last].last+100);
  ind:= cvg.ac[cvg.last].last;
  for i:= 1 to 3 do
    cvg.ac[cvg.last].c[ind,i]:= x[i-1];
end;

function TOdeNf.SF(const x: Vector3): Double;
begin
  result:= TheF1.ff([x[1], x[2], x[3]]);
end;

function TOdeNf.SFgrad(const x: vector3): Vector3;
begin
  result[1]:= theF1.firstD([x[1], x[2], x[3], 1,0,0]).dfa0;
  result[2]:= theF1.firstD([x[1], x[2], x[3], 0,1,0]).dfa0;
  result[3]:= theF1.firstD([x[1], x[2], x[3], 0,0,1]).dfa0;
end;

function TOdeNf.Shift(x1,x2: array of double): Double;
 var
    i: Integer;
begin
  result:= 0;
  for i:= 0 to 1 do
    result:= result + sqr(x2[i]-x1[i]);
  result:= sqrt(result);
end;     


procedure TOdeNf.VertGrad(var f: Double; x: array of double);
{partial @Derivative of the function that defines a first order Normalized
 ODE with respect to 'w'}
begin
  f:= theF1.firstD([x[0], x[1], x[2], 0, 0, 1]).dfa0;
end;

procedure TOdeNf.vg1(var f: Double; x: array of double);
begin
  f:= TheF1.ff([x[0], x[1], x[2]]);
end;

{------------------------------  End of TOdeNf ------------------------------------}

{------------------------------  TOdeNs begins ------------------------------------}


procedure TOdeNs.AdjustToSurf(var x: array of double);
{similar to AdjustToSurfPArcLenFO}
var
  gr: Vector4;
  a, b, cut: Double;
  i, count: Integer;
  rslt: Boolean;
begin
  count:=0;
  a:= theF1.ff(x);
  gradient(gr, x);
  cut:= 1.0e-12 * radTotal*EuclNorm(gr);
  rslt:= ( a < cut );
  if not rslt then begin
    while (not rslt) and (count < 20) do begin
      inc(count);
      gradient(gr, x);
      b:= EuclEnergy(gr);
      if b > 1.0e-24 then begin
        b:= -a/b;
        for i:= 0 to 3 do
          x[i]:= x[i]+b*gr[i+1];
      end;
      a:= theF1.ff([x[0],x[1],x[2],x[3]]); 
      rslt:= ( abs(a) < cut );
    end;
    if not rslt then begin
      b:= sqrt(b);
      if abs(a) > 1.0e-12*b then
        x[0]:= kElError;
    end;
  end;
end; 


function TOdeNs.CheckCameBack(firstX, prevX, curX: array of double): Boolean;
var
  a, b, d: Double;
begin
  result:= EuclDistanceToSegment(d, 1, firstX, prevX, curX);
  if result then begin
    a:= abs(2*firstX[3]- prevX[3]- curX[3]);
    b:= 0.02*abs(curX[3]);
    result := (d < 5*rad) and (a < b);
  end;
end;

function TOdeNs.CheckRange(nextX: array of double): Boolean;
{Check whether 'nextX' is in the domain. Here the angle 'w' is limited by 48*pi
 and the curvature by 10^7}
  var
    i: Integer;
begin
  result:= True;
  for i:= 0 to 1 do
    result:= result and (mi[i+1] < nextX[i]) and (nextX[i] < ma[i+1]);
  result:= result and (abs(nextX[2]) < 500) and (abs(nextX[3]) < 1.0e7);
end; 

procedure TOdeNs.ComputeCurve(par, dir: Vector2);
{compute solution to a arc-length implicit ODE of second order}
 var
    dirs: directions;
    i: Integer;
    x: Vector4;
begin
  SetLength(dirs.FOSO,6);
  dirs.last:= -1;
  GetInitDirections(dirs, dir, par); {find admissible initial second order conditions}
  if dirs.last >=0 then begin
    RK.SetEnvironment(4, rad, cExTolSO, @Derivative, @CheckRange,
          @RecordNode, nil, @CheckCameBack, @AdjustToSurf,
          @Shift);
    for i:= 0 to dirs.last do begin
      x[1]:= par[1];
      x[2]:= par[2];

      {the initial direction and curvature}
      x[3]:= dirs.FOSO[i,1];
      x[4]:= dirs.FOSO[i,2];

      AdjustToSurf(x); {make 'x' be almost exactly on the surface}
      if  CheckRange(x) then
        try
          MainForm.BeginWatch;
          sense:= 1;

          {compute curve forward. In second order cases we don't compute the curve
            backward: see Manual}
          RK.ComputeIntegralCurve(x, sense);
          DrawOneCurve2D(cv2.ac[cv2.last], 1);
          MainForm.EndWatch;
        Finally
          MainForm.EndWatch;
        end;
    end;
  end;
  dirs.FOSO:= nil;
end;

function TOdeNs.Derivative(var deriv: array of double; here:
    array of double): Boolean;
{gives the derivatives of the variables u,v,w,x along a curve solution
 of a Normalized implicit second order ODE}
  var
    i: Integer;
begin
  result := Direction(deriv, here);
  if result then begin
    if (sense= -1) then for i:= 0 to 3 do
      deriv[i]:= - deriv[i];
    result:= ( EuclNorm(deriv) > 1.0e-14 );
  end;
end;

function TOdeNs.Direction(var vel: array of double;
       x: array of double): Boolean;
{gives the derivatives of the variables u,v,w,x along a curve solution
 of an arc-length parametrized implicit second order ODE.}
  var
    i: Integer;
    co,si,nn: Double;
    gr: Vector4;
begin
  gradient(gr, x);
  sincos(x[2],si,co);
  vel[3]:= -(gr[1]*co + gr[2]*si+ gr[3]*x[3]);
  if gr[4] < 0 then begin  {this change of sign is intended to give
    continuity to the vector field that we are integrating when the
    criminant is crossed}
    vel[3]:= -vel[3];
    gr[4]:= -gr[4];
  end;
  vel[0]:= co*gr[4];
  vel[1]:= si*gr[4];
  vel[2]:= x[3]*gr[4];
  nn:= EuclNorm(vel);
  result:= nn > 1.0e-12;
  if result then
    for i:= 0 to 3 do
      vel[i]:= vel[i]/nn;
end; 


procedure TOdeNs.GetInitDirections(var dirs: directions; var dir: Vector2; par: Vector2);
{get admissible initial accelerations for an ODE of second order}
  const
    cSamples = 200;
  var
    i: Integer;
    phi, x, step, f1, f2: Double;

function RegulaFalsi: Double;
var
  x1, x2, v1, v2, val, y: Double;
  count: Integer;
begin
  x1:= x- step;
  x2:= x;
  v1:= f1;
  v2:= f2;
  count:= 0;
  repeat
    inc(count);
    y:= x1 + (x2-x1)*v1/(v1-v2);
    val:= theF1.ff([par[1], par[2], phi, y]);
    if val * v1 > 0 then begin
      x1:= y;
      v1:= val;
    end
    else begin
      x2:= y;
      v2:= val;
    end;
  until (count = 10) or (abs(val) < 1.0e-10);
  result:= y;
end;

begin
try
 {extract from 'dir' the angle with the u-axis it defines}
  if abs(dir[1])< 1.0e-14 then begin
    if dir[2] > 0 then
      phi:= cHalfPi
    else if dir[2] < 0 then
      phi:= 3*cHalfPi
    else  {bad initial direction}
      Exit;
  end
  else begin
    phi:= arctan2(dir[2],dir[1]);
    if phi < 0 then
      phi:= phi+ cTwoPi;
  end;
  dir[1]:= phi;
  step:= ma[3]/cSamples;
  x:= -ma[3] -step;
  f2:= TheF1.ff([par[1], par[2],phi, x]);
  for i:= 0 to 2*cSamples do begin
    {compute the signs of adjacent samples and if they differ, seek the zero}
    f1:= f2;
    x:= x + step;
    f2:= TheF1.ff([par[1], par[2],phi, x]);
    if  f1*f2 < 0 then begin
      inc(dirs.last);
      if High(dirs.FOSO) > dirs.last then
        SetLength(dirs.FOSO, dirs.last+10);
      dirs.FOSO[dirs.last,1]:= phi;
      {now we adjust it better}
      dirs.FOSO[dirs.last,2]:= RegulaFalsi;
    end;
  end;
  except on exception do
    MainForm.Inform(eBadPointEncountered, False);
  end;
end;  

procedure TOdeNs.Gradient(var f: array of Double; x: array of double);
begin
  f[0]:= theF1.firstD([x[0], x[1], x[2], x[3], 1, 0, 0, 0]).dfa0;
  f[1]:= theF1.firstD([x[0], x[1], x[2], x[3], 0, 1, 0, 0]).dfa0;
  f[2]:= theF1.firstD([x[0], x[1], x[2], x[3], 0, 0, 1, 0]).dfa0;
  f[3]:= theF1.firstD([x[0], x[1], x[2], x[3], 0, 0, 0, 1]).dfa0;
end;


    
function TOdeNs.Shift(x1,x2: array of double): Double;
{We desestimate the @Shift of the velocity}
  var
    i: Integer;
begin
  result:= 0;
  for i:= 0 to 2 do
    result:= result + sqr(x2[i]-x1[i]);
  result:= sqrt(result);
end;

end.           

     
