(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 11.2' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       158,          7]
NotebookDataLength[    193788,       3438]
NotebookOptionsPosition[    189570,       3363]
NotebookOutlinePosition[    189908,       3378]
CellTagsIndexPosition[    189865,       3375]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{

Cell[CellGroupData[{
Cell["\<\
It is supposed that some point P in the surface is taken as the origin, so \
that \[OpenCurlyQuote]luz\[CloseCurlyQuote] is the position vector of the \
point-like light source and \[OpenCurlyQuote]viewPoint\[CloseCurlyQuote] is \
that of the camera. Then the cosine of that angle is the argument passed \
below to the
function ArcCos.\
\>", "Subsubsection",
 CellChangeTimes->{{3.743775466206061*^9, 3.743775621423443*^9}, {
  3.743776619915036*^9, 
  3.7437766659940243`*^9}},ExpressionUUID->"6124fcac-9fe3-4bb3-969e-\
d6e33e9028fd"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"p", "=", 
  RowBox[{"ArcCos", "[", "0.8", "]"}]}]], "Input",
 CellChangeTimes->{{3.7435186035335283`*^9, 3.743518626917889*^9}, {
  3.743518704843423*^9, 3.743518743299752*^9}, {3.7435726190757227`*^9, 
  3.743572635158239*^9}, {3.7437438041987543`*^9, 3.74374382911452*^9}, {
  3.743749822094358*^9, 3.743749836162024*^9}, {3.743776032627306*^9, 
  3.743776045578672*^9}},ExpressionUUID->"998de7f7-42be-48cf-b17b-\
04593dd453ed"],

Cell[BoxData["0.6435011087932843`"], "Output",
 CellChangeTimes->{3.743518765493051*^9, 3.7435724450802383`*^9, 
  3.743572636227146*^9, 3.743593919404331*^9, 3.7435940257911043`*^9, 
  3.743595776113081*^9, 3.743597335048861*^9, 3.743597833839634*^9, 
  3.743597909833269*^9, 3.743598113435034*^9, 3.743743848485661*^9, 
  3.7437443237945232`*^9, 3.743747648420815*^9, 3.743749841766448*^9, 
  3.743750370477195*^9, 3.743772474563089*^9, 3.743776060958119*^9, 
  3.743777056959052*^9},ExpressionUUID->"f94f41d5-ae54-4bc6-aa2e-\
fb24b8dc4f50"]
}, Open  ]],

Cell["\<\
We  choose the third axis so that it is the bisector of the angle formed by \
those two vectors, so that the unit vectors from P to those vectors may be \
written as\
\>", "Text",
 CellChangeTimes->{{3.743776694097076*^9, 3.743776760245125*^9}, {
  3.7437772293670397`*^9, 
  3.743777271670474*^9}},ExpressionUUID->"06158f52-b38b-4b9c-bfb0-\
df3a41ba5a79"],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"light", " ", "=", " ", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"Cos", "[", 
       RowBox[{"p", "/", "2"}], "]"}], ",", " ", "0", ",", " ", 
      RowBox[{"Sin", "[", 
       RowBox[{"p", "/", "2"}], "]"}]}], "}"}]}], ";"}], " "}], "\n", 
 RowBox[{
  RowBox[{"viewpoint", " ", "=", " ", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", 
      RowBox[{"Cos", "[", 
       RowBox[{"p", "/", "2"}], "]"}]}], ",", " ", "0", ",", " ", 
     RowBox[{"Sin", "[", 
      RowBox[{"p", "/", "2"}], "]"}]}], "}"}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.743596900956101*^9, 3.743596922369762*^9}, {
  3.743597745452804*^9, 3.743597748578072*^9}, {3.743598098974394*^9, 
  3.743598106791038*^9}, {3.743776998471678*^9, 3.743776999201364*^9}, {
  3.7437774269226294`*^9, 
  3.7437774281676273`*^9}},ExpressionUUID->"fe36f75e-f595-447f-9120-\
7821b7b350fa"],

Cell["\<\
Now,  we compute the intensity of two lights that reach the camera from P. In \
that formula, \[OpenCurlyQuote]LN\[CloseCurlyQuote] is the cosine of the \
angle between unit \[OpenCurlyQuote]normal\[CloseCurlyQuote] to the surface \
at point P and \[OpenCurlyQuote]light\[CloseCurlyQuote]. \[OpenCurlyQuote]VN\
\[CloseCurlyQuote] is the cosine of the angle from \[OpenCurlyQuote]viewpoint\
\[CloseCurlyQuote] to \[OpenCurlyQuote]normal\[CloseCurlyQuote],  \
\[OpenCurlyQuote]refl\[CloseCurlyQuote] is the vector symmetric to \
\[OpenCurlyQuote]light \[OpenCurlyQuote]  with respect to \
\[OpenCurlyQuote]normal\[CloseCurlyQuote].
If  \[OpenCurlyQuote]VN >= 0\[CloseCurlyQuote]   we keep  the normal, and say \
that from \[OpenCurlyQuote]viewpoint\[CloseCurlyQuote] we are seeing the \
obverse side of the surface at P. Otherwise we change sign to \
\[OpenCurlyQuote]normal\[CloseCurlyQuote] and to \[OpenCurlyQuote]VN\
\[CloseCurlyQuote]. Thus, \[OpenCurlyQuote]normal\[CloseCurlyQuote] and \
\[OpenCurlyQuote]viewpoint\[CloseCurlyQuote] point
to the same side of the surface at P.  Therefore, if  \[OpenCurlyQuote]light\
\[CloseCurlyQuote]  does not point to the same side, that is if  \
\[OpenCurlyQuote]LN < 0\[CloseCurlyQuote],  then  the light coming from \
\[OpenCurlyQuote]light\[CloseCurlyQuote] to P cannot be reflected to \
\[OpenCurlyQuote]viewpoint\[CloseCurlyQuote] so that we put \
\[OpenCurlyQuote]LN = 0.\[CloseCurlyQuote]
The ray coming from \[OpenCurlyQuote]light\[CloseCurlyQuote] to P reflects \
specularly  on the surface in the direction given by
\[OpenCurlyQuote]refl= 2 LN normal - light.\[CloseCurlyQuote] If  then \
\[OpenCurlyQuote]LN = 0\[CloseCurlyQuote], we have that \[OpenCurlyQuote]refl \
. viewpoint = - LV < 0\[CloseCurlyQuote].  Thus,  the specular
 part of the light coming from \[OpenCurlyQuote]light\[CloseCurlyQuote] \
reflected cannot reach \[OpenCurlyQuote]viewpoint\[CloseCurlyQuote] and we \
put \[OpenCurlyQuote]beta=0.$  If, on the contrary, \[OpenCurlyQuote]LN >0\
\[CloseCurlyQuote],  we  put \[OpenCurlyQuote]beta = refl.viewpoint\
\[CloseCurlyQuote]. Now,  the point-like light source  \[OpenCurlyQuote]light\
\[CloseCurlyQuote] scatters at P in all directions, with an intensity that \
depends on the angle between \[OpenCurlyQuote]normal\[CloseCurlyQuote] and \
\[OpenCurlyQuote]light\[CloseCurlyQuote]. We put finally the value of the \
light reaching the camera from \[OpenCurlyQuote]light\[CloseCurlyQuote] as \
the charged sum  0.3*LN + 0.7*beta. 
 
 Also we suppose that there is an ambient light that consists of a fixed \
quantity, 0.2, and a quantity that depends on the orientation of the normal \
with respect to the camera, say:  alpha= 0.2 + 0.8*VN.\
\>", "Text",
 CellChangeTimes->{{3.743596937518608*^9, 3.743596969062446*^9}, {
   3.743776776536574*^9, 3.7437768223106422`*^9}, {3.7437768538568087`*^9, 
   3.7437769878577967`*^9}, {3.7437771028176117`*^9, 3.743777151200585*^9}, {
   3.743777284360598*^9, 3.743777518925812*^9}, 3.7437776234864607`*^9, {
   3.743777654235736*^9, 3.743777768214739*^9}, {3.743777808490899*^9, 
   3.743777930044421*^9}, {3.743778096490137*^9, 
   3.743778884641416*^9}},ExpressionUUID->"dd688018-ffd5-40ce-b416-\
47467de11072"],

Cell[BoxData[
 RowBox[{
  RowBox[{"alphabeta", "[", 
   RowBox[{"s_", ",", "t_"}], "]"}], ":=", " ", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
     "LN", ",", " ", "normal", ",", " ", "VN", ",", " ", "alpha", ",", " ", 
      "beta", ",", " ", "refl"}], "}"}], ",", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"normal", " ", "=", " ", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{
         RowBox[{"Cos", "[", "s", "]"}], " ", 
         RowBox[{"Cos", "[", "t", "]"}]}], ",", " ", 
        RowBox[{
         RowBox[{"Sin", "[", "s", "]"}], " ", 
         RowBox[{"Cos", "[", "t", "]"}]}], ",", " ", 
        RowBox[{"Sin", "[", "t", "]"}]}], "}"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"VN", " ", "=", " ", 
      RowBox[{"viewPoint", ".", "normal"}]}], ";", "\n", "    ", 
     RowBox[{"If", "[", 
      RowBox[{
       RowBox[{"VN", " ", "<", " ", "0"}], ",", " ", 
       RowBox[{
        RowBox[{"VN", " ", "=", " ", 
         RowBox[{"-", "VN"}]}], ";", " ", 
        RowBox[{"normal", " ", "=", " ", 
         RowBox[{"-", "normal"}]}]}]}], "]"}], ";", "\[IndentingNewLine]", 
     RowBox[{"LN", "=", " ", 
      RowBox[{"light", ".", "normal"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"If", "[", 
      RowBox[{
       RowBox[{"LN", " ", "<", " ", "0"}], ",", " ", 
       RowBox[{"LN", " ", "=", "0"}]}], "]"}], ";", "\[IndentingNewLine]", 
     RowBox[{"refl", "=", " ", 
      RowBox[{
       RowBox[{"2", " ", "LN", " ", "normal"}], " ", "-", " ", "light"}]}], 
     ";", "\[IndentingNewLine]", 
     RowBox[{"beta", "=", " ", 
      RowBox[{"refl", ".", "viewPoint"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"If", "[", 
      RowBox[{
       RowBox[{"beta", " ", "<", " ", "0"}], ",", " ", 
       RowBox[{"beta", "=", "0"}], ",", " ", 
       RowBox[{"beta", "=", " ", 
        RowBox[{
         RowBox[{"0.3", " ", "*", "LN"}], " ", "+", " ", 
         RowBox[{"0.7", "*", "beta"}]}]}]}], "]"}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"alpha", "=", " ", 
      RowBox[{"0.2", " ", "+", " ", 
       RowBox[{"0.8", " ", "VN"}]}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"{", 
      RowBox[{"alpha", ",", "beta"}], "}"}]}]}], "]"}], " "}]], "Input",
 CellChangeTimes->{{3.743596545548347*^9, 3.743596564088546*^9}, {
   3.743596977937416*^9, 3.7435970601302347`*^9}, {3.743597090431056*^9, 
   3.7435971513620577`*^9}, {3.743597181941595*^9, 3.743597234646997*^9}, {
   3.743597287883874*^9, 3.743597297343981*^9}, 3.7435974526894083`*^9, {
   3.7435974978016567`*^9, 3.743597571589223*^9}, {3.7435976334782257`*^9, 
   3.7435976830862293`*^9}, {3.743597755928776*^9, 3.743597765555675*^9}, {
   3.743597823631186*^9, 3.743597825255391*^9}, {3.743598060336822*^9, 
   3.743598079887854*^9}, {3.7437443039279222`*^9, 3.743744317809619*^9}, {
   3.743747628650352*^9, 3.743747641306246*^9}, {3.743750301481143*^9, 
   3.743750349652969*^9}, {3.743772449583589*^9, 3.743772463755272*^9}, {
   3.743772788022957*^9, 3.743772792150462*^9}, {3.743777012704438*^9, 
   3.743777043850164*^9}},ExpressionUUID->"7b1aafaf-13eb-4a84-a5af-\
b1756ba86932"],

Cell["\<\
We make a plot of the sum   \[OpenCurlyQuote]alpha + beta\[CloseCurlyQuote] \
depending on the  variables  s and t that set the direction of  the normal \
with respect to the fixed pair \[OpenCurlyQuote]light\[CloseCurlyQuote] and \
\[OpenCurlyQuote]viewpoint\[CloseCurlyQuote]. \
\>", "Text",
 CellChangeTimes->{{3.743778905645794*^9, 3.743779020251191*^9}, 
   3.743779051413787*^9},ExpressionUUID->"25a1a01e-a849-4602-b1c8-\
0064263790aa"],

Cell[BoxData[
 RowBox[{"Plot3D", "[", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"alphabeta", "[", 
      RowBox[{"s", ",", "t"}], "]"}], "[", 
     RowBox[{"[", "1", "]"}], "]"}], "+", 
    RowBox[{
     RowBox[{"alphabeta", "[", 
      RowBox[{"s", ",", "t"}], "]"}], "[", 
     RowBox[{"[", "2", "]"}], "]"}]}], ",", " ", 
   RowBox[{"{", 
    RowBox[{"s", ",", "0", ",", " ", 
     RowBox[{"2", " ", "Pi"}]}], "}"}], ",", " ", 
   RowBox[{"{", 
    RowBox[{"t", ",", 
     RowBox[{
      RowBox[{"-", "Pi"}], "/", "2"}], ",", 
     RowBox[{"Pi", "/", "2"}]}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.743514168490238*^9, 3.743514239600815*^9}, {
   3.743516919172364*^9, 3.743516970756659*^9}, {3.7435187949323072`*^9, 
   3.743518796178865*^9}, 3.743518827927651*^9, {3.743598154121831*^9, 
   3.7435981900086737`*^9}},ExpressionUUID->"ff27416d-c794-4e52-9f9f-\
5600ad1bcd9b"],

Cell[BoxData[
 Graphics3DBox[GraphicsComplex3DBox[CompressedData["
1:eJyMfXVUlcv3PraCiV3Y3aAYqINdYFzba10UVLC7xUbEVmxFxe68BvraCmJe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   "], {{
     {RGBColor[0.880722, 0.611041, 0.142051], EdgeForm[None], Specularity[
       GrayLevel[1], 3], 
      StyleBox[GraphicsGroup3DBox[
        TagBox[{Polygon3DBox[CompressedData["
1:eJw1nHf8UNMbx+8e36iMIjtUaItQVhmhMiI/I1RWGUVWEUJWIXvP7L3Kjops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           "]], Polygon3DBox[CompressedData["
1:eJw1nAnYVtPXxs/zPGd8IxqUBmlAiqKJEomUDJUyRUoTMkUqElKpVCRFyFAR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           "]], 
          Polygon3DBox[CompressedData["
1:eJwt1wv8z9Udx/Ff/r+/31/SxphcptnWFnLv75Z7EmHK/ZJLbtHDZQrlLiot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           "]], 
          Polygon3DBox[{{1827, 1110, 1289, 2407, 1704, 1705}, {1711, 1710, 
           2411, 1294, 1295, 2412}, {1701, 1700, 2403, 1285, 1286, 2404}}]},
         Annotation[#, "Charting`Private`Tag$19454#1"]& ]],
       Lighting->{{"Ambient", 
          RGBColor[0.30100577, 0.22414668499999998`, 0.090484535]}, {
         "Directional", 
          RGBColor[0.2642166, 0.18331229999999998`, 0.04261530000000001], 
          ImageScaled[{0, 2, 2}]}, {"Directional", 
          RGBColor[0.2642166, 0.18331229999999998`, 0.04261530000000001], 
          ImageScaled[{2, 2, 2}]}, {"Directional", 
          RGBColor[0.2642166, 0.18331229999999998`, 0.04261530000000001], 
          ImageScaled[{2, 0, 2}]}}]}, {}, {}, {}, {}}, {
     {GrayLevel[0], Line3DBox[CompressedData["
1:eJwl0j1PU2EUB/BTtIhvqJNYpaUw6yeASTeYTFw0MnQSDZgQB5mILjKgcdEG
tDaEAFEmTVU0TtIEWKkWtdiXj8AHMPF3Q9Knv/85N7nnuc+9+cL961OpiHhq
bXRFJHmc54Qb3RFn1QXmeIdXOMkRPuAYH/EWn/Eu37BszcgVzvEbi6xylfVj
EY/N+CjnjkQ8lxvmVtW1noi8XlGvpVfT+8Qt9fmjET/UffzJC6wzwz1e5C9e
4m/28w+zbCSzuM8B/mWeTQ6yxSG2eTUdsWBex9yO+jO31W0eeIbd5B72+EKv
qbep/i7XXavIO1xJenzJr3zCD3xvPZRfcYLzvMlZjnKaw7zHy7ztPIbNyCbn
J/+TM2Z94Rm9dZbU7+ynV/2Wp7nGU1zhSS7zBJd4nGX2JO+LXkeUaGS8ZjrZ
Hx1HLNAviv6MimsuLqYOv5f/2wdIuQ==
       "]]}, 
     {GrayLevel[0.2], 
      Line3DBox[{1621, 2268, 2269, 2267, 1060, 1620, 1904, 1905, 2537, 1778, 
       1622, 2322, 2323, 2614, 1779, 1623, 2386, 2436, 2021, 1780, 1624, 2378,
        2437, 1906, 1781, 1625, 2388, 2438, 2022, 1782, 1626, 1425, 2112, 
       1783, 1627, 2113, 2114, 1063, 1628, 2439, 1784, 1629, 1861, 2440, 1785,
        1630, 2441, 1786, 1631, 2442, 1787, 1632, 2426, 2443, 2115, 1788, 
       1633, 1217, 2168, 2243, 1789, 2244}], Line3DBox[CompressedData["
1:eJwVzjlSAlEUheELDU1DYqAswEgjrcIBzS0NyLDKxCFAIcTAAcvAFai4Bylz
DcXYYSWoixC/Dv76zz193+s33zlt9wsR0UYnjXjNIpaqEQu1iLG8XYlI+EaX
6f7svJif8YZF3VR3jF07t5g4EzjRtcx3uLZb0J3zBQa4xJlulh9RlEtcdWeD
V/Ctm+OEy7yKNfyY67zs7hn7qbyOZn6vbwc8Qkne5x2+t1vmrncN5VTuyXvy
AwbmCt7tf+ALn7jSbfETsnyHa/65wZv41R3l7+ZD/gd/kCJk
       "]], 
      Line3DBox[{1640, 2352, 2488, 1639, 1864, 2353, 2528, 2154, 2155, 1865, 
       2169, 2580, 2117, 2118, 1914, 2389, 2449, 2056, 1802, 2171, 2172, 2588,
        1915, 2176, 2070, 2174, 2071, 1916, 2173, 2175, 1866, 2027, 1803, 
       1917, 2539, 1908, 1804, 1641, 1909, 2354, 2538, 2245, 2312, 2313, 1867,
        2228, 2282, 2529, 2283, 2143, 2119, 2120, 1076, 1642, 2286, 2450, 
       1805, 2287, 2288, 1161, 1920, 2072, 2073, 1919, 2540, 1868, 2317, 1806,
        2213, 2594, 1913, 1807, 1643}], 
      Line3DBox[{1645, 2355, 2489, 1644, 2284, 2356, 2607, 2285, 1869, 1921, 
       2427, 2530, 2145, 2144, 2177, 1240, 2057, 1808, 1923, 2079, 2577, 1870,
        2180, 2178, 2179, 1924, 2541, 1918, 2324, 1809, 1646, 2451, 1810, 
       1647, 2357, 2490, 1648, 2227, 2358, 2596, 2237, 2250, 1871, 2251, 2428,
        2531, 2147, 2146, 2327, 2121, 2122, 2328, 2329, 2156, 2157, 2238, 
       2585, 1872, 2028, 1811, 2271, 2603, 2270, 1812, 1649, 2452, 1813, 
       1650}], Line3DBox[{1652, 1929, 1930, 2559, 1651, 1931, 1932, 2543, 
       1653, 1873, 1922, 1169, 1654, 2391, 2571, 2058, 1933, 2107, 2106, 2123,
        2080, 2181, 2589, 1874, 1925, 1875, 2291, 1655, 2326, 2608, 2289, 
       2325, 2290, 1656, 2453, 1814, 1657, 2359, 2491, 1658, 2292, 2293, 2609,
        1659, 2294, 1876, 1926, 1877, 2578, 2260, 2081, 2215, 2261, 2263, 
       2262, 2182, 2254, 1936, 2253, 1937, 2252, 1660, 2542, 1927, 2029, 1928,
        1661, 2454, 1815, 1662, 2455, 1816, 1663}], 
      Line3DBox[{1665, 1938, 1939, 2560, 1664, 1940, 1941, 2545, 1666, 1942, 
       2492, 2393, 1667, 2380, 2544, 1934, 1878, 2082, 2216, 2124, 2295, 2392,
        2581, 2030, 1935, 2345, 1668, 2349, 2618, 2342, 2348, 2343, 1669, 
       2546, 1944, 1945, 1670, 1946, 1947, 2547, 1671, 2334, 2335, 2617, 1672,
        2336, 1948, 2217, 2493, 2394, 2272, 2218, 2219, 2083, 1879, 2059, 
       1880, 2316, 1673, 2332, 2432, 2615, 2331, 2296, 1674, 2456, 1817, 1675,
        2457, 1818, 1676}], 
      Line3DBox[{1678, 1950, 1951, 2561, 1677, 1952, 2494, 2395, 1679, 1953, 
       2495, 2396, 1680, 2183, 2184, 2590, 1943, 2085, 2084, 1681, 2397, 2601,
        2255, 1954, 1682, 2360, 2496, 1683, 2548, 1956, 1957, 1684, 1958, 
       1959, 2562, 1685, 1960, 2497, 2398, 1686, 1961, 2256, 2498, 2315, 2330,
        2314, 1687, 2086, 2087, 1949, 2186, 2185, 1688, 2399, 2499, 1962, 
       1689, 2361, 2500, 1690, 2458, 1819, 1691}], 
      Line3DBox[{1695, 2459, 1820, 1693, 2460, 1821, 1697, 1965, 2221, 2595, 
       1822, 1966, 2090, 2091, 2461, 1968, 2190, 1823, 2031, 2402, 2462, 1824,
        1701, 2404, 2463, 1825, 2032, 2406, 2464, 1826, 2033, 1972, 2465, 
       1827, 1705, 2466, 1828, 1707, 1974, 2467, 1829, 1975, 2191, 2192, 1977,
        2074, 1830, 2034, 2410, 2468, 2222, 1831, 1711, 2412, 2469, 1832, 
       2035, 2414, 2470, 1833, 1714}], 
      Line3DBox[{1713, 1979, 2512, 2413, 1712, 1978, 2511, 2411, 1710, 1976, 
       2510, 2409, 1709, 2092, 2127, 2093, 1963, 2189, 2257, 2188, 1708, 2408,
        2509, 1973, 1706, 2508, 2364, 1704, 2407, 2507, 1971, 1703, 1970, 
       2506, 2405, 1702, 1969, 2505, 2403, 1700, 1967, 2504, 2401, 1699, 2125,
        2220, 2126, 1955, 2591, 2089, 2187, 2088, 1698, 2400, 2503, 1964, 
       1696, 2502, 2363, 1692, 2501, 2362, 1694}], 
      Line3DBox[{1716, 2365, 2513, 1715, 2337, 2471, 1834, 2338, 1980, 1981, 
       2592, 2193, 1835, 2194, 2094, 2095, 2472, 2037, 1836, 2036, 2415, 2473,
        1837, 1717, 2417, 2474, 1838, 2038, 2563, 1982, 1839, 2039, 1983, 
       2366, 2549, 1718, 2475, 1840, 1719, 2297, 2476, 1841, 2298, 1881, 2060,
        2075, 1842, 2266, 2264, 2265, 2477, 2223, 1843, 1720, 2435, 2478, 
       2344, 1844, 1721, 2479, 1845, 1722}], 
      Line3DBox[{1724, 2367, 2514, 1723, 2299, 2368, 2610, 2300, 2301, 1882, 
       1985, 2576, 2076, 2195, 2077, 1984, 2108, 2109, 2416, 2480, 2061, 1846,
        1725, 1183, 1986, 1847, 1726, 2550, 1987, 2318, 1848, 2040, 2564, 
       1988, 1849, 2041, 1989, 2369, 2551, 1727, 2370, 2515, 1728, 1990, 1991,
        2552, 1850, 2196, 1992, 2148, 2197, 2199, 2198, 1993, 2096, 2258, 
       2381, 2481, 1994, 2304, 2305, 2303, 2611, 2302, 2311, 1851, 1729, 2482,
        1852, 1730}], 
      Line3DBox[{1732, 2371, 2516, 1731, 1995, 1997, 2372, 2553, 1996, 2208, 
       1883, 2209, 2425, 2532, 2078, 2042, 2110, 2418, 2483, 2062, 1853, 2158,
        2159, 2586, 1884, 2001, 2275, 2276, 2274, 2604, 2273, 1854, 1733, 
       2484, 1855, 1734, 2373, 2517, 1735, 2306, 2374, 2612, 2307, 2308, 1885,
        2002, 2429, 2533, 2150, 2149, 2128, 2319, 1856, 2160, 2161, 2239, 
       2587, 1886, 2151, 2231, 2232, 2230, 2597, 2229, 2240, 1857, 1736, 2485,
        1858, 1737}], 
      Line3DBox[{1750, 2014, 2013, 2556, 1749, 1898, 2241, 1897, 2536, 2242, 
       1748, 2224, 2130, 2152, 2129, 2582, 1747, 2278, 2320, 2277, 2321, 1746,
        2279, 2605, 1896, 2003, 1895, 2097, 1745, 2555, 1894, 1893, 1744, 
       2602, 2010, 2009, 1743, 1859, 2486, 1742, 1892, 1891, 2535, 1741, 2162,
        1890, 2203, 1889, 2534, 2202, 1740, 2007, 2063, 2572, 2419, 1739, 
       2200, 2593, 2000, 2210, 1999, 2205, 2111, 2132, 2131, 2204, 2583, 1888,
        1998, 1887, 1738, 2554, 2005, 2004, 2006}], 
      Line3DBox[{734, 566, 2047, 531, 2579, 1763, 530, 693, 692, 1762, 1901, 
       2521, 2383, 1761, 2520, 2375, 1760, 2421, 2519, 2046, 2012, 1759, 726, 
       2011, 676, 1758, 2574, 529, 2045, 674, 1757, 2134, 2133, 2584, 1756, 
       1900, 2064, 762, 1755, 831, 2233, 2598, 2431, 1754, 2008, 2044, 2565, 
       2420, 1753, 2382, 2518, 2201, 1899, 1752, 673, 1327, 673}], 
      Line3DBox[{1777, 2137, 2138, 2066, 2020, 2558, 1776, 2346, 2347, 2333, 
       2616, 2434, 1775, 1903, 2527, 2385, 1774, 2526, 2377, 1773, 2433, 2525,
        2339, 2341, 2340, 1772, 2309, 2613, 2310, 2163, 2136, 1771, 2606, 
       2280, 2281, 2153, 2135, 1770, 1860, 2487, 1769, 2018, 2050, 2567, 2424,
        1768, 2524, 2376, 1767, 2017, 2049, 2566, 2423, 1766, 2384, 2523, 
       2016, 1902, 1765, 2422, 2522, 2048, 2015, 1764, 2557, 2019, 2065, 2207,
        2206, 2259}], Line3DBox[{799, 678, 2043, 527, 2573, 1751, 528, 691}]}, 
     {GrayLevel[0.2], 
      Line3DBox[{919, 1207, 1061, 2537, 920, 1568, 1153, 1525, 1526, 1527, 
       2568, 940, 1158, 1457, 2528, 1074, 946, 1570, 1571, 2607, 1078, 954, 
       1247, 2543, 1084, 963, 1254, 2545, 1092, 972, 2494, 1264, 1099, 982, 
       2502, 1108, 2460, 994, 1611, 1612, 1114, 2471, 1005, 1586, 1587, 2610, 
       1120, 1015, 1319, 1320, 2553, 1126, 1023, 1438, 1192, 2583, 1503, 1395,
        1509, 1396, 1520, 1505, 1510, 1327, 1504, 1136, 1038, 2522, 1339, 
       1142, 1048}], Line3DBox[CompressedData["
1:eJwVzq9LQ3EUhvEz/MUsEzFocwgiDOu0GAw208Bm8g9wm8VoNCwtbSoWhzYR
NDoYiGCyTQfT5ASHwaEs2PRzw8NzznvuOd+b3SkWdlMRUcDRSMT6eMTiaER+
MuJYPxyLWNYv6S/1h+pH1OQzsjuuo49rVMwa/JaOOOcL3Nqrytdwr26afaj7
2EjuyVqyb/UPNtGRPcj+1OHGFi94ryf/1E9j1r+2eV7+xF9mJ3zq22cuYsqs
xL9mB9zBmfk2v7tf5h72MMBrkiOHrN0u572zihVMJLftXXEZGbzob3gfc/gH
Xnkx/g==
       "]], Line3DBox[CompressedData["
1:eJwVzT0sg1EUgOHT9CdpMTUhMRAjU4VRR+nOYqmIxM/ErmzFYjGglM1glHTx
szEZlOpAYikbk0Ri9nR48p5zv5v7jSyuz6wlImKWejLiKhVxTTUbMZ2OuDHf
smcv2e/MTe7Zd7bk7NV8rm/6oMv05CI69g8++XH3QL84JJeJeHTvu/ueDrn/
pEf2X2q07M/8mTd02J0XPbYn9ETbeubtiqbYpM/bdd/y5lO98H3LPEeVfnYY
4NK3bV1hlDJjNJzv6iq9/lvQd8aZJ+1sQidZIGOf0iKD/vkPpootqA==
       "]], 
      Line3DBox[CompressedData["
1:eJwVzjtLAmAYhuFXLcmmAls6EEEGgVP/I+gwuPcDaiungsRNF90idKhF95ra
OlKBtRWaS9LaUks1dDncPPf7fMeFrZ2N7URErOM4FbE/EnGAcibiQh6ihKr5
cjRiMR2RQ0d3ZZ4dj3jiN/waz3xX3qLCXzFpT1Xe6/bQ5UVZk3PWHnmdP8hV
HGFK/67Le2sgM3IMy/ov85o93/LFn+/8LWneRAoN80Cf5YXhXWjpPnUNZ5qY
1yXcd8LP+LR7T/k5z+MN/eE53YezK7zN/+SSvT/yFxO6nvWinME/jFcuwQ==

       "]], 
      Line3DBox[{925, 1213, 1214, 2438, 1350, 1215, 1466, 608, 2167, 609, 489,
        627, 2175, 624, 1354, 1235, 1242, 2541, 1236, 1602, 886, 2608, 1577, 
       1576, 1619, 917, 2618, 1614, 1613, 974, 2496, 1101, 985, 2505, 1285, 
       1286, 2463, 997, 1304, 1305, 2474, 1008, 1312, 2550, 1313, 1600, 879, 
       1562, 2604, 1561, 1024, 1195, 2535, 1455, 1196, 1385, 762, 1540, 1200, 
       1377, 2567, 1342, 1343, 1052}], 
      Line3DBox[{926, 1424, 1425, 1426, 1467, 1468, 1378, 1409, 2569, 1408, 
       1352, 1220, 1234, 2539, 1221, 947, 1079, 2451, 956, 1086, 2453, 965, 
       1259, 2546, 1260, 975, 1271, 2548, 1272, 986, 2506, 1287, 1364, 1288, 
       2464, 998, 1306, 1369, 2563, 1307, 1009, 1314, 1371, 2564, 1315, 1017, 
       1127, 2484, 1025, 1132, 2486, 1032, 1439, 2584, 1456, 1440, 1041, 1145,
        2487, 1053}], 
      Line3DBox[{8, 1627, 23, 459, 2053, 38, 1641, 53, 1647, 68, 1657, 83, 
       1670, 424, 98, 1684, 427, 113, 1703, 432, 2033, 128, 438, 2039, 143, 
       441, 2041, 158, 1734, 173, 1743, 188, 1757, 563, 203, 1770, 218}], 
      Line3DBox[{927, 1427, 1063, 928, 1469, 1470, 1379, 1390, 2575, 1070, 
       942, 1222, 1541, 2538, 1075, 948, 2490, 1080, 957, 2491, 1087, 966, 
       1261, 2547, 1094, 976, 1273, 2562, 1361, 1102, 987, 2507, 1289, 1110, 
       2465, 999, 1308, 2549, 1115, 1010, 1316, 2551, 1121, 1018, 2517, 1128, 
       1026, 1333, 2602, 1552, 1133, 1033, 1418, 2574, 1419, 1387, 1523, 1569,
        1138, 1042, 1566, 2606, 1567, 1565, 1146, 1054}], 
      Line3DBox[{929, 1064, 2439, 930, 1223, 1441, 571, 2212, 699, 592, 1442, 
       2529, 828, 949, 829, 1535, 2596, 1081, 958, 1578, 2609, 1579, 1088, 
       967, 1609, 2617, 1610, 1095, 977, 2497, 1274, 1103, 988, 2508, 1111, 
       2466, 1000, 1116, 2475, 1011, 2515, 1122, 1019, 1593, 1594, 2612, 1129,
        1027, 1197, 2555, 1334, 487, 727, 726, 1521, 733, 1043, 2613, 868, 
       1596, 1147, 1055}], Line3DBox[CompressedData["
1:eJwVzi1IQ2EYhuFXnIrnWLYJhhk0GAyiwmBBsJgHw7Ri0T6DSTAYDKaTLOYp
KvsLgiYZmCxmLUuysLJkEFzwMtw8f98L3+rh8X5jKiKqaE5HnOci1nE5H3Ej
X/CbyOSefD0TUZaf+Ec+4+/ol/2ezs1GPNC+/ZRe2QtJRFGfRwFvtpHtxc3Y
vqRLsYA97OLV9muboIaBm3dd4i5FXdeUR/ot77dxoGvrvnVr3nziSPejm+ie
5Vu5RYe6Nu0gp1t236Ub/lriG3yCD+8q3pzwi///kHfsZ/wK/gDyNSlG
       "]], 
      Line3DBox[{933, 1066, 2441, 934, 1072, 2447, 944, 838, 2450, 951, 547, 
       2122, 548, 778, 2253, 318, 464, 2059, 246, 329, 1949, 786, 248, 647, 
       342, 1963, 648, 250, 789, 1977, 353, 252, 2060, 465, 373, 2148, 581, 
       557, 2319, 880, 2320, 822, 1035, 2520, 1140, 1045, 2526, 1149, 1057}], 
      Line3DBox[CompressedData["
1:eJwNzrFKQgEUxvFjVpBYLiFEJNhgQ1hjL5BTTkJLYw+Qb+CsLfUABQmhEhQY
Cu4u1uDQIAhWD2D5Bka/4c//O989596bv6xWrhIRUUY7GXG4GnG9EdGRR/Ib
bs3vXEdhPaLBBzzh0VrEDYd5J2VX7unOMJMXbp/kT/nUTgkn9op8jCN8+NbQ
3tJOxe4ff+nGurT5HJt4NM/5B3u4QA7P+m3v+ZUXeHC/ZW5y17OaroApMvqW
foV3/UdbHuhfcIes7p7T+qnbLr8ii29zj/vYxz8wFip8
       "]], 
      Line3DBox[{937, 1428, 1429, 2443, 1430, 1157, 1542, 771, 2249, 772, 491,
        1237, 2540, 1162, 1599, 877, 1558, 2603, 1557, 960, 1089, 2454, 969, 
       1096, 2456, 979, 2500, 1105, 991, 2511, 1294, 1295, 2469, 1002, 1615, 
       1616, 2478, 1617, 911, 1589, 2611, 1588, 1597, 870, 1529, 2597, 1528, 
       1539, 760, 2536, 1465, 604, 810, 692, 1506, 1336, 1608, 2616, 1606, 
       1607, 1059}], 
      Line3DBox[{938, 1216, 1217, 1471, 1472, 1473, 1380, 1411, 2570, 1410, 
       1353, 1227, 1511, 2594, 1228, 952, 1082, 2452, 961, 1090, 2455, 970, 
       1097, 2457, 980, 1106, 2458, 992, 2512, 1296, 1366, 1297, 2470, 1003, 
       1118, 2479, 1013, 1124, 2482, 1021, 1130, 2485, 1029, 1337, 2556, 1522,
        1338, 1375, 1420, 2579, 1421, 1388, 1423, 1422, 1389, 1345, 2558, 
       1618, 1346, 1347}], 
      Line3DBox[{1047, 1141, 2557, 1344, 1037, 1135, 1397, 1386, 1417, 2573, 
       1416, 1030, 1131, 1326, 2554, 1325, 1022, 1125, 2516, 1014, 1119, 2514,
        1004, 1113, 2513, 993, 2459, 1107, 2501, 981, 1098, 1359, 2561, 1263, 
       971, 1091, 1358, 2560, 1253, 962, 1083, 1356, 2559, 1246, 953, 1077, 
       2489, 945, 1073, 2488, 939, 1068, 2599, 1534, 1533, 918, 1060, 1206, 
       1348}]}, {}, {}}},
   VertexNormals->CompressedData["
1:eJx0e3dcz1/0fwspK7KSSkhpSEjzfUhGSZHISmgYTRqiSUNZUSkrtNCSimTU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    "]],
  Axes->True,
  AxesLabel->{None, None, None},
  AxesOrigin->{Automatic, Automatic, Automatic},
  BoxRatios->{1, 1, 0.4},
  DisplayFunction->Identity,
  FaceGrids->None,
  FaceGridsStyle->Automatic,
  ImageSize->{482.2257329851469, 194.},
  Method->{"DefaultBoundaryStyle" -> Directive[
      GrayLevel[0.3]], "RotationControl" -> "Globe"},
  PlotRange->
   NCache[{{0, 2 Pi}, {Rational[-1, 2] Pi, Rational[1, 2] Pi}, {
     0.7600001703072273, 1.5599118091280344`}}, {{
     0, 6.283185307179586}, {-1.5707963267948966`, 1.5707963267948966`}, {
     0.7600001703072273, 1.5599118091280344`}}],
  PlotRangePadding->{
    Scaled[0.02], 
    Scaled[0.02], 
    Scaled[0.02]},
  Ticks->{Automatic, Automatic, Automatic},
  ViewPoint->{0.004156651460206768, -3.3766963806283807`, 0.2188704532362284},
  
  ViewVertical->{0., 0., 1.}]], "Input",
 CellChangeTimes->{{3.743779117407633*^9, 
  3.7437791188226967`*^9}},ExpressionUUID->"b679d139-858e-46ad-b71c-\
3edd42ac56a7"],

Cell["\<\
We see that the maximum of that sum is reached for  s= Pi. Thus, we plot the \
sum for  s= Pi and  -Pi/2 < t < Pi/2:\
\>", "Text",
 CellChangeTimes->{{3.743779081139629*^9, 3.743779122956805*^9}, {
  3.7437791843086443`*^9, 
  3.7437792135251217`*^9}},ExpressionUUID->"438019f0-b97f-4e6b-95f8-\
c2a699b462c2"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Plot", "[", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"alphabeta", "[", 
      RowBox[{"Pi", ",", "t"}], "]"}], "[", 
     RowBox[{"[", "1", "]"}], "]"}], "+", 
    RowBox[{
     RowBox[{"alphabeta", "[", 
      RowBox[{"Pi", ",", "t"}], "]"}], "[", 
     RowBox[{"[", "2", "]"}], "]"}]}], ",", "  ", 
   RowBox[{"{", 
    RowBox[{"t", ",", 
     RowBox[{
      RowBox[{"-", "Pi"}], "/", "2"}], ",", 
     RowBox[{"Pi", "/", "2"}]}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.743744682665655*^9, 3.743744698993888*^9}, {
  3.74374476366912*^9, 3.743744763869796*^9}, {3.7437477042285767`*^9, 
  3.743747714705127*^9}, {3.743748480433722*^9, 3.743748514152038*^9}, {
  3.7437485735011787`*^9, 3.743748576982758*^9}, {3.7437764011360283`*^9, 
  3.743776409101467*^9}},ExpressionUUID->"5e5fcf1b-6448-4588-afa5-\
f8aa6d552a9c"],

Cell[BoxData[
 GraphicsBox[{{{}, {}, 
    TagBox[
     {RGBColor[0.368417, 0.506779, 0.709798], AbsoluteThickness[1.6], Opacity[
      1.], LineBox[CompressedData["
1:eJwt2Xk4Fd8bAHCEJNmSElEqCUWFFHlJSCH5ChVZypLspES2ZEuSNWuSbBFS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       "]]},
     Annotation[#, "Charting`Private`Tag$23159#1"]& ]}, {}, {}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->{True, True},
  AxesLabel->{None, None},
  AxesOrigin->{0, 0.7605142849459227},
  DisplayFunction->Identity,
  Frame->{{False, False}, {False, False}},
  FrameLabel->{{None, None}, {None, None}},
  FrameTicks->{{Automatic, 
     Charting`ScaledFrameTicks[{Identity, Identity}]}, {Automatic, 
     Charting`ScaledFrameTicks[{Identity, Identity}]}},
  GridLines->{None, None},
  GridLinesStyle->Directive[
    GrayLevel[0.5, 0.4]],
  ImagePadding->All,
  ImageSize->{525., Automatic},
  Method->{
   "DefaultBoundaryStyle" -> Automatic, "DefaultMeshStyle" -> 
    AbsolutePointSize[6], "ScalingFunctions" -> None, 
    "CoordinatesToolOptions" -> {"DisplayFunction" -> ({
        (Identity[#]& )[
         Part[#, 1]], 
        (Identity[#]& )[
         Part[#, 2]]}& ), "CopiedValueFunction" -> ({
        (Identity[#]& )[
         Part[#, 1]], 
        (Identity[#]& )[
         Part[#, 2]]}& )}},
  PlotRange->
   NCache[{{Rational[-1, 2] Pi, Rational[1, 2] Pi}, {0.7605142849459227, 
     1.5599999982323207`}}, {{-1.5707963267948966`, 1.5707963267948966`}, {
    0.7605142849459227, 1.5599999982323207`}}],
  PlotRangeClipping->True,
  PlotRangePadding->{{
     Scaled[0.02], 
     Scaled[0.02]}, {
     Scaled[0.05], 
     Scaled[0.05]}},
  Ticks->{Automatic, Automatic}]], "Output",
 CellChangeTimes->{
  3.743744701485468*^9, 3.743744765273798*^9, 3.743747650084487*^9, 
   3.743747724690919*^9, {3.7437484907043552`*^9, 3.743748515406705*^9}, 
   3.74374857820842*^9, 3.743750408812887*^9, 3.743772811753152*^9, 
   3.7437760618832073`*^9, 3.743776414716332*^9, 
   3.743777058131071*^9},ExpressionUUID->"bca15b4c-7f2e-448d-a717-\
6b8cc4160eeb"]
}, Open  ]],

Cell["The maximum is approximately  reached at  t = 0.38:", "Text",
 CellChangeTimes->{{3.743779228823022*^9, 3.7437792690381308`*^9}, {
  3.74377941245173*^9, 
  3.743779415893757*^9}},ExpressionUUID->"6aec0e7f-48f4-45c4-a17d-\
7d29b8122fdc"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"alphabeta", "[", 
    RowBox[{"Pi", ",", "0.38"}], "]"}], "[", 
   RowBox[{"[", "1", "]"}], "]"}], "+", 
  RowBox[{
   RowBox[{"alphabeta", "[", 
    RowBox[{"Pi", ",", "0.38"}], "]"}], "[", 
   RowBox[{"[", "2", "]"}], "]"}]}]], "Input",
 CellChangeTimes->{{3.743744826527484*^9, 3.743744841889415*^9}, {
  3.7437448737152157`*^9, 3.743744921046142*^9}, {3.743747814384853*^9, 
  3.743747843666905*^9}, {3.743748632708583*^9, 3.743748733037949*^9}, {
  3.743750466844809*^9, 3.743750490034*^9}, {3.7437763199681787`*^9, 
  3.743776323661174*^9}, {3.7437764274570627`*^9, 3.743776491968842*^9}, {
  3.743776525848431*^9, 3.7437765322396917`*^9}, {3.743777079885104*^9, 
  3.7437770833900223`*^9}, {3.7437792850171022`*^9, 
  3.743779425732774*^9}},ExpressionUUID->"1719ef6f-34ab-43f1-81b7-\
5d1c7e77bcd7"],

Cell[BoxData["1.5586431845391893`"], "Output",
 CellChangeTimes->{{3.743744831744928*^9, 3.743744922793922*^9}, 
   3.743747650118266*^9, {3.743747822979433*^9, 3.743747845009037*^9}, {
   3.743748635359253*^9, 3.743748734099901*^9}, {3.7437504732175093`*^9, 
   3.743750490781115*^9}, 3.743772503443878*^9, 3.743772811807926*^9, 
   3.743776061992929*^9, 3.743776324767078*^9, 3.743776434199114*^9, {
   3.7437764653622*^9, 3.743776492910173*^9}, 3.743776533164706*^9, {
   3.743777069256146*^9, 3.743777085408359*^9}, {3.7437792895155354`*^9, 
   3.743779426683867*^9}},ExpressionUUID->"8422c181-02fc-458e-84e7-\
7d5abe08753a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"1", "/", "%"}]], "Input",
 CellChangeTimes->{{3.7437449609452877`*^9, 3.743744969751704*^9}, {
  3.74374671733267*^9, 3.7437467267033052`*^9}, {3.74375050166321*^9, 
  3.743750502944848*^9}, {3.743772509835033*^9, 
  3.743772510066469*^9}},ExpressionUUID->"a1feaff4-fc54-455a-a8aa-\
f324e68f4eda"],

Cell[BoxData["0.6415836606603766`"], "Output",
 CellChangeTimes->{
  3.743744973265089*^9, 3.743746729076248*^9, 3.743747650195483*^9, 
   3.743750503925103*^9, 3.743772511015345*^9, 3.7437728278489447`*^9, 
   3.7437760621032124`*^9, 3.7437764977623253`*^9, 3.74377653729574*^9, {
   3.743777069358397*^9, 3.7437770880464478`*^9}, 
   3.74377943207858*^9},ExpressionUUID->"7466cd9e-4c68-481b-92a6-\
3d8185620374"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Computation of a Gamma for the Pickers. ", "Subsubsection",
 CellChangeTimes->{{3.743779483753496*^9, 
  3.743779502742751*^9}},ExpressionUUID->"c2f3e625-8c69-4bcc-9989-\
1f5bcaf26115"],

Cell["\<\
Below is the computation of a gamma correction, f,  that sends the values \
(0..L) to the 
values (a..L), where a= 40,  L = 255. It is a cubic polynomial .  We impose \
the conditions that f[0]= a, 
f[L] = L,  f1[0]= 1, f1[L]= 0:\
\>", "Text",
 CellChangeTimes->{{3.743779559385997*^9, 3.743779702958181*^9}, {
  3.743779847952263*^9, 3.7437798536442747`*^9}, {3.7437799013533792`*^9, 
  3.7437799807703733`*^9}, {3.743780098049305*^9, 
  3.743780102055243*^9}},ExpressionUUID->"fd2964b2-72e0-40a1-ad5c-\
1ea5a7917f65"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"f", "[", "t_", "]"}], "=", " ", 
   RowBox[{"a", "+", " ", 
    RowBox[{"b", " ", "t"}], " ", "+", " ", 
    RowBox[{"c", " ", 
     RowBox[{"t", "^", "2"}]}], " ", "+", " ", 
    RowBox[{"d", " ", 
     RowBox[{"t", "^", "3"}]}]}]}], ";", " ", 
  RowBox[{
   RowBox[{"f1", "[", "t_", "]"}], "=", " ", 
   RowBox[{"b", " ", "+", " ", 
    RowBox[{"2", " ", "c", " ", "t"}], " ", "+", " ", 
    RowBox[{"3", " ", "d", " ", 
     RowBox[{"t", "^", "2"}]}]}]}], ";", " ", 
  RowBox[{
  Cell["", "Text",ExpressionUUID->"fc47181b-e955-4c6a-b1a3-b8e76a91b507"], 
   Cell["", "Text",ExpressionUUID->"9c9707fe-e8c3-413b-8536-92b0710de05c"], 
   " "}]}]], "Input",
 CellChangeTimes->{{3.743769670993062*^9, 3.743769746232009*^9}, {
  3.7437699911910963`*^9, 3.7437700023474894`*^9}, {3.7437797127322407`*^9, 
  3.743779806012712*^9}},ExpressionUUID->"102cb8c5-bea9-4746-adf9-\
6d392c456599"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Solve", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"f", "[", "L", "]"}], ",", " ", 
      RowBox[{"f1", "[", "0", "]"}], ",", " ", 
      RowBox[{"f1", "[", "L", "]"}]}], "}"}], "\[Equal]", 
    RowBox[{"{", 
     RowBox[{"L", ",", " ", "1", ",", " ", "0"}], "}"}]}], ",", " ", 
   RowBox[{"{", 
    RowBox[{"b", ",", "c", ",", "d"}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.743769702063547*^9, 3.743769705516247*^9}, {
  3.743769759685858*^9, 3.7437698043706284`*^9}, {3.7437698937773943`*^9, 
  3.743769933517912*^9}},ExpressionUUID->"96cd7254-fc11-467d-a812-\
431c7fb62579"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{"b", "\[Rule]", "1"}], ",", 
    RowBox[{"c", "\[Rule]", 
     FractionBox["3", "1445"]}], ",", 
    RowBox[{"d", "\[Rule]", 
     RowBox[{"-", 
      FractionBox["7", "663255"]}]}]}], "}"}], "}"}]], "Output",
 CellChangeTimes->{3.743770006060441*^9, 3.7437760623159437`*^9, 
  3.74377706944563*^9, 
  3.743779838233786*^9},ExpressionUUID->"3a934b8b-e2a9-490b-a632-\
ba7c7cd9beff"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"L", "=", " ", "255"}]], "Input",
 CellChangeTimes->{{3.743770066392682*^9, 3.743770097430231*^9}, {
  3.743770377357185*^9, 3.7437703813572893`*^9}, {3.7437704811279583`*^9, 
  3.743770481764786*^9}},ExpressionUUID->"0082346c-2c3e-4bd0-84bf-\
3c97d252284b"],

Cell[BoxData["255"], "Output",
 CellChangeTimes->{3.7437704837924423`*^9, 3.743776062333641*^9, 
  3.7437770694602757`*^9, 3.74377983825419*^9, 
  3.7437800118240013`*^9},ExpressionUUID->"99f93f85-1802-4c0e-93a7-\
552b627d13dd"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"a", "=", " ", "40"}]], "Input",
 CellChangeTimes->{{3.7437704861514797`*^9, 
  3.7437704881352386`*^9}},ExpressionUUID->"3e282c24-4e0f-4096-9702-\
27a15841b615"],

Cell[BoxData["40"], "Output",
 CellChangeTimes->{3.743770489312078*^9, 3.743776062433703*^9, 
  3.743777069512989*^9, 3.743779838333897*^9, 
  3.743780011927732*^9},ExpressionUUID->"bf9bd980-5fcd-4192-9997-\
135fae813903"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"g", "[", "t_", "]"}], "=", " ", 
  RowBox[{"a", "+", "t", "+", 
   FractionBox[
    RowBox[{
     RowBox[{"(", 
      RowBox[{"255", "-", 
       RowBox[{"3", " ", "a"}]}], ")"}], " ", 
     SuperscriptBox["t", "2"]}], "65025"], "+", 
   FractionBox[
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "255"}], "+", 
       RowBox[{"2", " ", "a"}]}], ")"}], " ", 
     SuperscriptBox["t", "3"]}], "16581375"]}]}]], "Input",
 CellChangeTimes->{{3.743770258137878*^9, 3.743770259434595*^9}, {
   3.743770303641288*^9, 3.743770306341242*^9}, {3.743770356832985*^9, 
   3.743770362231192*^9}, 3.743770438182729*^9, 
   3.7437705199219503`*^9},ExpressionUUID->"c423b437-7d8d-4125-880c-\
8d476c859a63"],

Cell[BoxData[
 RowBox[{"40", "+", "t", "+", 
  FractionBox[
   RowBox[{"3", " ", 
    SuperscriptBox["t", "2"]}], "1445"], "-", 
  FractionBox[
   RowBox[{"7", " ", 
    SuperscriptBox["t", "3"]}], "663255"]}]], "Output",
 CellChangeTimes->{3.743770521256621*^9, 3.743776062518668*^9, 
  3.743777069566433*^9, 3.743779838410887*^9, 
  3.743780011947504*^9},ExpressionUUID->"48411a20-c23b-4c7d-8155-\
1d2a4089ccb7"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Plot", "[", 
  RowBox[{
   RowBox[{"g", "[", "t", "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"t", ",", "0", ",", "L"}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.7437701802476187`*^9, 3.74377019422824*^9}, {
  3.743770288516848*^9, 
  3.74377032070708*^9}},ExpressionUUID->"fd605eb9-3b97-4452-98dd-\
4444677a04f9"],

Cell[BoxData[
 GraphicsBox[{{{}, {}, 
    TagBox[
     {RGBColor[0.368417, 0.506779, 0.709798], AbsoluteThickness[1.6], Opacity[
      1.], LineBox[CompressedData["
1:eJwV1Ak0VukfB3CVZai0TKOVso5kz5SkfMtrQgtDLxGv3Wt53/uUGkSZlJI0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       "]]},
     Annotation[#, "Charting`Private`Tag$23896#1"]& ]}, {}, {}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->{True, True},
  AxesLabel->{None, None},
  AxesOrigin->{0, 40.00000520408169},
  DisplayFunction->Identity,
  Frame->{{False, False}, {False, False}},
  FrameLabel->{{None, None}, {None, None}},
  FrameTicks->{{Automatic, 
     Charting`ScaledFrameTicks[{Identity, Identity}]}, {Automatic, 
     Charting`ScaledFrameTicks[{Identity, Identity}]}},
  GridLines->{None, None},
  GridLinesStyle->Directive[
    GrayLevel[0.5, 0.4]],
  ImagePadding->All,
  Method->{
   "DefaultBoundaryStyle" -> Automatic, "DefaultMeshStyle" -> 
    AbsolutePointSize[6], "ScalingFunctions" -> None, 
    "CoordinatesToolOptions" -> {"DisplayFunction" -> ({
        (Identity[#]& )[
         Part[#, 1]], 
        (Identity[#]& )[
         Part[#, 2]]}& ), "CopiedValueFunction" -> ({
        (Identity[#]& )[
         Part[#, 1]], 
        (Identity[#]& )[
         Part[#, 2]]}& )}},
  PlotRange->{{0, 255}, {40.00000520408169, 254.99999999999986`}},
  PlotRangeClipping->True,
  PlotRangePadding->{{
     Scaled[0.02], 
     Scaled[0.02]}, {
     Scaled[0.05], 
     Scaled[0.05]}},
  Ticks->{Automatic, Automatic}]], "Output",
 CellChangeTimes->{{3.7437701951523647`*^9, 3.743770208183134*^9}, 
   3.7437702900681543`*^9, 3.743770321581318*^9, {3.743770366031543*^9, 
   3.743770391713284*^9}, 3.743770530255221*^9, 3.743776062726845*^9, 
   3.7437770696785803`*^9, 3.743779838566082*^9, 
   3.7437800120866137`*^9},ExpressionUUID->"ed8165db-b3db-4706-b70f-\
25819d84be65"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"u", "=", " ", 
   RowBox[{"t", "/", "L"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.7437711985941772`*^9, 
  3.743771229358719*^9}},ExpressionUUID->"6c3841ed-6acc-4597-b823-\
683112af604a"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"test", "=", " ", 
  RowBox[{"a", " ", "+", " ", "t", " ", "+", " ", 
   RowBox[{
    RowBox[{"u", "^", "2"}], 
    RowBox[{"(", 
     RowBox[{"255", "-", 
      RowBox[{"3", "a"}], " ", "+", 
      RowBox[{"u", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"2", " ", "a"}], "-", "255"}], ")"}]}]}], 
     ")"}]}]}]}]], "Input",
 CellChangeTimes->{{3.743771152402133*^9, 3.743771164979383*^9}, {
  3.743771240041545*^9, 
  3.743771296947206*^9}},ExpressionUUID->"339fd2cf-1737-4816-bef8-\
7520c0454d99"],

Cell[BoxData[
 RowBox[{"40", "+", "t", "+", 
  FractionBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{"135", "-", 
      FractionBox[
       RowBox[{"35", " ", "t"}], "51"]}], ")"}], " ", 
    SuperscriptBox["t", "2"]}], "65025"]}]], "Output",
 CellChangeTimes->{{3.7437712981029882`*^9, 3.7437713055955753`*^9}, 
   3.74377606283325*^9, 3.7437770697498217`*^9, 
   3.743780022564662*^9},ExpressionUUID->"7ac5fe84-5c1f-4877-9a5e-\
84fcd2ff42bc"]
}, Open  ]],

Cell["Now, we verify that this formula is the required one", "Text",
 CellChangeTimes->{{3.7437800293433228`*^9, 3.743780070460863*^9}, 
   3.7437801187156878`*^9},ExpressionUUID->"71e1b465-efcf-4e9b-846e-\
80e771d3e6d1"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Simplify", "[", 
  RowBox[{"test", " ", "-", " ", 
   RowBox[{"g", "[", "t", "]"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.743771316463297*^9, 
  3.743771330892194*^9}},ExpressionUUID->"da4ce78a-3594-48bb-bb7d-\
7a440460a51e"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.743771331718968*^9, 3.743776062915208*^9, 
  3.74377706980586*^9, 
  3.74378002258083*^9},ExpressionUUID->"593e0723-a1ab-4409-b4ed-0dd687c60870"]
}, Open  ]]
}, Open  ]]
},
WindowSize->{808, 911},
WindowMargins->{{7, Automatic}, {29, Automatic}},
FrontEndVersion->"11.2 for Linux x86 (64-bit) (September 10, 2017)",
StyleDefinitions->"Default.nb"
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{}
*)
(*CellTagsIndex
CellTagsIndex->{}
*)
(*NotebookFileOutline
Notebook[{
Cell[CellGroupData[{
Cell[580, 22, 545, 11, 123, "Subsubsection",ExpressionUUID->"6124fcac-9fe3-4bb3-969e-d6e33e9028fd"],
Cell[CellGroupData[{
Cell[1150, 37, 454, 8, 46, "Input",ExpressionUUID->"998de7f7-42be-48cf-b17b-04593dd453ed"],
Cell[1607, 47, 543, 8, 50, "Output",ExpressionUUID->"f94f41d5-ae54-4bc6-aa2e-fb24b8dc4f50"]
}, Open  ]],
Cell[2165, 58, 366, 8, 58, "Text",ExpressionUUID->"06158f52-b38b-4b9c-bfb0-df3a41ba5a79"],
Cell[2534, 68, 901, 24, 55, "Input",ExpressionUUID->"fe36f75e-f595-447f-9120-7821b7b350fa"],
Cell[3438, 94, 3238, 49, 380, "Text",ExpressionUUID->"dd688018-ffd5-40ce-b416-47467de11072"],
Cell[6679, 145, 3127, 69, 277, "Input",ExpressionUUID->"7b1aafaf-13eb-4a84-a5af-b1756ba86932"],
Cell[9809, 216, 452, 8, 58, "Text",ExpressionUUID->"25a1a01e-a849-4602-b1c8-0064263790aa"],
Cell[10264, 226, 895, 24, 70, "Input",ExpressionUUID->"ff27416d-c794-4e52-9f9f-5600ad1bcd9b"],
Cell[11162, 252, 151147, 2477, 207, InheritFromParent,ExpressionUUID->"b679d139-858e-46ad-b71c-3edd42ac56a7"],
Cell[162312, 2731, 319, 7, 58, "Text",ExpressionUUID->"438019f0-b97f-4e6b-95f8-c2a699b462c2"],
Cell[CellGroupData[{
Cell[162656, 2742, 861, 22, 46, "Input",ExpressionUUID->"5e5fcf1b-6448-4588-afa5-f8aa6d552a9c"],
Cell[163520, 2766, 11263, 204, 364, "Output",ExpressionUUID->"bca15b4c-7f2e-448d-a717-6b8cc4160eeb"]
}, Open  ]],
Cell[174798, 2973, 243, 4, 35, "Text",ExpressionUUID->"6aec0e7f-48f4-45c4-a17d-7d29b8122fdc"],
Cell[CellGroupData[{
Cell[175066, 2981, 852, 18, 46, "Input",ExpressionUUID->"1719ef6f-34ab-43f1-81b7-5d1c7e77bcd7"],
Cell[175921, 3001, 629, 9, 50, "Output",ExpressionUUID->"8422c181-02fc-458e-84e7-7d5abe08753a"]
}, Open  ]],
Cell[CellGroupData[{
Cell[176587, 3015, 321, 6, 46, "Input",ExpressionUUID->"a1feaff4-fc54-455a-a8aa-f324e68f4eda"],
Cell[176911, 3023, 414, 7, 50, "Output",ExpressionUUID->"7466cd9e-4c68-481b-92a6-3d8185620374"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[177374, 3036, 191, 3, 45, "Subsubsection",ExpressionUUID->"c2f3e625-8c69-4bcc-9989-1f5bcaf26115"],
Cell[177568, 3041, 529, 11, 81, "Text",ExpressionUUID->"fd2964b2-72e0-40a1-ad5c-1ea5a7917f65"],
Cell[178100, 3054, 929, 23, 49, "Input",ExpressionUUID->"102cb8c5-bea9-4746-adf9-6d392c456599"],
Cell[CellGroupData[{
Cell[179054, 3081, 641, 16, 46, "Input",ExpressionUUID->"96cd7254-fc11-467d-a812-431c7fb62579"],
Cell[179698, 3099, 450, 13, 68, "Output",ExpressionUUID->"3a934b8b-e2a9-490b-a632-ba7c7cd9beff"]
}, Open  ]],
Cell[CellGroupData[{
Cell[180185, 3117, 281, 5, 46, "Input",ExpressionUUID->"0082346c-2c3e-4bd0-84bf-3c97d252284b"],
Cell[180469, 3124, 228, 4, 50, "Output",ExpressionUUID->"99f93f85-1802-4c0e-93a7-552b627d13dd"]
}, Open  ]],
Cell[CellGroupData[{
Cell[180734, 3133, 185, 4, 46, "Input",ExpressionUUID->"3e282c24-4e0f-4096-9702-27a15841b615"],
Cell[180922, 3139, 222, 4, 50, "Output",ExpressionUUID->"bf9bd980-5fcd-4192-9997-135fae813903"]
}, Open  ]],
Cell[CellGroupData[{
Cell[181181, 3148, 747, 21, 69, InheritFromParent,ExpressionUUID->"c423b437-7d8d-4125-880c-8d476c859a63"],
Cell[181931, 3171, 414, 11, 73, "Output",ExpressionUUID->"48411a20-c23b-4c7d-8155-1d2a4089ccb7"]
}, Open  ]],
Cell[CellGroupData[{
Cell[182382, 3187, 343, 9, 46, "Input",ExpressionUUID->"fd605eb9-3b97-4452-98dd-4444677a04f9"],
Cell[182728, 3198, 4849, 97, 254, "Output",ExpressionUUID->"ed8165db-b3db-4706-b70f-25819d84be65"]
}, Open  ]],
Cell[187592, 3298, 224, 6, 46, "Input",ExpressionUUID->"6c3841ed-6acc-4597-b823-683112af604a"],
Cell[CellGroupData[{
Cell[187841, 3308, 541, 16, 46, "Input",ExpressionUUID->"339fd2cf-1737-4816-bef8-7520c0454d99"],
Cell[188385, 3326, 446, 12, 80, "Output",ExpressionUUID->"7ac5fe84-5c1f-4877-9a5e-84fcd2ff42bc"]
}, Open  ]],
Cell[188846, 3341, 221, 3, 35, "Text",ExpressionUUID->"71e1b465-efcf-4e9b-846e-80e771d3e6d1"],
Cell[CellGroupData[{
Cell[189092, 3348, 252, 6, 46, "Input",ExpressionUUID->"da4ce78a-3594-48bb-bb7d-7a440460a51e"],
Cell[189347, 3356, 195, 3, 50, "Output",ExpressionUUID->"593e0723-a1ab-4409-b4ed-0dd687c60870"]
}, Open  ]]
}, Open  ]]
}
]
*)

